#include "./src/IO/hdf5_interface.h"


void hdf5_write_0d_double(const char* namefile, const char* namegroup, const char* namefield, double data){
	hsize_t dims[1];

	hid_t file_id;
	hid_t group_id;
	hid_t dataset_id;

	hid_t dataspace_id;
	
	herr_t status;

	int exists;
	struct stat buffer;

	int rank;
#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#else
	rank = 0;
#endif

	exists = (stat(namefile, &buffer) == 0);

	if (rank == 0){
		if (exists){
			// The file exists --> open it
			file_id = H5Fopen(namefile, H5F_ACC_RDWR, H5P_DEFAULT);
		}
		else{
			// The file does not exist --> create it
			file_id = H5Fcreate(namefile, H5F_ACC_RDWR, H5P_DEFAULT, H5P_DEFAULT);
		}

		// Test if the group exists (return FALSE if the file has just been created)
		exists = (exists && ((int)(H5Lexists(file_id, namegroup, H5P_DEFAULT)) > 0));

		if (exists){
			// The group exists --> open it
			group_id = H5Gopen2(file_id, namegroup, H5P_DEFAULT);
		}
		else{
			// The groupe does not exist --> create it
			group_id = H5Gcreate1(file_id, namegroup, 0);
		}

		// Create dataspace for global data
		dims[0] = 1;

		dataspace_id = H5Screate_simple(1, dims, NULL);

		// Test if the dataset exists (return FALSE if the file or the group has just been created)
		exists = (exists && ((int)(H5Lexists(group_id, namefield, H5P_DEFAULT)) > 0));

		if (exists){
			// The dataset exists --> open it
			dataset_id = H5Dopen2(group_id, namefield, H5P_DEFAULT);
		}
		else{
			// The dataset does not exist --> create global dataset inside the dataspace
			dataset_id = H5Dcreate(group_id, namefield, H5T_NATIVE_DOUBLE, dataspace_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
		}

		// Write the data
		status = H5Dwrite(dataset_id, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, &data);

		// Close the dataset
		status = H5Dclose(dataset_id);

		// Close the dataspace
		status = H5Sclose(dataspace_id);

		// Close the group
		status = H5Gclose(group_id);

		// Close the file
		status = H5Fclose(file_id);

	}

#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
#endif
	H5garbage_collect();
#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
#endif
}


void hdf5_write_0d_int(const char* namefile, const char* namegroup, const char* namefield, int data){
	hsize_t dims[1];
	
	hid_t file_id;
	hid_t group_id;
	hid_t dataset_id;

	hid_t dataspace_id;

	herr_t status;

	int exists;
	struct stat buffer;
	
	int rank;
#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#else
	rank = 0;
#endif

	exists = (stat(namefile, &buffer) == 0);

	if (rank == 0){
		if (exists){
			// The file exists --> open it
			file_id = H5Fopen(namefile, H5F_ACC_RDWR, H5P_DEFAULT);
		}
		else{
			// The file does not exist --> create it
			file_id = H5Fcreate(namefile, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
		}

		// Test if the group exists (return FALSE if the file has just been created)
		exists = (exists && ((int)(H5Lexists(file_id, namegroup, H5P_DEFAULT)) > 0));

		if (exists){
			// The group exists --> open it
			group_id = H5Gopen2(file_id, namegroup, H5P_DEFAULT);
		}
		else{
			// The groupe does not exist --> create it
			group_id = H5Gcreate1(file_id, namegroup, 0);
		}

		// Create dataspace for global data
		dims[0] = 1;
		dataspace_id = H5Screate_simple(1, dims, NULL);

		// Test if the dataset exists (return FALSE if the file or the group has just been created)
		exists = (exists && ((int)(H5Lexists(group_id, namefield, H5P_DEFAULT)) > 0));

		if (exists){
			// The dataset exists --> open it
			dataset_id = H5Dopen2(group_id, namefield, H5P_DEFAULT);
		}
		else{
			// The dataset does not exist --> create global dataset inside the dataspace
			dataset_id = H5Dcreate(group_id, namefield, H5T_NATIVE_INT, dataspace_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
		}

		// Write data
		status = H5Dwrite(dataset_id, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &data);

		// Close the dataset
		status = H5Dclose(dataset_id);

		// Close the dataspace
		status = H5Sclose(dataspace_id);

		// Close the group
		status = H5Gclose(group_id);

		// Close the file
		status = H5Fclose(file_id);
	}

#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
#endif
	H5garbage_collect();
#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
#endif
}


void hdf5_write_2d(const char* namefile, const char* namegroup, const char* namefield, double** floc, PARAMETERS parameters){
	hsize_t dims[2];
	hsize_t locdims[2];
	hsize_t offsets[2];

	hid_t plist_access_id, plist_write_id;
	hid_t file_id;
	hid_t group_id;
	hid_t dataset_id;

	hid_t dataspace_id;
	hid_t hyperslab_global_id, hyperslab_local_id;
	hid_t memspace_id;
	
	herr_t status;

	int exists;
	struct stat buffer;
	
	hid_t att_id;
	hid_t attspace_id;

	int NVXLOC, NVYLOC, IVX0LOC, IVY0LOC;
	NVXLOC = (int)(parameters.nvxloc);
	NVYLOC = (int)(parameters.nvyloc);
	IVX0LOC = (int)(parameters.ivx0loc);
	IVY0LOC = (int)(parameters.ivy0loc);

#ifdef key_phdf5
	// Set access property list for IO access
	plist_access_id = H5Pcreate(H5P_FILE_ACCESS);
	MPI_Barrier(MPI_COMM_WORLD);
	status = H5Pset_fapl_mpio(plist_access_id, MPI_COMM_WORLD, MPI_INFO_NULL);

	// Impose strong file closure at the end of the routine
	status = H5Pset_fclose_degree(plist_access_id, H5F_CLOSE_STRONG);
#else
	plist_access_id = H5P_DEFAULT;
#endif

	// Test if file exists
	exists = (stat(namefile, &buffer) == 0);

	if (exists){
		// The file exists --> open it
		file_id = H5Fopen(namefile, H5F_ACC_RDWR, plist_access_id);
	}
	else{
		// The file does not exist --> create it
		file_id = H5Fcreate(namefile, H5F_ACC_TRUNC, H5P_DEFAULT, plist_access_id);
	}

	// Test if the group exists (return FALSE if the file has just been created)
	exists = (exists && ((int)(H5Lexists(file_id, namegroup, H5P_DEFAULT)) > 0));

	if (exists){
		// The group exists --> open it
		group_id = H5Gopen2(file_id, namegroup, H5P_DEFAULT);
	}
	else{
		// The group does not exist --> create it
		group_id = H5Gcreate1(file_id, namegroup, 0);
	}

	// Create dataspace for global data
	dims[0] = parameters.nvx;
	dims[1] = parameters.nvy;
#ifdef key_phdf5
	// Consider it as a hyperslab
	hyperslab_global_id = H5Screate_simple(2, dims, NULL);
#else
	dataspace_id = H5Screate_simple(2, dims, NULL);
#endif

	// Test if the dataset exists (return FALSE if the file or the group has just been created)
	exists = (exists && ((int)(H5Lexists(group_id, namefield, H5P_DEFAULT)) > 0));

	if (exists){
		// The dataset exists --> open it
		dataset_id = H5Dopen1(group_id, namefield);
	}
	else{
		// The dataset does not exist --> create it
#ifdef key_phdf5
		dataset_id = H5Dcreate(group_id, namefield, H5T_NATIVE_DOUBLE, hyperslab_global_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
#else
		dataset_id = H5Dcreate(group_id, namefield, H5T_NATIVE_DOUBLE, dataspace_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
#endif
	}

#ifdef key_phdf5
	// Each processor defines its own dataset in memory and write its part in the hyperslab associated to
	// the global dataset previously defined
	locdims[0] = NVXLOC;
	locdims[1] = NVYLOC;
	offsets[0] = IVX0LOC;
	offsets[1] = IVY0LOC;
	memspace_id = H5Screate_simple(2, locdims, NULL);

	// Select the hyperslab corresponding to the global dataset
	hyperslab_local_id = H5Dget_space(dataset_id);
    status = H5Sselect_hyperslab(hyperslab_local_id, H5S_SELECT_SET, offsets, NULL, locdims, NULL);

	// Create property list for collective dataset write
	plist_write_id = H5Pcreate(H5P_DATASET_XFER);
	status = H5Pset_dxpl_mpio(plist_write_id, H5FD_MPIO_COLLECTIVE);

	// Write local data
	status = H5Dwrite(dataset_id, H5T_NATIVE_DOUBLE, memspace_id, hyperslab_local_id, plist_write_id, floc[0]);
#else
	// Write data
	status = H5Dwrite(dataset_id, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, floc[0]);
#endif

	attspace_id = H5Screate(H5S_SCALAR);
	if (exists){
		att_id = H5Aopen(dataset_id, "nvx", H5P_DEFAULT);
		if (att_id < 0)
			att_id = H5Acreate2(dataset_id, "nvx", H5T_NATIVE_INT, attspace_id, H5P_DEFAULT, H5P_DEFAULT);
	}
	else
		att_id = H5Acreate2(dataset_id, "nvx", H5T_NATIVE_INT, attspace_id, H5P_DEFAULT, H5P_DEFAULT);
	status = H5Awrite(att_id, H5T_NATIVE_INT, &(parameters.nvx));
	status = H5Sclose(attspace_id);
	status = H5Aclose(att_id);

	attspace_id = H5Screate(H5S_SCALAR);
	if (exists == 1){
		att_id = H5Aopen(dataset_id, "nvy", H5P_DEFAULT);
		if (att_id < 0)
			att_id = H5Acreate2(dataset_id, "nvy", H5T_NATIVE_INT, attspace_id, H5P_DEFAULT, H5P_DEFAULT);
	}
	else
		att_id = H5Acreate2(dataset_id, "nvy", H5T_NATIVE_INT, attspace_id, H5P_DEFAULT, H5P_DEFAULT);
	status = H5Awrite(att_id, H5T_NATIVE_INT, &(parameters.nvy));
	status = H5Sclose(attspace_id);
	status = H5Aclose(att_id);
	
#ifdef key_phdf5
	// Close the write access property, the hyperslab and the memory space for local data
	status = H5Pclose(plist_write_id);
	status = H5Sclose(hyperslab_local_id);
	status = H5Sclose(memspace_id);
#endif

	// Close the dataset
	status = H5Dclose(dataset_id);

	// Close the hyperslab/dataspace for global data
#ifdef key_phdf5
	status = H5Sclose(hyperslab_global_id);
#else
	status = H5Sclose(dataspace_id);
#endif

	// Close the group
	status = H5Gclose(group_id);

#ifdef key_phdf5
	// Close the access property
	status = H5Pclose(plist_access_id);

#endif

	// Close the file
	status = H5Fclose(file_id);

#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
#endif
	H5garbage_collect();
#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
#endif
}




void hdf5_write_3d(const char* namefile, const char* namegroup, const char* namefield, double*** floc, PARAMETERS parameters){
	hsize_t dims[3];
	hsize_t locdims[3];
	hsize_t offsets[3];

	hid_t plist_access_id, plist_write_id;
	hid_t file_id;
	hid_t group_id;
	hid_t dataset_id;

	hid_t dataspace_id;
	hid_t hyperslab_global_id, hyperslab_local_id;
	hid_t memspace_id;
	
	herr_t status;

	int exists;
	struct stat buffer;
	
	hid_t att_id;
	hid_t attspace_id;

	int NVXLOC, NVYLOC, NVZLOC, IVX0LOC, IVY0LOC, IVZ0LOC;
	NVXLOC = (int)(parameters.nvxloc);
	NVYLOC = (int)(parameters.nvyloc);
	NVZLOC = (int)(parameters.nvzloc);
	IVX0LOC = (int)(parameters.ivx0loc);
	IVY0LOC = (int)(parameters.ivy0loc);
	IVZ0LOC = (int)(parameters.ivz0loc);

	
#ifdef key_phdf5
	// Set access property list for IO access
	plist_access_id = H5Pcreate(H5P_FILE_ACCESS);
	status = H5Pset_fapl_mpio(plist_access_id, MPI_COMM_WORLD, MPI_INFO_NULL);
	
	// Impose strong file closure at the end of the routine
	status = H5Pset_fclose_degree(plist_access_id, H5F_CLOSE_STRONG);
#else
	plist_access_id = H5P_DEFAULT;
#endif

	// Test if file exists
	exists = (stat(namefile, &buffer) == 0);

	if (exists){
		// The file exists --> open it
		file_id = H5Fopen(namefile, H5F_ACC_RDWR, plist_access_id);
	}
	else{
		// The file does not exist --> create it
		file_id = H5Fcreate(namefile, H5F_ACC_TRUNC, H5P_DEFAULT, plist_access_id);
	}

	// Test if the group exists (return FALSE if the file has just been created)
	exists = (exists && ((int)(H5Lexists(file_id, namegroup, H5P_DEFAULT)) > 0));

	if (exists){
		// The group exists --> open it
		group_id = H5Gopen2(file_id, namegroup, H5P_DEFAULT);
	}
	else{
		// The group does not exist --> create it
		group_id = H5Gcreate1(file_id, namegroup, 0);
	}

	// Create dataspace for global data
	dims[0] = parameters.nvx;
	dims[1] = parameters.nvy;
	dims[2] = parameters.nvz;
#ifdef key_phdf5
	// Consider it as a hyperslab
	hyperslab_global_id = H5Screate_simple(3, dims, NULL);
#else
	dataspace_id = H5Screate_simple(3, dims, NULL);
#endif

	// Test if the dataset exists (return FALSE if the file or the group has just been created)
	exists = (exists && ((int)(H5Lexists(group_id, namefield, H5P_DEFAULT)) > 0));

	if (exists){
		// The dataset exists --> open it
		dataset_id = H5Dopen1(group_id, namefield);
	}
	else{
		// The dataset does not exist --> create it
#ifdef key_phdf5
		dataset_id = H5Dcreate(group_id, namefield, H5T_NATIVE_DOUBLE, hyperslab_global_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
	}
#else
		dataset_id = H5Dcreate(group_id, namefield, H5T_NATIVE_DOUBLE, dataspace_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
	}
#endif

#ifdef key_phdf5
	// Each processor defines its own dataset in memory and write its part in the hyperslab associated to
	// the global dataset previously defined
	locdims[0] = NVXLOC;
	locdims[1] = NVYLOC;
	locdims[2] = NVZLOC;
 	offsets[0] = IVX0LOC;
	offsets[1] = IVY0LOC;
	offsets[2] = IVZ0LOC;
	memspace_id = H5Screate_simple(3, locdims, NULL);

	// Select the hyperslab corresponding to the global dataset
	hyperslab_local_id = H5Dget_space(dataset_id);
    status = H5Sselect_hyperslab(hyperslab_local_id, H5S_SELECT_SET, offsets, NULL, locdims, NULL);

	// Create property list for collective dataset write
	plist_write_id = H5Pcreate(H5P_DATASET_XFER);
	status = H5Pset_dxpl_mpio(plist_write_id, H5FD_MPIO_COLLECTIVE);

	// Write local data
	status = H5Dwrite(dataset_id, H5T_NATIVE_DOUBLE, memspace_id, hyperslab_local_id, plist_write_id, floc[0][0]);
#else
	// Write data
	status = H5Dwrite(dataset_id, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, floc[0][0]);
#endif

	attspace_id = H5Screate(H5S_SCALAR);
	if (exists){
		att_id = H5Aopen(dataset_id, "nvx", H5P_DEFAULT);
		if (att_id < 0)
			att_id = H5Acreate2(dataset_id, "nvx", H5T_NATIVE_INT, attspace_id, H5P_DEFAULT, H5P_DEFAULT);
	}
	else
		att_id = H5Acreate2(dataset_id, "nvx", H5T_NATIVE_INT, attspace_id, H5P_DEFAULT, H5P_DEFAULT);
	status = H5Awrite(att_id, H5T_NATIVE_INT, &(parameters.nvx));
	status = H5Sclose(attspace_id);

	status = H5Aclose(att_id);

	attspace_id = H5Screate(H5S_SCALAR);
	if (exists == 1){
		att_id = H5Aopen(dataset_id, "nvy", H5P_DEFAULT);
		if (att_id < 0)
			att_id = H5Acreate2(dataset_id, "nvy", H5T_NATIVE_INT, attspace_id, H5P_DEFAULT, H5P_DEFAULT);
	}
	else
		att_id = H5Acreate2(dataset_id, "nvy", H5T_NATIVE_INT, attspace_id, H5P_DEFAULT, H5P_DEFAULT);
	status = H5Awrite(att_id, H5T_NATIVE_INT, &(parameters.nvy));
	status = H5Sclose(attspace_id);
	status = H5Aclose(att_id);

	attspace_id = H5Screate(H5S_SCALAR);
	if (exists == 1){
		att_id = H5Aopen(dataset_id, "nvz", H5P_DEFAULT);
		if (att_id < 0)
			att_id = H5Acreate2(dataset_id, "nvz", H5T_NATIVE_INT, attspace_id, H5P_DEFAULT, H5P_DEFAULT);
	}
	else
		att_id = H5Acreate2(dataset_id, "nvz", H5T_NATIVE_INT, attspace_id, H5P_DEFAULT, H5P_DEFAULT);
	status = H5Awrite(att_id, H5T_NATIVE_INT, &(parameters.nvz));
	status = H5Sclose(attspace_id);
	status = H5Aclose(att_id);
	
#ifdef key_phdf5
	// Close the write access property, the hyperslab and the memory space for local data
	status = H5Pclose(plist_write_id);
	status = H5Sclose(hyperslab_local_id);
	status = H5Sclose(memspace_id);
#endif

	// Close the dataset
	status = H5Dclose(dataset_id);

	// Close the hyperslab/dataspace for global data
#ifdef key_phdf5
	status = H5Sclose(hyperslab_global_id);
#else
	status = H5Sclose(dataspace_id);
#endif

	// Close the group
	status = H5Gclose(group_id);

#ifdef key_phdf5
	// Close the access property
	status = H5Pclose(plist_access_id);
#endif

	// Close the file
	status = H5Fclose(file_id);

#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
#endif
	H5garbage_collect();
#ifdef key_mpi
	MPI_Barrier(MPI_COMM_WORLD);
#endif
}


void hdf5_read_0d_double(const char* namefile, const char* namegroup, const char* namefield, double *data){
	char msg[MAXSTRINGLENGTH];
	hid_t file_id, dataset_id, group_id;
	herr_t status;
	int exists;
	struct stat buffer;

	int rank;
#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#else
	rank = 0;
#endif

	// Test if the file exists
	exists = (stat(namefile, &buffer) == 0);

	if (!exists){
		sprintf(msg, "Input file %s does not exist\n", namefile);
		error(msg);
	}

	if (rank == 0){
		file_id = H5Fopen(namefile, H5F_ACC_RDWR, H5P_DEFAULT);

		// Test if the group exists
		exists = ((int)(H5Lexists(file_id, namegroup, H5P_DEFAULT)) > 0);

		if (exists)
			group_id = H5Gopen2(file_id, namegroup, H5P_DEFAULT);
		else{
			// Close the file
			status = H5Fclose(file_id);
			// Stop the code
			sprintf(msg, "Problem with 2nd argument (namegroup) in hdf5_read_0d_double: the HDF5 group %s cannot be found\n", namegroup);
			error(msg);
		}

		// Test if the dataset exists
		exists = ((int)(H5Lexists(group_id, namefield, H5P_DEFAULT)) > 0);

		if (exists)
			dataset_id = H5Dopen2(group_id, namefield, H5P_DEFAULT);
		else{
			// Close the group
			status = H5Gclose(group_id);
			// Close the file
			status = H5Fclose(file_id);
			// Stop the code
			sprintf(msg, "Problem with 3rd argument (namefield) in hdf5_read_0d_double: the HDF5 dataset %s cannot be found\n", namefield);
			error(msg);
		}

		// Read the data
		status = H5Dread(dataset_id, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, data);

		status = H5Dclose(dataset_id);
		status = H5Gclose(group_id);
		status = H5Fclose(file_id);
	}
#if key_mpi
	MPI_Bcast(data, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
#endif
}


void hdf5_read_0d_int(const char* namefile, const char* namegroup, const char* namefield, int *data){
	char msg[MAXSTRINGLENGTH];
	hid_t file_id, dataset_id, group_id;
	herr_t status;
	int exists;
	struct stat buffer;

	int rank;
#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#else
	rank = 0;
#endif

	// Test if the file exists
	exists = (stat(namefile, &buffer) == 0);

	if (!exists){
		sprintf(msg, "Input file %s does not exist\n", namefile);
		error(msg);
	}

	if (rank == 0){
		file_id = H5Fopen(namefile, H5F_ACC_RDWR, H5P_DEFAULT);

		// Test if the group exists
		exists = ((int)(H5Lexists(file_id, namegroup, H5P_DEFAULT)) > 0);


		if (exists)
			group_id = H5Gopen2(file_id, namegroup, H5P_DEFAULT);
		else{
			// Close the file
			status = H5Fclose(file_id);
			// Stop the code
			sprintf(msg, "Problem with 2nd argument (namegroup) in hdf5_read_0d_int: the HDF5 group %s cannot be found\n", namegroup);
			error(msg);
		}

		// Test if the dataset exists
		exists = ((int)(H5Lexists(group_id, namefield, H5P_DEFAULT)) > 0);
		if (exists)
			dataset_id = H5Dopen2(group_id, namefield, H5P_DEFAULT);
		else{
			// Close the group
			status = H5Gclose(group_id);
			// Close the file
			status = H5Fclose(file_id);
			// Stop the code
			sprintf(msg, "Problem with 3rd argument (namefield) in hdf5_read_0d_int: the HDF5 dataset %s cannot be found\n", namefield);
			error(msg);
		}

		// Read the data
		status = H5Dread(dataset_id, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, data);

		H5Dclose(dataset_id);
		H5Gclose(group_id);
		H5Fclose(file_id);
	}
#if key_mpi
	MPI_Bcast(data, 1, MPI_INT, 0, MPI_COMM_WORLD);
#endif
}


void hdf5_read_2d(const char* namefile, const char* namegroup, const char* namefield, double ***data, PARAMETERS *parameters){
	hsize_t start[2];
	hsize_t count[2];
	hsize_t stride[2];

	hid_t plist_access_id;
	hid_t file_id;
	hid_t group_id;
	hid_t dataset_id;

	hid_t dataspace_id;
	
	hid_t hyperslab_id; 
	hid_t memspace_id;
	
	herr_t status;
	hid_t att_id, attspace_id;
	int exists;
	char msg[MAXSTRINGLENGTH];
	struct stat buffer;

	int NVXLOC, NVYLOC, IVX0LOC, IVY0LOC;

	int rank, commsize;
#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &commsize);
#else
	rank = 0;
	commsize = 1;
#endif

#ifdef key_phdf5
	// Set access property list for IO access
	plist_access_id = H5Pcreate(H5P_FILE_ACCESS);
	status = H5Pset_fapl_mpio(plist_access_id, MPI_COMM_WORLD, MPI_INFO_NULL);
#else
	plist_access_id = H5P_DEFAULT;
#endif

	// Test if the file exists
	exists = (stat(namefile, &buffer) == 0);

	if (!exists){
		sprintf(msg, "Problem with 1st argument (namefile) in hdf5_read_2d_double: the HDF5 file %s cannot be found\n", namefile);
		error(msg);
	}

	// Open the file
	file_id = H5Fopen(namefile, H5F_ACC_RDWR, plist_access_id);

	// Test if the group exists
	exists = ((int)(H5Lexists(file_id, namegroup, H5P_DEFAULT)) > 0);

	if (!exists){
		// Close the file
		status = H5Fclose(file_id);

		// Stop the code
		sprintf(msg, "Problem with 2nd argument (namegroup) in hdf5_read_2d_double: the HDF5 group %s cannot be found\n", namegroup);
		error(msg);
	}

	// Open the group
	group_id = H5Gopen2(file_id, namegroup, H5P_DEFAULT);

	// Test if the dataset exists
	exists = ((int)(H5Lexists(group_id, namefield, H5P_DEFAULT)) > 0);

	if (!exists){
		// Close the group
		status = H5Gclose(group_id);

		// Close the file
		status = H5Fclose(file_id);

		// Stop the code
		sprintf(msg, "Problem with 3rd argument (namefield) in hdf5_read_2d_double: the HDF5 dataset %s cannot be found\n", namefield);
		error(msg);
	}

	// Open the dataset
	dataset_id = H5Dopen1(group_id, namefield);

	// Test if the attribute nvx exists
	exists = ((int)(H5Aexists(dataset_id, "nvx")) > 0);

	if (!exists){
		// Close the dataset
		status = H5Dclose(dataset_id);

		// Close the group
		status = H5Gclose(group_id);

		// Close the file
		status = H5Fclose(file_id);

		// Stop the code
		sprintf(msg, "Problem with the dataset %s: its 'nvx' attribute cannot be identified\n", namefield);
		error(msg);
	}
	att_id = H5Aopen_name(dataset_id, "nvx");
	status = H5Aread(att_id, H5T_NATIVE_INT, &parameters->nvx);
	H5Aclose(att_id);

	exists = ((int)(H5Aexists(dataset_id, "nvy")) > 0);

	if (!exists){
		// Close the dataset
		status = H5Dclose(dataset_id);

		// Close the group
		status = H5Gclose(group_id);

		// Close the file
		status = H5Fclose(file_id);

		// Stop the code
		sprintf(msg, "Problem with the dataset %s: its 'nvy' attribute cannot be identified\n", namefield);
		error(msg);
	}
	att_id = H5Aopen_name(dataset_id, "nvy");
	status = H5Aread(att_id, H5T_NATIVE_INT, &parameters->nvy);
	H5Aclose(att_id);

	parameters->nvz = 0;

	// Build the partition table associated to the current communicator MPI_COMM_WORLD
	// Warning: it can be different from the partition table used for writting the file
#ifdef key_fftw_mpi
	// Build the partition table
	// The global array of size nx*ny will be splitted in parts of size nxloc*ny
	int itemp, itemp2, istatus;
	parameters->alloc_local = fftw_mpi_local_size_2d(parameters->nvx, parameters->nvy, MPI_COMM_WORLD, &parameters->nvxloc, &parameters->ivx0loc);
	parameters->nvyloc = parameters->nvy;
	parameters->ivy0loc = 0;
#else
	parameters->nvxloc = parameters->nvx;
	parameters->ivx0loc = 0;
	parameters->nvyloc = parameters->nvy;
	parameters->ivy0loc = 0;
#endif
	parameters->nvzloc = 1;
	parameters->ivz0loc = 0;

	NVXLOC = (int)(parameters->nvxloc);
	NVYLOC = (int)(parameters->nvyloc);
	IVX0LOC = (int)(parameters->ivx0loc);
	IVY0LOC = (int)(parameters->ivy0loc);

	start[0] = IVX0LOC;
	start[1] = IVY0LOC;
	count[0] = NVXLOC;
	count[1] = NVYLOC;
	stride[0] = 1;
	stride[1] = 1;

	// Do allocation for local data
	*data = allocation_2d(NVXLOC, NVYLOC);

#ifdef key_phdf5
	// Create the dataspace corresponding to the hyperslab
	hyperslab_id = H5Dget_space(dataset_id);
    status = H5Sselect_hyperslab(hyperslab_id, H5S_SELECT_SET, start, stride, count, NULL);

	// Create the memory space for the local slab
	memspace_id = H5Screate_simple(2, count, NULL);
#else
	// Default data space and memory space
	hyperslab_id = H5S_ALL;
	memspace_id = H5S_ALL;
#endif

	// Read local data
	status = H5Dread(dataset_id, H5T_NATIVE_DOUBLE, memspace_id, hyperslab_id, H5P_DEFAULT, (void *)&(*data)[0][0]);

#ifdef key_phdf5
	status = H5Sclose(memspace_id);
	status = H5Sclose(hyperslab_id);
#endif
	status = H5Dclose(dataset_id);

	status = H5Gclose(group_id);

	status = H5Fclose(file_id);

#ifdef key_phdf5
	// Clean the access property
	status = H5Pclose(plist_access_id);
#endif
}


void hdf5_read_3d(const char* namefile, const char* namegroup, const char* namefield, double**** data, PARAMETERS *parameters){
	hsize_t start[3];
	hsize_t count[3];
	hsize_t stride[3];

	hid_t plist_access_id;
	hid_t file_id;
	hid_t group_id;
	hid_t dataset_id;

	hid_t dataspace_id;
	
	hid_t hyperslab_id; 
	hid_t memspace_id;
	
	herr_t status;
	hid_t att_id, attspace_id;
	int exists;
	char msg[MAXSTRINGLENGTH];
	struct stat buffer;

	int NVXLOC, NVYLOC, NVZLOC, IVX0LOC, IVY0LOC, IVZ0LOC;

	int rank, commsize;
#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &commsize);
#else
	rank = 0;
	commsize = 1;
#endif

#ifdef key_phdf5
	// Set access property list for IO access
	plist_access_id = H5Pcreate(H5P_FILE_ACCESS);
	status = H5Pset_fapl_mpio(plist_access_id, MPI_COMM_WORLD, MPI_INFO_NULL);
#else
	plist_access_id = H5P_DEFAULT;
#endif

	// Test if the file exists
	exists = (stat(namefile, &buffer) == 0);

	if (!exists){
		sprintf(msg, "Problem with 1st argument (namefile) in hdf5_read_3d_double: the HDF5 file %s cannot be found\n", namefile);
		error(msg);
	}

	// Open the file
	file_id = H5Fopen(namefile, H5F_ACC_RDWR, plist_access_id);

	// Test if the group exists
	exists = ((int)(H5Lexists(file_id, namegroup, H5P_DEFAULT)) > 0);

	if (!exists){
		// Close the file
		status = H5Fclose(file_id);

		// Stop the code
		sprintf(msg, "Problem with 2nd argument (namegroup) in hdf5_read_3d_double: the HDF5 group %s cannot be found\n", namegroup);
		error(msg);
	}

	// Open the group
	group_id = H5Gopen2(file_id, namegroup, H5P_DEFAULT);

	// Test if the dataset exists
	exists = ((int)(H5Lexists(group_id, namefield, H5P_DEFAULT)) > 0);

	if (!exists){
		// Close the group
		status = H5Gclose(group_id);

		// Close the file
		status = H5Fclose(file_id);

		// Stop the code
		sprintf(msg, "Problem with 3rd argument (namefield) in hdf5_read_3d_double: the HDF5 dataset %s cannot be found\n", namefield);
		error(msg);
	}

	// Open the dataset
	dataset_id = H5Dopen1(group_id, namefield);

	// Test if the attribute nvx exists
	exists = ((int)(H5Aexists(dataset_id, "nvx")) > 0);

	if (!exists){
		// Close the dataset
		status = H5Dclose(dataset_id);

		// Close the group
		status = H5Gclose(group_id);

		// Close the file
		status = H5Fclose(file_id);

		// Stop the code
		sprintf(msg, "Problem with the dataset %s: its 'nvx' attribute cannot be identified\n", namefield);
		error(msg);
	}
	att_id = H5Aopen_name(dataset_id, "nvx");
	status = H5Aread(att_id, H5T_NATIVE_INT, &(parameters->nvx));
	H5Aclose(att_id);

	exists = ((int)(H5Aexists(dataset_id, "nvy")) > 0);

	if (!exists){
		// Close the dataset
		status = H5Dclose(dataset_id);

		// Close the group
		status = H5Gclose(group_id);

		// Close the file
		status = H5Fclose(file_id);

		// Stop the code
		sprintf(msg, "Problem with the dataset %s: its 'nvy' attribute cannot be identified\n", namefield);
		error(msg);
	}
	att_id = H5Aopen_name(dataset_id, "nvy");
	status = H5Aread(att_id, H5T_NATIVE_INT, &(parameters->nvy));
	H5Aclose(att_id);

	exists = ((int)(H5Aexists(dataset_id, "nvz")) > 0);

	if (!exists){
		// Close the dataset
		status = H5Dclose(dataset_id);

		// Close the group
		status = H5Gclose(group_id);

		// Close the file
		status = H5Fclose(file_id);

		// Stop the code
		sprintf(msg, "Problem with the dataset %s: its 'nvz' attribute cannot be identified\n", namefield);
		error(msg);
	}
	att_id = H5Aopen_name(dataset_id, "nvz");
	status = H5Aread(att_id, H5T_NATIVE_INT, &(parameters->nvz));
	H5Aclose(att_id);

	// Build the partition table associated to the current communicator MPI_COMM_WORLD
	// Warning: it can be different from the partition table used for writting the file
#ifdef key_fftw_mpi
	// Build the partition table
	// The global array of size nx*ny will be splitted in parts of size nxloc*ny
	int itemp, itemp2, istatus;
	parameters->alloc_local = fftw_mpi_local_size_3d(parameters->nvx, parameters->nvy, parameters->nvz, MPI_COMM_WORLD, &parameters->nvxloc, &parameters->ivx0loc);
	parameters->nvyloc = parameters->nvy;
	parameters->ivy0loc = 0;
	parameters->nvzloc = parameters->nvz;
	parameters->ivz0loc = 0;
#else
	parameters->nvxloc = parameters->nvx;
	parameters->ivx0loc = 0;
	parameters->nvyloc = parameters->nvy;
	parameters->ivy0loc = 0;
	parameters->nvzloc = parameters->nvz;
	parameters->ivz0loc = 0;
#endif

	NVXLOC = (int)(parameters->nvxloc);
	NVYLOC = (int)(parameters->nvyloc);
	IVX0LOC = (int)(parameters->ivx0loc);
	IVY0LOC = (int)(parameters->ivy0loc);
	NVZLOC = (int)(parameters->nvzloc);
	IVZ0LOC = (int)(parameters->ivz0loc);

	start[0] = IVX0LOC;
	start[1] = IVY0LOC;
	start[2] = IVZ0LOC;
	count[0] = NVXLOC;
	count[1] = NVYLOC;
	count[2] = NVZLOC;
	stride[0] = 1;
	stride[1] = 1;
	stride[2] = 1;

	// Do allocation for local data
	*data = allocation_3d(NVXLOC, NVYLOC, NVZLOC);

#ifdef key_phdf5
	// Create the data space corresponding to the hyperslab
	hyperslab_id = H5Dget_space(dataset_id);
    status = H5Sselect_hyperslab(hyperslab_id, H5S_SELECT_SET, start, stride, count, NULL);

	// Create the memory space for the local slab
	memspace_id = H5Screate_simple(3, count, NULL);
#else
	hyperslab_id = H5S_ALL;
	memspace_id = H5S_ALL;
#endif

	// Read local data
	status = H5Dread(dataset_id, H5T_NATIVE_DOUBLE, memspace_id, hyperslab_id, H5P_DEFAULT, (void *)&(*data)[0][0][0]);

#ifdef key_phdf5
	status = H5Sclose(memspace_id);
	status = H5Sclose(hyperslab_id);
#endif
	status = H5Dclose(dataset_id);

	status = H5Gclose(group_id);

	status = H5Fclose(file_id);

#ifdef key_phdf5
	// Clean the access property
	status = H5Pclose(plist_access_id);
#endif
}





int hdf5_search_in_file(const char *namefile, const char *dataname){
	char msg[MAXSTRINGLENGTH];
	hid_t file_id, data_id;
	herr_t status;
	int exists;

	struct stat buffer;

	int rank;
#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#else
	rank = 0;
#endif

	exists = (1==0);
	
	if (rank == 0){
		// Test if file exists
		exists = (stat(namefile, &buffer) == 0);

		if (exists){
			// Open the file
			file_id = H5Fopen(namefile, H5F_ACC_RDWR, H5P_DEFAULT);

			// Test if the data exists in the file
			exists = ((int)H5Lexists(file_id, dataname, H5P_DEFAULT) > 0);

			// Close the file
			H5Fclose(file_id);
		}
	}
#ifdef key_mpi
	MPI_Bcast(&exists, 1, MPI_INT, 0, MPI_COMM_WORLD);
#endif

	return exists;
}
