#include "./src/IO/xdmf_interface.h"


mxml_node_t* xdmf_create_new_root(void){
	mxml_node_t *root, *format, *xdmf;
	root = mxmlNewXML("1.0");
	format = mxmlNewElement(root, "!DOCTYPE Xdmf SYSTEM \"Xdmf.dtd\" []");
	xdmf = mxmlNewElement(root, "Xdmf");
	mxmlElementSetAttr(xdmf, "Version", "2.0");
	return root;
}


mxml_node_t* xdmf_add_domain(mxml_node_t *root){
	mxml_node_t *xdmf, *domain;
	xdmf = mxmlFindElement(root, root, "Xdmf", NULL, NULL, MXML_DESCEND);
	domain = mxmlNewElement(xdmf, "Domain");
	return domain;
}


mxml_node_t* xdmf_add_mesh(mxml_node_t *domain, const char *meshh5file, const char *meshh5group){
	mxml_node_t *grid;
	mxml_node_t *topology, *topology_dataitem;
	mxml_node_t *geometry, *geometry_dataitem;
	int exists = 0;
	int ndims, s, nnodes, ntriangles, nquadrangles;
	herr_t status;
	hid_t file_id, group_id, dataset_id, dataspace_id;
	hsize_t *dims, *maxdims;
	char buf[MAXSTRINGLENGTH];
	char *path, *meshfile;
	FILE *f;
	
	// Verify if the mesh file exists
	if ((f = fopen(meshh5file, "r"))){
		fclose(f);
		exists = 1;
	}
	if (exists != 1){
		sprintf(buf, "Error: File %s cannot be found", meshh5file);
		error(buf);
	}
	file_id = H5Fopen(meshh5file, H5F_ACC_RDONLY, H5P_DEFAULT);
	
	// File is opened : try to open the mesh group
	group_id = H5Lexists(file_id, meshh5group, H5P_DEFAULT);
	if (group_id > 0)
		group_id = H5Gopen2(file_id, meshh5group, H5P_DEFAULT);
	else{
		sprintf(buf, "Error: H5Group %s cannot be found in file %s", meshh5group, meshh5file);
		error(buf);
	}
	// Group is opened : try to open some datasets
	
	// Open dataset "nodes"
	dataset_id = H5Lexists(group_id, "nodes", H5P_DEFAULT);
	if (dataset_id > 0)
		dataset_id = H5Dopen2(group_id, "nodes", H5P_DEFAULT);
	else{
		sprintf("Error: H5Dataset \"nodes\" in %s H5Group %s cannot be found (file %s)", meshh5group, meshh5file);
		error(buf);
	}
	// Dataset "nodes" is opened : extract dimensionality of the mesh and number of nodes
	dataspace_id = H5Dget_space(dataset_id);
	ndims = H5Sget_simple_extent_ndims(dataspace_id);
	dims = malloc(sizeof(hsize_t)*ndims);
	maxdims = malloc(sizeof(hsize_t)*ndims);
	s = H5Sget_simple_extent_dims(dataspace_id, dims, maxdims);
	nnodes = (int)dims[0];
	free(dims);
	free(maxdims);
	H5Sclose(dataspace_id);
	H5Dclose(dataset_id);
	
	// Open dataset "triangles"
	dataset_id = H5Lexists(group_id, "triangles", H5P_DEFAULT);
	if (dataset_id > 0){
		dataset_id = H5Dopen2(group_id, "triangles", H5P_DEFAULT);
		// Dataset "triangles" is opened : extract number of triangles
		dataspace_id = H5Dget_space(dataset_id);
		ndims = H5Sget_simple_extent_ndims(dataspace_id);
		dims = malloc(sizeof(hsize_t)*ndims);
		maxdims = malloc(sizeof(hsize_t)*ndims);
		s = H5Sget_simple_extent_dims(dataspace_id, dims, maxdims);
		ntriangles = (int)maxdims[0];
		free(dims);
		free(maxdims);
		H5Sclose(dataspace_id);
		H5Dclose(dataset_id);
	}
	else
		ntriangles = 0;
	
	// Open dataset "quadrangles"
	dataset_id = H5Lexists(group_id, "quadrangles", H5P_DEFAULT);
	if (dataset_id > 0){
		dataset_id = H5Dopen2(group_id, "quadrangles", H5P_DEFAULT);
		// Dataset "quadrangles" is opened : extract number of quadrangles
		dataspace_id = H5Dget_space(dataset_id);
		ndims = H5Sget_simple_extent_ndims(dataspace_id);
		dims = malloc(sizeof(hsize_t)*ndims);
		maxdims = malloc(sizeof(hsize_t)*ndims);
		s = H5Sget_simple_extent_dims(dataspace_id, dims, maxdims);
		nquadrangles = (int)maxdims[0];
		free(dims);
		free(maxdims);
		H5Sclose(dataspace_id);
		H5Dclose(dataset_id);
	}
	else
		nquadrangles = 0;
	
	H5Gclose(group_id);
	H5Fclose(file_id);
	
	split_path_file(&path, &meshfile, (char*)meshh5file);
	
	if (!((ntriangles > 0) && (nquadrangles == 0)))
		error("Only 2D meshes with triangles only can be handled at present time");
	
	grid = mxmlNewElement(domain, "Grid");
	mxmlElementSetAttr(grid, "Name", meshh5group);
	mxmlElementSetAttr(grid, "GridType", "Uniform");
	
	topology = mxmlNewElement(grid, "Topology");
	mxmlElementSetAttr(topology, "TopologyType", "Triangle");
	sprintf(buf, "%d", ntriangles);
	mxmlElementSetAttr(topology, "NumberOfElements", buf);
	
	topology_dataitem = mxmlNewElement(topology, "DataItem");
	sprintf(buf, "%d", ntriangles*3);
	mxmlElementSetAttr(topology_dataitem, "Dimensions", buf);
	mxmlElementSetAttr(topology_dataitem, "Format", "HDF");
	
	sprintf(buf, "./%s:/%s/triangles", meshfile, meshh5group);
	mxmlNewText(topology_dataitem, 0, buf);
	
	geometry = mxmlNewElement(grid, "Geometry");
	mxmlElementSetAttr(geometry, "GeometryType", "XY");
	
	geometry_dataitem = mxmlNewElement(geometry, "DataItem");
	sprintf(buf, "%d", nnodes*2);
	mxmlElementSetAttr(geometry_dataitem, "Dimensions", buf);
	mxmlElementSetAttr(geometry_dataitem, "Format", "HDF");
	
	sprintf(buf, "./%s:/%s/nodes", meshfile, meshh5group);
	mxmlNewText(geometry_dataitem, 0, buf);
	
	free_1d_char(path);
	free_1d_char(meshfile);
	
	return grid;
}


mxml_node_t* xdmf_add_scalar_data(mxml_node_t *grid, const char* datah5file, const char* datah5group, const char* datah5dataset, int centering){
	mxml_node_t *attribute, *attribute_dataitem;
	char buf[MAXSTRINGLENGTH];
	char *path, *datafile;

	int exists = 0;
	int ndims, s;
	herr_t status;
	hid_t file_id, group_id, dataset_id, dataspace_id;
	hsize_t *dims, *maxdims;
	FILE *f;
	
	// Verify if the data file exists
	if ((f = fopen(datah5file, "r"))){
		fclose(f);
		exists = 1;
	}
	if (exists != 1){
		sprintf(buf, "Error: File %s cannot be found", datah5file);
		error(buf);
	}
	file_id = H5Fopen(datah5file, H5F_ACC_RDONLY, H5P_DEFAULT);

	// File is opened : try to open the data group
	group_id = H5Lexists(file_id, datah5group, H5P_DEFAULT);
	if (group_id > 0)
		group_id = H5Gopen2(file_id, datah5group, H5P_DEFAULT);
	else{
		sprintf(buf, "Error: H5Group %s cannot be found in file %s", datah5group, datah5file);
		error(buf);
	}
	// Group is opened : try to open the dataset
	dataset_id = H5Lexists(group_id, datah5dataset, H5P_DEFAULT);
	if (dataset_id > 0)
		dataset_id = H5Dopen2(group_id, datah5dataset, H5P_DEFAULT);
	else{
		sprintf("Error: H5Dataset \"%s\" in %s H5Group %s cannot be found (file %s)", datah5dataset, datah5group, datah5file);
		error(buf);
	}

	// Dataset is opened : extract size of data
	dataspace_id = H5Dget_space(dataset_id);
	ndims = H5Sget_simple_extent_ndims(dataspace_id);
	dims = malloc(sizeof(hsize_t)*ndims);
	maxdims = malloc(sizeof(hsize_t)*ndims);
	s = H5Sget_simple_extent_dims(dataspace_id, dims, maxdims);
	s = (int)dims[0];
	
	free(dims);
	free(maxdims);
	H5Sclose(dataspace_id);
	H5Dclose(dataset_id);
	H5Gclose(group_id);
	H5Fclose(file_id);
	
	split_path_file(&path, &datafile, (char*)datah5file);

	attribute = mxmlNewElement(grid, "Attribute");
	mxmlElementSetAttr(attribute, "Name", datah5dataset);
	mxmlElementSetAttr(attribute, "AttributeType", "Scalar");
	switch (centering){
		case NODE_CENTERED :
			mxmlElementSetAttr(attribute, "Center", "Node");
			break;
		case CELL_CENTERED :
			mxmlElementSetAttr(attribute, "Center", "Cell");
			break;
		default :
			error("Unknown kind of data centering");
	}
	attribute_dataitem = mxmlNewElement(attribute, "DataItem");
	sprintf(buf, "%d", s);
	mxmlElementSetAttr(attribute_dataitem, "Dimensions", buf);
	mxmlElementSetAttr(attribute_dataitem, "Format", "HDF");
	sprintf(buf, "./%s:/%s/%s", datafile, datah5group, datah5dataset);
	mxmlNewText(attribute_dataitem, 0, buf);
	
	free_1d_char(path);
	free_1d_char(meshfile);
	
	return attribute;
}



mxml_node_t* xdmf_add_vector_data(mxml_node_t *grid, const char **scalar_attribute_names, int n_scalar_attribute_names, const char *vector_attribute_name){
	mxml_node_t *vector_attribute, *vector_attribute_dataitem, *function_dataitem;
	mxml_node_t *scalar_attribute, *scalar_attribute_dataitem;
	char buf[MAXSTRINGLENGTH];
	char centering1[MAXSTRINGLENGTH];
	char centering2[MAXSTRINGLENGTH];
	char ndata1[MAXSTRINGLENGTH];
	char ndata2[MAXSTRINGLENGTH];
	char source[MAXSTRINGLENGTH];
	char thefunction[MAXSTRINGLENGTH];
	
	strcpy(thefunction, "join(");
	for (int i = 0; i < n_scalar_attribute_names; i++){
		sprintf(thefunction, "%s$%d", thefunction, i);
		scalar_attribute = mxmlFindElement(grid, grid, scalar_attribute_names[i], NULL, NULL, MXML_DESCEND);
		scalar_attribute_dataitem = mxmlFindElement(scalar_attribute, scalar_attribute, "DataItem", NULL, NULL, MXML_DESCEND);
		strcpy(buf, mxmlElementGetAttr(scalar_attribute, "AttributeType"));
		if (strcmp(buf, "Scalar") != 0)
			error("Error: trying to join some data that are not scalar");
		
		strcpy(buf, mxmlElementGetAttr(scalar_attribute_dataitem, "Format"));
		if (strcmp(buf, "HDF") != 0)
			error("Error: trying to join some data that do not come from *.h5 files");
		
		if (i == 0){
			strcpy(centering1, mxmlElementGetAttr(scalar_attribute, "Center"));
			strcpy(ndata1, mxmlElementGetAttr(scalar_attribute_dataitem, "Dimensions"));
		}
		else{
			strcpy(centering2, mxmlElementGetAttr(scalar_attribute, "Center"));
			strcpy(ndata2, mxmlElementGetAttr(scalar_attribute_dataitem, "Dimensions"));
			if (strcmp(centering1, centering2) != 0)
				error("Error: trying to join some data that do not have the same centering convention");
			if (strcmp(ndata1, ndata2) != 0)
				error("Error: trying to join some data that do not have the same size");
		}
	}
		
	sprintf(thefunction, "%s)", thefunction);
	
	vector_attribute = mxmlNewElement(grid, "Attribute");
	mxmlElementSetAttr(vector_attribute, "Name", vector_attribute_name);
	mxmlElementSetAttr(vector_attribute, "Center", centering1);
	mxmlElementSetAttr(vector_attribute, "AttributeType", "Vector");

	function_dataitem = mxmlNewElement(vector_attribute, "DataItem");
	mxmlElementSetAttr(function_dataitem, "ItemType", "Function");	
	mxmlElementSetAttr(function_dataitem, "Function", thefunction);
	sprintf(buf, "%s %d", ndata1, n_scalar_attribute_names);
	mxmlElementSetAttr(function_dataitem, "Dimensions", buf);
	
	for (int i = 0; i < n_scalar_attribute_names; i++){
		scalar_attribute = mxmlFindElement(grid, grid, scalar_attribute_names[i], NULL, NULL, MXML_DESCEND);
		scalar_attribute_dataitem = mxmlFindElement(scalar_attribute, scalar_attribute, "DataItem", NULL, NULL, MXML_DESCEND);
		
		vector_attribute_dataitem = mxmlNewElement(function_dataitem, "DataItem");
		mxmlElementSetAttr(scalar_attribute_dataitem, "Name", scalar_attribute_names[i]);
		sprintf(buf, "%s 1", ndata1);
		mxmlElementSetAttr(scalar_attribute_dataitem, "Dimensions", buf);
		mxmlElementSetAttr(scalar_attribute_dataitem, "Format", "HDF");
		strcpy(source, mxmlGetText(scalar_attribute_dataitem, NULL));
		mxmlNewText(scalar_attribute_dataitem, 0, source);
	}
}


void xdmf_write(mxml_node_t *root, const char *namefile){
	FILE *f;
	f = fopen(namefile, "w");
	mxmlSaveFile(root, f, MXML_NO_CALLBACK);
	fclose(f);
}
