#include "./src/core/cross_section.h"


double sinc(double x){
	if (fabs(x) < THRESHOLD)
		return 1.;
	else
		return sin(x)/x;
}


double phi2(double x, PARAMETERS parameters){
	if (fabs(parameters.gamma) < THRESHOLD){
		// 2D Maxwellian molecules model
		return 2.*parameters.cutoff * sinc(parameters.cutoff*x);
	}
	else
		error("Only 2D Maxwellian molecules cross-section is handled at present time.");
}


double phi3(double x, PARAMETERS parameters){
	if (fabs(parameters.gamma-1.) < THRESHOLD){
		// 3D Hard Spheres model
		return pow(parameters.cutoff,2) * (2*sinc(parameters.cutoff*x) - pow(sinc(0.5*parameters.cutoff*x),2));
	}
	else
		error("Only 3D Hard Spheres cross-section is handled at present time.");
}


double psi3(double x, PARAMETERS parameters){
	if (fabs(parameters.gamma-1.) < THRESHOLD){
		// 3D Hard Spheres model
		return 2. * pow(parameters.cutoff,2) * pow(sinc(0.5*parameters.cutoff*x),2);
	}
	else
		error("Only 3D Hard Spheres cross-section is handled at present time.");
}


void expanded_collision_modes(double **alpha, double **alphaprime, PARAMETERS parameters){
	double pi = 4.*atan(1.);
	double theta, phi, thetaprime, s, psi, ex, ey, ez;
	int kvxabs, kvyabs, kvzabs;
	double px, py, pz, pnorm;
	double modex, modey, modez;
	int ind_k, ind_p, ind_pq, k;
	int itemp;

	int NVXLOC = (int)parameters.nvxloc;
	int IVX0LOC = (int)parameters.ivx0loc;
	int NVYLOC = (int)parameters.nvyloc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int NVZLOC = (int)parameters.nvzloc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	
	if (parameters.ndim == DIM2){
		for (int p = 0; p < parameters.mint; p++){
			ind_p = p;

			// Since alpha and alphaprime are computed by using the same function,
			// the integral on [0,pi] for obtaining beta is replaced by 2 * integral on [0,pi/2]
			theta = 0.5*p*pi/parameters.mint;
			
			#pragma omp parallel for private(k, kvxabs, kvyabs, modex, modey, ex, ey, s) shared(NVXLOC, NVYLOC, parameters, MVX, MVY, IVX0LOC, IVY0LOC, theta, alpha, alphaprime, ind_p)
			for (k = 0; k < NVXLOC*NVYLOC; k++){
#ifdef key_force_vectorize
				kvxabs = parameters.kvxabstab[k] - MVX*(parameters.kvxabstab[k] >= MVX/2);
				kvyabs = parameters.kvyabstab[k] - MVY*(parameters.kvyabstab[k] >= MVY/2);
#else
				kvxabs = k/NVYLOC;
				kvyabs = k - kvxabs*NVYLOC;
				kvxabs += IVX0LOC;
				kvyabs += IVY0LOC;
				if (kvxabs >= MVX/2)
					kvxabs -= MVX;
				if (kvyabs >= MVY/2)
					kvyabs -= MVY;
#endif
				modex = kvxabs;
				modey = kvyabs;
				
				ex = cos(theta);
				ey = sin(theta);
				
				s = modex*ex+modey*ey;
				alpha[ind_p][k] = 2*parameters.Ccross*phi2(s, parameters);

				ex = -sin(theta);
				ey = cos(theta);
				s = modex*ex+modey*ey;
				alphaprime[ind_p][k] = phi2(s, parameters);
			}
		}
	}
	else if (parameters.ndim == DIM3){
		for (int p = 0; p < parameters.mint1; p++){
			theta = p*pi/parameters.mint1;
			for (int q = 0; q < parameters.mint2; q++){
				ind_pq = q+parameters.mint2*p;

				phi = q*pi/parameters.mint2;
				ex = sin(theta)*cos(phi);
				ey = sin(theta)*sin(phi);
				ez = cos(theta);
				
				#pragma omp parallel for private(k, kvxabs, kvyabs, kvzabs, itemp, modex, modey, modez, s, px, py, pz, pnorm) shared(NVXLOC, NVYLOC, NVZLOC, parameters, MVX, MVY, MVZ, IVX0LOC, IVY0LOC, IVZ0LOC, ex, ey, ez, ind_pq, alpha, alphaprime, theta)
				for (k = 0; k < NVXLOC*NVYLOC*NVZLOC; k++){
#ifdef key_force_vectorize
					kvxabs = parameters.kvxabstab[k] - MVX*(parameters.kvxabstab[k] >= MVX/2);
					kvyabs = parameters.kvyabstab[k] - MVY*(parameters.kvyabstab[k] >= MVY/2);
					kvzabs = parameters.kvzabstab[k] - MVZ*(parameters.kvzabstab[k] >= MVZ/2);
#else
					itemp = k/NVZLOC;
					kvzabs = k - itemp*NVZLOC;
					kvxabs = itemp/NVYLOC;
					kvyabs = itemp-kvxabs*NVYLOC;
					kvxabs += IVX0LOC;
					kvyabs += IVY0LOC;
					kvzabs += IVZ0LOC;
					if (kvxabs >= MVX/2)
						kvxabs -= MVX;
					if (kvyabs >= MVY/2)
						kvyabs -= MVY;
					if (kvzabs >= MVZ/2)
						kvzabs -= MVZ;
#endif
					modex = kvxabs;
					modey = kvyabs;
					modez = kvzabs;

					s = modex*ex+modey*ey+modez*ez;
					
					alpha[ind_pq][k] = 4*parameters.Ccross*sin(theta)*phi3(s, parameters);

					px = modex - (modex*ex+modey*ey+modez*ez)*ex;
					py = modey - (modex*ex+modey*ey+modez*ez)*ey;
					pz = modez - (modex*ex+modey*ey+modez*ez)*ez;
					pnorm = sqrt(px*px+py*py+pz*pz);

					alphaprime[ind_pq][k] = psi3(pnorm, parameters);
				}
			}
		}
	}
}


double collision_kernel(double r, double theta, PARAMETERS parameters){
	double res;

	res = parameters.Ccross * pow(r, parameters.gamma);

	return res;
}



double ftest_int_2d(double *xitilde, double *sigma){
	double res;

	res = xitilde[0]*sigma[0]+xitilde[1]*sigma[1];

	return res;
}


double ftest_int_3d(double *xitilde, double *sigma){
	double res;

	//res = sqrt(pow(xitilde[0], 2) + pow(xitilde[1], 2) + pow(xitilde[2], 2))*(atan(sigma[0])*1./(3+sigma[1])*sqrt(fabs(sigma[2])));
	//res = xitilde[0]*sigma[0]+xitilde[1]*sigma[1]+xitilde[2]*sigma[2];
	res = 1.;

	return res;
}


void compute_points_weights_force_convolution(double R, int degmax_radial, int degmax_sphere_1, int degmax_sphere_2, int ndim, double ***xitilde, double ***sigma, double **W, int *npoints){
	double pi = 4.*atan(1.);
	int P1, P2, P3, p;
	double *LegP1_weights, *LegP1_points;
	double *LegP2_weights, *LegP2_points, *LegP3_weights, *LegP3_points;
	double *LebP2_weights, **LebP2_points, *LebP3_weights, **LebP3_points;

	P1 = degmax_radial;
	LegP1_weights = allocation_1d(P1);
	LegP1_points = allocation_1d(P1);
	read_legendre_roots(P1, LegP1_points, LegP1_weights);

	if (ndim == 2){
		P2 = degmax_sphere_1;
		LegP2_weights = allocation_1d(P2);
		LegP2_points = allocation_1d(P2);
		read_legendre_roots(P2, LegP2_points, LegP2_weights);

		P3 = degmax_sphere_2;
		LegP3_weights = allocation_1d(P3);
		LegP3_points = allocation_1d(P3);
		read_legendre_roots(P3, LegP3_points, LegP3_weights);

		*npoints = P1*P2*P3;
		*xitilde = allocation_2d(*npoints, ndim);
		*sigma = allocation_2d(*npoints, ndim);
		*W = allocation_1d(*npoints);

		for (int p1 = 0; p1 < P1; p1++){
			for (int p2 = 0; p2 < P2; p2++){
				for (int p3 = 0; p3 < P3; p3++){
					p = p1 + P1 * (p2 + P2*p3);

					(*xitilde)[p][0] = 0.5*R*(LegP1_points[p1]+1.) * cos(pi*(LegP2_points[p2]+1.));
					(*xitilde)[p][1] = 0.5*R*(LegP1_points[p1]+1.) * sin(pi*(LegP2_points[p2]+1.));

					(*sigma)[p][0] = cos(pi*(LegP3_points[p3]+1.));
					(*sigma)[p][1] = sin(pi*(LegP3_points[p3]+1.));

					(*W)[p] = LegP1_weights[p1] * LegP2_weights[p2] * LegP3_weights[p3] * 0.25*pow(R*pi,2) * (LegP1_points[p1]+1.);
				}
			}
		}

		free_1d(LegP2_points);
		free_1d(LegP2_weights);
		free_1d(LegP3_points);
		free_1d(LegP3_weights);
	}
	else if (ndim == 3){
		if (degmax_sphere_1 > 65)
			error("Lebedev quadrature cannot handle exact computation for spherical harmonics of order > 65.");
		P2 = order_table(degmax_sphere_1);
		LebP2_points = allocation_2d(3, P2);
		LebP2_weights = allocation_1d(P2);
		ld_by_order(P2, LebP2_points[0], LebP2_points[1], LebP2_points[2], LebP2_weights);

		if (degmax_sphere_2 > 65)
			error("Lebedev quadrature cannot handle exact computation for spherical harmonics of order > 65.");
		P3 = order_table(degmax_sphere_2);
		LebP3_points = allocation_2d(3, P3);
		LebP3_weights = allocation_1d(P3);
		ld_by_order(P3, LebP3_points[0], LebP3_points[1], LebP3_points[2], LebP3_weights);

		*npoints = P1*P2*P3;
		*xitilde = allocation_2d(*npoints, ndim);
		*sigma = allocation_2d(*npoints, ndim);
		*W = allocation_1d(*npoints);

		for (int p1 = 0; p1 < P1; p1++){
			for (int p2 = 0; p2 < P2; p2++){
				for (int p3 = 0; p3 < P3; p3++){
					p = p1 + P1 * (p2 + P2*p3);

					for (int d = 0; d < ndim; d++){
						(*xitilde)[p][d] = 0.5*R*(LegP1_points[p1]+1.) * LebP2_points[d][p2];
						(*sigma)[p][d] = LebP3_points[d][p3];
					}
					(*W)[p] = LegP1_weights[p1] * LebP2_weights[p2] * LebP3_weights[p3] * 2.*pow(R,3)*pow(pi*(LegP1_points[p1]+1.), 2);
				}
			}
		}

		free_2d(LebP2_points);
		free_1d(LebP2_weights);
		free_2d(LebP3_points);
		free_1d(LebP3_weights);
	}


	free_1d(LegP1_weights);
	free_1d(LegP1_points);
	return;
}



void compute_ABCW_force_convolution(double *W, double ***A, double ***B, double ***C, PARAMETERS parameters){
	double **xitilde;
	double **sigma;
	double *weights;
	int P;

	int NVXLOC = (int)parameters.nvxloc;
	int IVX0LOC = (int)parameters.ivx0loc;
	int NVYLOC = (int)parameters.nvyloc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int NVZLOC = (int)parameters.nvzloc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(parameters.nvz, 1);
	double *modek;
	double normxi;
	double mode;

	int kvxabs, kvyabs, kvzabs;
	int itemp;

	modek = allocation_1d(parameters.ndim+1);
	printf("%d integration points\n", parameters.mint);


	compute_points_weights_force_convolution(parameters.cutoff, parameters.P1_int, parameters.P2_int, parameters.P3_int, parameters.ndim+1, &xitilde, &sigma, &weights, &P);

	if (P != parameters.mint)
		error("Something is wrong with the computation of Legendre-Lebedev quadrature weights and points.");
	
	for (int p = 0; p < parameters.mint; p++){
		normxi = 0.;
		for (int d = 0; d <= parameters.ndim; d++)
			normxi += pow(xitilde[p][d], 2);
		normxi = sqrt(normxi);

		for (int k = 0; k < NVXLOC*NVYLOC*NVZLOC; k++){
#ifdef key_force_vectorize
			kvxabs = parameters.kvxabstab[k] - MVX*(parameters.kvxabstab[k] >= MVX/2);
			kvyabs = parameters.kvyabstab[k] - MVY*(parameters.kvyabstab[k] >= MVY/2);
			kvzabs = parameters.kvzabstab[k] - MVZ*(parameters.kvzabstab[k] >= MVZ/2);
#else
			itemp = k/NVZLOC;
			kvzabs = k - itemp*NVZLOC;
			kvxabs = itemp/NVYLOC;
			kvyabs = itemp-kvxabs*NVYLOC;
			kvxabs += IVX0LOC;
			kvyabs += IVY0LOC;
			kvzabs += IVZ0LOC;
			if (kvxabs >= MVX/2)
				kvxabs -= MVX;
			if (kvyabs >= MVY/2)
				kvyabs -= MVY;
			if (kvzabs >= MVZ/2)
				kvzabs -= MVZ;
#endif
			modek[0] = kvxabs;
			modek[1] = kvyabs;
			if (parameters.ndim == DIM3)
				modek[2] = kvzabs;

			mode = 0.;
			for (int d = 0; d <= parameters.ndim; d++)
				mode += modek[d]*(0.5*xitilde[p][d] + 0.5*normxi*sigma[p][d]);
			A[p][k][0] = cos(mode);
			A[p][k][1] = sin(mode);

			mode = 0.;
			for (int d = 0; d <= parameters.ndim; d++)
				mode += modek[d]*(0.5*xitilde[p][d] - 0.5*normxi*sigma[p][d]);
			B[p][k][0] = cos(mode);
			B[p][k][1] = sin(mode);

			mode = 0.;
			for (int d = 0; d <= parameters.ndim; d++)
				mode += modek[d]*xitilde[p][d];
			C[p][k][0] = cos(mode);
			C[p][k][1] = sin(mode);
		}

		
		W[p] = weights[p] * collision_kernel(normxi, 0., parameters);
	}

	free_2d(xitilde);
	free_2d(sigma);
	free_1d(weights);
	free_1d(modek);
}




double integrate_Gamba_2d(double R, int P1, int P2, int P3){
	double pi = 4.*atan(1.);
	char namefile_Leg[MAXSTRINGLENGTH];
	double *LegP1_weights, *LegP1_points, *LegP2_weights, *LegP2_points, *LegP3_weights, *LegP3_points;
	double res = 0.;
	double theta, thetatilde;
	double sigma[2];
	double sigmatilde[2];
	double xitilde[2];

	LegP1_weights = allocation_1d(P1+1);
	LegP1_points = allocation_1d(P1+1);
	LegP2_weights = allocation_1d(P2+1);
	LegP2_points = allocation_1d(P2+1);
	LegP3_weights = allocation_1d(P3+1);
	LegP3_points = allocation_1d(P3+1);

	read_legendre_roots(P1+1, LegP1_points, LegP1_weights);
	read_legendre_roots(P2+1, LegP2_points, LegP2_weights);
	read_legendre_roots(P3+1, LegP3_points, LegP3_weights);

	for (int p1 = 0; p1 <= P1; p1++){
		for (int p2 = 0; p2 <= P2; p2++){
			for (int p3 = 0; p3 <= P3; p3++){
				theta = pi * (LegP3_points[p3]+1.);

				thetatilde = pi * (LegP3_points[p2]+1.);

				sigma[0] = cos(theta);
				sigma[1] = sin(theta);

				sigmatilde[0] = cos(thetatilde);
				sigmatilde[1] = sin(thetatilde);

				xitilde[0] = 0.5*R*(LegP1_points[p1]+1.)*sigmatilde[0];
				xitilde[1] = 0.5*R*(LegP1_points[p1]+1.)*sigmatilde[1];
				
				res += LegP1_weights[p1]*LegP2_weights[p2]*LegP3_weights[p3] *  0.5*R*pi*pi * 0.5*R*(LegP1_points[p1]+1.) * ftest_int_2d(xitilde, sigma);
				//res += LegP1_weights[p1]*LegP2_weights[p2]*LegP3_weights[p3] *  0.5*R*pi*pi * ftest_int_2d(xitilde, sigma);
			}
		}
	}

	free_1d(LegP1_weights);
	free_1d(LegP1_points);
	free_1d(LegP2_weights);
	free_1d(LegP2_points);
	free_1d(LegP3_weights);
	free_1d(LegP3_points);

	return res;
}


double integrate_Gamba_3d(double R, int P1_Leg, int degmax2_spherical_harmonics, int degmax3_spherical_harmonics){
	double pi = 4.*atan(1.);

	double res = 0.;

	char namefile_Leg[MAXSTRINGLENGTH];

	double *LegP1_weights, *LegP1_points;
	double **LebP2_points, *LebP2_weights;
	double **LebP3_points, *LebP3_weights;

	int P2_Leb, P3_Leb;

	double xitilde[3];
	double sigma[3];
	double sigmatilde[3];

	LegP1_weights = allocation_1d(P1_Leg+1);
	LegP1_points = allocation_1d(P1_Leg+1);

	read_legendre_roots(P1_Leg+1, LegP1_points, LegP1_weights);

	if (degmax2_spherical_harmonics > 65)
		error("Lebedev quadrature cannot handle exact computation for spherical harmonics of order > 65.");

	P2_Leb = order_table(degmax2_spherical_harmonics);
	LebP2_points = allocation_2d(3, P2_Leb);
	LebP2_weights = allocation_1d(P2_Leb);

	ld_by_order(P2_Leb, LebP2_points[0], LebP2_points[1], LebP2_points[2], LebP2_weights);

	if (degmax3_spherical_harmonics > 65)
		error("Lebedev quadrature cannot handle exact computation for spherical harmonics of order > 65.");

	P3_Leb = order_table(degmax3_spherical_harmonics);
	LebP3_points = allocation_2d(3, P3_Leb);
	LebP3_weights = allocation_1d(P3_Leb);

	ld_by_order(P3_Leb, LebP3_points[0], LebP3_points[1], LebP3_points[2], LebP3_weights);

	for (int p1 = 0; p1 <= P1_Leg; p1++){
		for (int p2 = 0; p2 < P2_Leb; p2++){
			for (int p3 = 0; p3 < P3_Leb; p3++){
				for (int d = 0; d < 3; d++){
					sigmatilde[d] = LebP2_points[d][p2];
					sigma[d] = LebP3_points[d][p3];
					xitilde[d] = 0.5*R*(LegP1_points[p1]+1.)*sigmatilde[d];
				}

				res += ((LegP1_weights[p1]*LebP2_weights[p2]*LebP3_weights[p3]) * pow(0.5*R*(LegP1_points[p1]+1.), 2) * 0.5*R * ftest_int_3d(xitilde, sigma));
			}
		}
	}

	res *= pow(4.*pi, 2);
	


	free_1d(LegP1_points);
	free_1d(LegP1_weights);
	free_1d(LebP2_weights);
	free_2d(LebP2_points);
	free_1d(LebP3_weights);
	free_2d(LebP3_points);

	return res;
}