#include "cuda_collision.cuh"

#ifdef key_cufft

extern "C" void collision_Qtotal_cufft(double *, double *, double **, double **, double, PARAMETERS);
extern "C" void collision_Qtotal_quick_convolution_cufft(double *, double *, double **, double **, double *, double, PARAMETERS);


__global__ void compute_product_alphareal_Gcomplex(double *d_alphareal, cufftDoubleComplex *d_Gcomplex, cufftDoubleComplex *d_result, int MV){
    // Given an input G and a double array alpha, we compute the product G * alpha

    // Useful in the computation of Q1c, Q2q, Q3q, Q4q

    int ind = blockIdx.x * blockDim.x + threadIdx.x;
    if (ind < MV){
        d_result[ind].x = d_Gcomplex[ind].x * d_alphareal[ind];
        d_result[ind].y = d_Gcomplex[ind].y * d_alphareal[ind];
    }
}


__global__ void compute_product_alphareal_betacomplex(cufftDoubleComplex *d_alphareal, cufftDoubleComplex *d_betacomplex, cufftDoubleComplex *d_result, int MV){
    // Given an complex input beta and a double array alpha (stored in a complex array), we compute the product beta * alpha

    // Useful in the computation of Q1c, Q2q, Q3q, Q4q

    int ind = blockIdx.x * blockDim.x + threadIdx.x;
    if (ind < MV){
        d_result[ind].x = d_betacomplex[ind].x * d_alphareal[ind].x;
        d_result[ind].y = d_betacomplex[ind].y * d_alphareal[ind].x;
    }
}



__global__ void compute_product_alphareal_Gcomplexconj(double *d_alphareal, cufftDoubleComplex *d_Gcomplex, cufftDoubleComplex *d_result, int MV){
    // Given an input G and a double array alpha, we compute the product bar(G) * alpha

    // Useful in the computation of Q1q

    int ind = blockIdx.x * blockDim.x + threadIdx.x;
    if (ind < MV){
        d_result[ind].x = d_Gcomplex[ind].x * d_alphareal[ind];
        d_result[ind].y = -d_Gcomplex[ind].y * d_alphareal[ind];
    }
}




__global__ void update_with_product_C_alphareal_betareal(cufftDoubleComplex *d_alphareal, cufftDoubleComplex *d_betareal, cufftDoubleComplex *d_gammareal, double C, int MV){
    // Given two input arrays alpha and beta being real (even if they are stored in complex arrays), we compute C * alpha * beta and add it to gamma

    // Useful in the computation of Q1c, Q2c, Q2q, Q3q, Q4q

    int ind = blockIdx.x * blockDim.x + threadIdx.x;
    if (ind < MV)
        d_gammareal[ind].x += C * d_alphareal[ind].x * d_betareal[ind].x;
}


__global__ void update_with_product_C_alphareal_betareal_gammareal(cufftDoubleComplex *d_alphareal, cufftDoubleComplex *d_betareal, cufftDoubleComplex *d_gammareal, cufftDoubleComplex *d_deltareal, double C, int MV){
    // Given 3 input arrays alpha, beta and gamma being real (even if they are stored in complex arrays), we compute C * alpha * beta * gamma and add it to delta

    // Useful in the computation of Q2q

    int ind = blockIdx.x * blockDim.x + threadIdx.x;
    if (ind < MV)
        d_deltareal[ind].x += C * d_alphareal[ind].x * d_betareal[ind].x * d_gammareal[ind].x;
}


__global__ void update_with_product_C_alphareal_betacomplex_gammacomplex_deltacomplex(cufftDoubleComplex *d_alphareal, cufftDoubleComplex *d_betacomplex, cufftDoubleComplex *d_gammacomplex, cufftDoubleComplex *d_deltacomplex, cufftDoubleComplex *d_result, double C, int j, int MV){
    // Given a real array alpha and 3 complex array beta, gamma, delta, we compute C * alpha * beta * gamma * delta
    int ind = blockIdx.x * blockDim.x + threadIdx.x;

    if (ind < MV){
        cufftDoubleComplex z1, z2, z3;

        z1.x = d_alphareal[j].x * d_betacomplex[ind].x;
        z1.y = d_alphareal[j].x * d_betacomplex[ind].y;

        z2.x = z1.x * d_gammacomplex[ind].x - z1.y * d_gammacomplex[ind].y;
        z2.y = z1.x * d_gammacomplex[ind].y + z1.y * d_gammacomplex[ind].x;

        z3.x = z2.x * d_deltacomplex[ind].x - z2.y * d_deltacomplex[ind].y;
        z3.y = z2.x * d_deltacomplex[ind].y + z2.y * d_deltacomplex[ind].x;

        d_result[ind].x += z3.x * C;
        d_result[ind].y += z3.y * C;
    }
}


__global__ void compute_product_alphareal_alphaprimereal_Gcomplex(double *d_alphareal, double *d_betareal, cufftDoubleComplex *d_Gcomplex, cufftDoubleComplex *d_result, int MV){
    // Given an input G and two double arrays alpha and beta, we compute the product G * alpha * beta

    // Useful in the computation of Q2c

    int ind = blockIdx.x * blockDim.x + threadIdx.x;
    if (ind < MV){
        d_result[ind].x = d_Gcomplex[ind].x * d_alphareal[ind] * d_betareal[ind];
        d_result[ind].y = d_Gcomplex[ind].y * d_alphareal[ind] * d_betareal[ind];
    }
}


__global__ void kill_imaginary_part(cufftDoubleComplex *d_Gcomplex, int MV){
    int ind = blockIdx.x * blockDim.x + threadIdx.x;
    if (ind < MV)
        d_Gcomplex[ind].y = 0.;
}


__global__ void kill_real_part(cufftDoubleComplex *d_Gcomplex, int MV){
    int ind = blockIdx.x * blockDim.x + threadIdx.x;
    if (ind < MV)
        d_Gcomplex[ind].x = 0.;
}


__global__ void normalize(cufftDoubleComplex *d_Gcomplex, int MV){
    int ind = blockIdx.x * blockDim.x + threadIdx.x;
    if (ind < MV){
        d_Gcomplex[ind].x /= ((double)MV);
        d_Gcomplex[ind].y /= ((double)MV);
    }
}


__global__ void compute_crossmodes_p2pi(cufftDoubleComplex *d_crossmode, int *d_modej, int *d_MVTAB, int ndim){
    int k = blockIdx.x * blockDim.x + threadIdx.x;

    int MVX = d_MVTAB[0];
    int MVY = d_MVTAB[1];
    int MVZ;
    if (ndim == 3)
        MVZ = d_MVTAB[2];
    else
        MVZ = 1;
    int MV = MVX*MVY*MVZ;
    
    if (k < MV){
        double pi = 4.*atan(1.);
    
        int itemp;
        int modek[3];
        double crossmode;

        itemp = k/MVZ;
        modek[2] = k - itemp*MVZ;
        modek[0] = itemp/MVY;
        modek[1] = itemp-modek[0]*MVY;
        for (int d = 0; d < ndim; d++){
            if (modek[d] >= d_MVTAB[d]/2)
                modek[d] -= d_MVTAB[d];
        }

        crossmode = 0.;
        for (int d = 0; d < ndim; d++)
			crossmode += (2.*pi*((double)modek[d])*((double)d_modej[d])/((double)d_MVTAB[d]));
        
        d_crossmode[k].x = cos(crossmode);
        d_crossmode[k].y = sin(crossmode);
    }
}


__global__ void add_real_part(cufftDoubleComplex *d_alpha, cufftDoubleComplex *d_result, double C, int MV){
    int ind = blockIdx.x * blockDim.x + threadIdx.x;
    if (ind < MV)
        d_result[ind].x += (C * d_alpha[ind].x);
}



void compute_Q1c_cufft(cufftDoubleComplex *d_Ginput_spectral, cufftDoubleComplex *d_Q1c_real, double **alpha, double **alphaprime, int MINT, int *MVTAB, int ndim){
    extern int nblocks;
    extern int nthreads_per_block;

    double pi = 4.*atan(1.);

    double C = pow(pi, ndim-1)/((double)MINT);

    int MVX = MVTAB[0];
    int MVY = MVTAB[1];
    int MVZ;
    if (ndim == 3)
        MVZ = MVTAB[2];
    else
        MVZ = 1;
    int MV = MVX*MVY*MVZ;

    double *d_alphap, *d_alphaprimep;
    int mem_size_real = sizeof(double) * MV;
    cudaMalloc((void**)&d_alphap, mem_size_real);
    cudaMalloc((void**)&d_alphaprimep, mem_size_real);

    cufftDoubleComplex *d_gp_spectral, *d_gprimep_spectral;
    cufftDoubleComplex *d_gp_real, *d_gprimep_real;
    int mem_size_complex = sizeof(cufftDoubleComplex) * MV;
    cudaMalloc((void**)&d_gp_spectral, mem_size_complex);
    cudaMalloc((void**)&d_gprimep_spectral, mem_size_complex);
    cudaMalloc((void**)&d_gp_real, mem_size_complex);
    cudaMalloc((void**)&d_gprimep_real, mem_size_complex);

    cufftHandle plan;
    if (ndim == 2)
        cufftPlan2d(&plan, MVX, MVY, CUFFT_Z2Z);
    else if (ndim == 3)
        cufftPlan3d(&plan, MVX, MVY, MVZ, CUFFT_Z2Z);

    // Initialize Q1c_spectral
    kill_real_part<<<nblocks,nthreads_per_block>>>(d_Q1c_real, MV);
    kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_Q1c_real, MV);

    for (int p = 0; p < MINT; p++){
        // Transfer alpha[p] and alphaprime[p] from CPU to GPU
        cudaMemcpy(d_alphap, alpha[p], mem_size_real, cudaMemcpyHostToDevice);
        cudaMemcpy(d_alphaprimep, alphaprime[p], mem_size_real, cudaMemcpyHostToDevice);

        // Compute ghat_p(k) = alpha_p(k) * fhat(k), gprimehat_p(k) = alphaprime_p(k) * fhat(k)
		// alpha_p and alphaprime_p are real numbers
        compute_product_alphareal_Gcomplex<<<nblocks,nthreads_per_block>>>(d_alphap, d_Ginput_spectral, d_gp_spectral, MV);
        compute_product_alphareal_Gcomplex<<<nblocks,nthreads_per_block>>>(d_alphaprimep, d_Ginput_spectral, d_gprimep_spectral, MV);

        // Do inverse FFTs ghat_p -> g_p, gprimehat_p -> gprime_p
        cufftExecZ2Z(plan, d_gp_spectral, d_gp_real, CUFFT_INVERSE);
        cufftExecZ2Z(plan, d_gprimep_spectral, d_gprimep_real, CUFFT_INVERSE);
        normalize<<<nblocks,nthreads_per_block>>>(d_gp_real, MV);
        normalize<<<nblocks,nthreads_per_block>>>(d_gprimep_real, MV);

        // Reminder: g_p and gprime_p are real numbers
        kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_gp_real, MV);
        kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_gprimep_real, MV);

        // Add C * g_p * gprime_p to Q1c
        update_with_product_C_alphareal_betareal<<<nblocks,nthreads_per_block>>>(d_gp_real, d_gprimep_real, d_Q1c_real, C, MV);
    }

    // Destroy cuFFT plan since it is now useless
    cufftDestroy(plan);

    // Free arrays on GPU that are now useless
    cudaFree(d_alphap);
    cudaFree(d_alphaprimep);
    cudaFree(d_gp_spectral);
    cudaFree(d_gprimep_spectral);
    cudaFree(d_gp_real);
    cudaFree(d_gprimep_real);
}



void compute_Q2c_cufft(cufftDoubleComplex *d_Ginput_spectral, cufftDoubleComplex *d_Ginput_real, cufftDoubleComplex *d_Q2c_real, double **alpha, double **alphaprime, int MINT, int *MVTAB, int ndim){
    extern int nblocks;
    extern int nthreads_per_block;

    double pi = 4.*atan(1.);

    double C = pow(pi, ndim-1)/((double)MINT);

    int MVX = MVTAB[0];
    int MVY = MVTAB[1];
    int MVZ;
    if (ndim == 3)
        MVZ = MVTAB[2];
    else
        MVZ = 1;
    int MV = MVX*MVY*MVZ;

    double *d_alphap, *d_alphaprimep;
    int mem_size_real = sizeof(double) * MV;
    cudaMalloc((void**)&d_alphap, mem_size_real);
    cudaMalloc((void**)&d_alphaprimep, mem_size_real);

    cufftDoubleComplex *d_gp_spectral, *d_gp_real;
    int mem_size_complex = sizeof(cufftDoubleComplex) * MV;
    cudaMalloc((void**)&d_gp_spectral, mem_size_complex);
    cudaMalloc((void**)&d_gp_real, mem_size_complex);
    cufftHandle plan;
    if (ndim == 2)
        cufftPlan2d(&plan, MVX, MVY, CUFFT_Z2Z);
    else if (ndim == 3)
        cufftPlan3d(&plan, MVX, MVY, MVZ, CUFFT_Z2Z);

    // Initialize Q2c_spectral
    kill_real_part<<<nblocks,nthreads_per_block>>>(d_Q2c_real, MV);
    kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_Q2c_real, MV);

    for (int p = 0; p < MINT; p++){
        // Transfer alpha[p] and alphaprime[p] from CPU to GPU
        cudaMemcpy(d_alphap, alpha[p], mem_size_real, cudaMemcpyHostToDevice);
        cudaMemcpy(d_alphaprimep, alphaprime[p], mem_size_real, cudaMemcpyHostToDevice);

        // Compute ghat_p(k) = alpha_p(k) * alphaprime_p(k) * fhat(k)
		// alpha_p and alphaprime_p are real numbers
        compute_product_alphareal_alphaprimereal_Gcomplex<<<nblocks,nthreads_per_block>>>(d_alphap, d_alphaprimep, d_Ginput_spectral, d_gp_spectral, MV);

        // Do inverse FFTs ghat_p -> g_p
        cufftExecZ2Z(plan, d_gp_spectral, d_gp_real, CUFFT_INVERSE);
        normalize<<<nblocks,nthreads_per_block>>>(d_gp_real, MV);

        // Reminder: g_p is real number
        kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_gp_real, MV);

        // Add C * g_p * f to Q2c
        update_with_product_C_alphareal_betareal<<<nblocks,nthreads_per_block>>>(d_gp_real, d_Ginput_real, d_Q2c_real, C, MV);
    }

    // Destroy cuFFT plan since it is now useless
    cufftDestroy(plan);

    // Free arrays on GPU that are now useless
    cudaFree(d_alphap);
    cudaFree(d_alphaprimep);
    cudaFree(d_gp_spectral);
    cudaFree(d_gp_real);
}


void compute_Q1q_cufft(cufftDoubleComplex *d_Ginput_real, cufftDoubleComplex *d_Q1q_real, double **alpha, double **alphaprime, int MINT, int *MVTAB, int ndim){
    extern int nblocks;
    extern int nthreads_per_block;

    double pi = 4.*atan(1.);

    double C = pow(pi, ndim-1)/((double)MINT);
    int modej[3];
    int itemp;

    int MVX = MVTAB[0];
    int MVY = MVTAB[1];
    int MVZ;
    if (ndim == 3)
        MVZ = MVTAB[2];
    else
        MVZ = 1;
    int MV = MVX*MVY*MVZ;

    cufftHandle plan;
    if (ndim == 2)
        cufftPlan2d(&plan, MVX, MVY, CUFFT_Z2Z);
    else if (ndim == 3)
        cufftPlan3d(&plan, MVX, MVY, MVZ, CUFFT_Z2Z);

    double *d_alphap, *d_alphaprimep;
    int mem_size_real = sizeof(double) * MV;
    cudaMalloc((void**)&d_alphap, mem_size_real);
    cudaMalloc((void**)&d_alphaprimep, mem_size_real);

    cufftDoubleComplex *d_crossmodej;
    cufftDoubleComplex *d_gpj_real, *d_gpj_spectral;
    cufftDoubleComplex *d_gprimepj_real, *d_gprimepj_spectral;
    cufftDoubleComplex *d_bigGpj_real, *d_bigGpj_spectral;
    cufftDoubleComplex *d_bigGprimepj_real, *d_bigGprimepj_spectral;
    cufftDoubleComplex *d_Q1q_spectral;
    int mem_size_complex = sizeof(cufftDoubleComplex) * MV;
    cudaMalloc((void**)&d_Q1q_spectral, mem_size_complex);
    cudaMalloc((void**)&d_crossmodej, mem_size_complex);
    cudaMalloc((void**)&d_gpj_real, mem_size_complex);
    cudaMalloc((void**)&d_gpj_spectral, mem_size_complex);
    cudaMalloc((void**)&d_gprimepj_real, mem_size_complex);
    cudaMalloc((void**)&d_gprimepj_spectral, mem_size_complex);
    cudaMalloc((void**)&d_bigGpj_real, mem_size_complex);
    cudaMalloc((void**)&d_bigGpj_spectral, mem_size_complex);
    cudaMalloc((void**)&d_bigGprimepj_real, mem_size_complex);
    cudaMalloc((void**)&d_bigGprimepj_spectral, mem_size_complex);

    int *d_modej, *d_MVTAB;
    cudaMalloc((void**)&d_modej, sizeof(int)*3);
    cudaMalloc((void**)&d_MVTAB, sizeof(int)*3);

    cudaMemcpy(d_MVTAB, MVTAB, sizeof(int)*3, cudaMemcpyHostToDevice);

    // Initialize Q1q_spectral
    kill_real_part<<<nblocks,nthreads_per_block>>>(d_Q1q_spectral, MV);
    kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_Q1q_spectral, MV);

    for (int j = 0; j < MV; j++){
        itemp = j/MVZ;
        modej[2] = j - itemp*MVZ;
        modej[0] = itemp/MVY;
        modej[1] = itemp-modej[0]*MVY;
        for (int d = 0; d < ndim; d++){
            if (modej[d] >= MVTAB[d]/2)
                modej[d] -= MVTAB[d];
        }

        cudaMemcpy(d_modej, modej, sizeof(int)*3, cudaMemcpyHostToDevice);

        // Compute exp(2*i*pi*k.j/N) for all k
        compute_crossmodes_p2pi<<<nblocks,nthreads_per_block>>>(d_crossmodej, d_modej, d_MVTAB, ndim);

        for (int p = 0; p < MINT; p++){
            // Transfer alpha[p] and alphaprime[p] from CPU to GPU
            cudaMemcpy(d_alphap, alpha[p], mem_size_real, cudaMemcpyHostToDevice);
            cudaMemcpy(d_alphaprimep, alphaprime[p], mem_size_real, cudaMemcpyHostToDevice);

            // Compute g_p,j_hat = alpha[p] * exp(-2i*pi*k.j/N) and g'_p,j_hat = alpha'[p] * exp(-2i*pi*k.j/N)
            compute_product_alphareal_Gcomplexconj<<<nblocks,nthreads_per_block>>>(d_alphap, d_crossmodej, d_gpj_spectral, MV);
            compute_product_alphareal_Gcomplexconj<<<nblocks,nthreads_per_block>>>(d_alphaprimep, d_crossmodej, d_gprimepj_spectral, MV);

            // g_p,j_hat and g'_p,j_hat are ready --> Perform inverse FFT to get g_p,j and g'_p,j
            cufftExecZ2Z(plan, d_gpj_spectral, d_gpj_real, CUFFT_INVERSE);
            cufftExecZ2Z(plan, d_gprimepj_spectral, d_gprimepj_real, CUFFT_INVERSE);
            normalize<<<nblocks,nthreads_per_block>>>(d_gpj_real, MV);
            normalize<<<nblocks,nthreads_per_block>>>(d_gprimepj_real, MV);

			// WARNING: g_p,j and g'_p,j are COMPLEX NUMBERS !!

            // Compute G_p,j = g_p,j * F and G'_p,j = g'_p,j * F
			// Reminder : F is real but g_p,j and g'_p,j are complex
            compute_product_alphareal_betacomplex<<<nblocks,nthreads_per_block>>>(d_Ginput_real, d_gpj_real, d_bigGpj_real, MV);
            compute_product_alphareal_betacomplex<<<nblocks,nthreads_per_block>>>(d_Ginput_real, d_gprimepj_real, d_bigGprimepj_real, MV);

            // Perform FFT to get G_p,j_hat and G'_p,j_hat
            cufftExecZ2Z(plan, d_bigGpj_real, d_bigGpj_spectral, CUFFT_FORWARD);
            cufftExecZ2Z(plan, d_bigGprimepj_real, d_bigGprimepj_spectral, CUFFT_FORWARD);

            // Update Q1q_hat[k] with C * G_p,j_hat[k] * G'_p,j_hat[k] * exp(2i*pi*k.j/N) * Gdist[j]
            update_with_product_C_alphareal_betacomplex_gammacomplex_deltacomplex<<<nblocks,nthreads_per_block>>>(d_Ginput_real, d_bigGpj_spectral, d_bigGprimepj_spectral, d_crossmodej, d_Q1q_spectral, C, j, MV);
        } // End of loop over p
    } // End of loop over j

    // Perform a backward FFT to get Q1q
    cufftExecZ2Z(plan, d_Q1q_spectral, d_Q1q_real, CUFFT_INVERSE);
    normalize<<<nblocks,nthreads_per_block>>>(d_Q1q_real, MV);
    kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_Q1q_real, MV);

    // Destroy cuFFT plan since it is now useless
    cufftDestroy(plan);

    // Free arrays on GPU that are now useless
    cudaFree(d_alphap);
    cudaFree(d_alphaprimep);

    cudaFree(d_Q1q_spectral);
    cudaFree(d_crossmodej);
    cudaFree(d_gpj_real);
    cudaFree(d_gpj_spectral);
    cudaFree(d_gprimepj_real);
    cudaFree(d_gprimepj_spectral);
    cudaFree(d_bigGpj_real);
    cudaFree(d_bigGpj_spectral);
    cudaFree(d_bigGprimepj_real);
    cudaFree(d_bigGprimepj_spectral);

    cudaFree(d_modej);
    cudaFree(d_MVTAB);
}


void compute_Q2q_cufft(cufftDoubleComplex *d_Ginput_spectral, cufftDoubleComplex *d_Ginput_real, cufftDoubleComplex *d_Q2q_real, double **alpha, double **alphaprime, int MINT, int *MVTAB, int ndim){
    extern int nblocks;
    extern int nthreads_per_block;

    double pi = 4.*atan(1.);

    double C = pow(pi, ndim-1)/((double)MINT);

    int MVX = MVTAB[0];
    int MVY = MVTAB[1];
    int MVZ;
    if (ndim == 3)
        MVZ = MVTAB[2];
    else
        MVZ = 1;
    int MV = MVX*MVY*MVZ;

    double *d_alphap, *d_alphaprimep;
    int mem_size_real = sizeof(double) * MV;
    cudaMalloc((void**)&d_alphap, mem_size_real);
    cudaMalloc((void**)&d_alphaprimep, mem_size_real);

    cufftDoubleComplex *d_gp_spectral, *d_gprimep_spectral;
    cufftDoubleComplex *d_gp_real, *d_gprimep_real;
    int mem_size_complex = sizeof(cufftDoubleComplex) * MV;
    cudaMalloc((void**)&d_gp_spectral, mem_size_complex);
    cudaMalloc((void**)&d_gprimep_spectral, mem_size_complex);
    cudaMalloc((void**)&d_gp_real, mem_size_complex);
    cudaMalloc((void**)&d_gprimep_real, mem_size_complex);

    cufftHandle plan;
    if (ndim == 2)
        cufftPlan2d(&plan, MVX, MVY, CUFFT_Z2Z);
    else if (ndim == 3)
        cufftPlan3d(&plan, MVX, MVY, MVZ, CUFFT_Z2Z);
    
    // Initialize Q2q_spectral
    kill_real_part<<<nblocks,nthreads_per_block>>>(d_Q2q_real, MV);
    kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_Q2q_real, MV);
    
    for (int p = 0; p < MINT; p++){
        // Transfer alpha[p] and alphaprime[p] from CPU to GPU
        cudaMemcpy(d_alphap, alpha[p], mem_size_real, cudaMemcpyHostToDevice);
        cudaMemcpy(d_alphaprimep, alphaprime[p], mem_size_real, cudaMemcpyHostToDevice);

        // Compute ghat_p(k) = alpha_p(k) * fhat(k), gprimehat_p(k) = alphaprime_p(k) * fhat(k)
		// alpha_p and alphaprime_p are real numbers
        compute_product_alphareal_Gcomplex<<<nblocks,nthreads_per_block>>>(d_alphap, d_Ginput_spectral, d_gp_spectral, MV);
        compute_product_alphareal_Gcomplex<<<nblocks,nthreads_per_block>>>(d_alphaprimep, d_Ginput_spectral, d_gprimep_spectral, MV);

        // Do inverse FFTs ghat_p -> g_p, gprimehat_p -> gprime_p
        cufftExecZ2Z(plan, d_gp_spectral, d_gp_real, CUFFT_INVERSE);
        cufftExecZ2Z(plan, d_gprimep_spectral, d_gprimep_real, CUFFT_INVERSE);
        normalize<<<nblocks,nthreads_per_block>>>(d_gp_real, MV);
        normalize<<<nblocks,nthreads_per_block>>>(d_gprimep_real, MV);

        // Reminder: g_p and gprime_p are real numbers
        kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_gp_real, MV);
        kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_gprimep_real, MV);

        // Add C * g_p * gprime_p * f to Q2q
        update_with_product_C_alphareal_betareal_gammareal<<<nblocks,nthreads_per_block>>>(d_gp_real, d_gprimep_real, d_Ginput_real, d_Q2q_real, C, MV);
    }

    // Destroy cuFFT plan since it is now useless
    cufftDestroy(plan);

    // Free arrays on GPU that are now useless
    cudaFree(d_alphap);
    cudaFree(d_alphaprimep);

    cudaFree(d_gp_spectral);
    cudaFree(d_gprimep_spectral);
    cudaFree(d_gp_real);
    cudaFree(d_gprimep_real);
}



void compute_Q3q_cufft(cufftDoubleComplex *d_Ginput_spectral, cufftDoubleComplex *d_Ginput_real, cufftDoubleComplex *d_Q3q_real, double **alpha, double **alphaprime, int MINT, int *MVTAB, int ndim){
    extern int nblocks;
    extern int nthreads_per_block;

    double pi = 4.*atan(1.);

    double C = pow(pi, ndim-1)/((double)MINT);

    int MVX = MVTAB[0];
    int MVY = MVTAB[1];
    int MVZ;
    if (ndim == 3)
        MVZ = MVTAB[2];
    else
        MVZ = 1;
    int MV = MVX*MVY*MVZ;

    double *d_alphap, *d_alphaprimep;
    int mem_size_real = sizeof(double) * MV;
    cudaMalloc((void**)&d_alphap, mem_size_real);
    cudaMalloc((void**)&d_alphaprimep, mem_size_real);

    cufftDoubleComplex *d_gprimep_spectral, *d_gprimep_real;
    cufftDoubleComplex *d_bigGprimep_spectral, *d_bigGprimep_real;
    cufftDoubleComplex *d_hp_spectral, *d_hp_real;
    int mem_size_complex = sizeof(cufftDoubleComplex) * MV;
    cudaMalloc((void**)&d_gprimep_spectral, mem_size_complex);
    cudaMalloc((void**)&d_gprimep_real, mem_size_complex);
    cudaMalloc((void**)&d_bigGprimep_spectral, mem_size_complex);
    cudaMalloc((void**)&d_bigGprimep_real, mem_size_complex);
    cudaMalloc((void**)&d_hp_spectral, mem_size_complex);
    cudaMalloc((void**)&d_hp_real, mem_size_complex);

    cufftHandle plan;
    if (ndim == 2)
        cufftPlan2d(&plan, MVX, MVY, CUFFT_Z2Z);
    else if (ndim == 3)
        cufftPlan3d(&plan, MVX, MVY, MVZ, CUFFT_Z2Z);
    
    // Initialize Q2q_spectral
    kill_real_part<<<nblocks,nthreads_per_block>>>(d_Q3q_real, MV);
    kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_Q3q_real, MV);

    for (int p = 0; p < MINT; p++){
        // Transfer alpha[p] and alphaprime[p] from CPU to GPU
        cudaMemcpy(d_alphap, alpha[p], mem_size_real, cudaMemcpyHostToDevice);
        cudaMemcpy(d_alphaprimep, alphaprime[p], mem_size_real, cudaMemcpyHostToDevice);

        // Compute gprimehat_p(k) = alphaprime_p(k) * fhat(k)
		// alphaprime_p is a real number
        compute_product_alphareal_Gcomplex<<<nblocks,nthreads_per_block>>>(d_alphaprimep, d_Ginput_spectral, d_gprimep_spectral, MV);

        // Do inverse FFT gprimehat_p -> gprime_p
        cufftExecZ2Z(plan, d_gprimep_spectral, d_gprimep_real, CUFFT_INVERSE);
        normalize<<<nblocks,nthreads_per_block>>>(d_gprimep_real, MV);

        // Reminder: gprime_p is real number
        kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_gprimep_real, MV);

        // Compute Gprime_p(k) = gprime_p(k) * f(k)
		// gprime_p and f are real numbers
        kill_real_part<<<nblocks,nthreads_per_block>>>(d_bigGprimep_real, MV);
        kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_bigGprimep_real, MV);
        update_with_product_C_alphareal_betareal<<<nblocks,nthreads_per_block>>>(d_gprimep_real, d_Ginput_real, d_bigGprimep_real, 1., MV);

        // Do FFT Gprime_p -> Gprimehat_p
        cufftExecZ2Z(plan, d_bigGprimep_real, d_bigGprimep_spectral, CUFFT_FORWARD);

        // Compute hhat_p(k) = alpha_p(k) * Gprimehat_p(k)
		// alpha_p is a real number
        compute_product_alphareal_Gcomplex<<<nblocks,nthreads_per_block>>>(d_alphap, d_bigGprimep_spectral, d_hp_spectral, MV);

        // Do inverse FFTs hhat_p -> h_p
        cufftExecZ2Z(plan, d_hp_spectral, d_hp_real, CUFFT_INVERSE);
        normalize<<<nblocks,nthreads_per_block>>>(d_hp_real, MV);

        // Reminder: h_p is real number
        kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_hp_real, MV);

        // Update Q3q with C * h_p * f
        update_with_product_C_alphareal_betareal<<<nblocks,nthreads_per_block>>>(d_hp_real, d_Ginput_real, d_Q3q_real, C, MV);
    }

    // Destroy cuFFT plan since it is now useless
    cufftDestroy(plan);

    // Free arrays on GPU that are now useless
    cudaFree(d_alphap);
    cudaFree(d_alphaprimep);

    cudaFree(d_gprimep_spectral);
    cudaFree(d_gprimep_real);
    cudaFree(d_bigGprimep_spectral);
    cudaFree(d_bigGprimep_real);
    cudaFree(d_hp_spectral);
    cudaFree(d_hp_real);
}



void compute_Q4q_cufft(cufftDoubleComplex *d_Ginput_spectral, cufftDoubleComplex *d_Ginput_real, cufftDoubleComplex *d_Q4q_real, double **alpha, double **alphaprime, int MINT, int *MVTAB, int ndim){
    extern int nblocks;
    extern int nthreads_per_block;

    double pi = 4.*atan(1.);

    double C = pow(pi, ndim-1)/((double)MINT);

    int MVX = MVTAB[0];
    int MVY = MVTAB[1];
    int MVZ;
    if (ndim == 3)
        MVZ = MVTAB[2];
    else
        MVZ = 1;
    int MV = MVX*MVY*MVZ;

    double *d_alphap, *d_alphaprimep;
    int mem_size_real = sizeof(double) * MV;
    cudaMalloc((void**)&d_alphap, mem_size_real);
    cudaMalloc((void**)&d_alphaprimep, mem_size_real);

    cufftDoubleComplex *d_gp_spectral, *d_gp_real;
    cufftDoubleComplex *d_bigGp_spectral, *d_bigGp_real;
    cufftDoubleComplex *d_hprimep_spectral, *d_hprimep_real;
    int mem_size_complex = sizeof(cufftDoubleComplex) * MV;
    cudaMalloc((void**)&d_gp_spectral, mem_size_complex);
    cudaMalloc((void**)&d_gp_real, mem_size_complex);
    cudaMalloc((void**)&d_bigGp_spectral, mem_size_complex);
    cudaMalloc((void**)&d_bigGp_real, mem_size_complex);
    cudaMalloc((void**)&d_hprimep_spectral, mem_size_complex);
    cudaMalloc((void**)&d_hprimep_real, mem_size_complex);

    cufftHandle plan;
    if (ndim == 2)
        cufftPlan2d(&plan, MVX, MVY, CUFFT_Z2Z);
    else if (ndim == 3)
        cufftPlan3d(&plan, MVX, MVY, MVZ, CUFFT_Z2Z);
    
    // Initialize Q2q_spectral
    kill_real_part<<<nblocks,nthreads_per_block>>>(d_Q4q_real, MV);
    kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_Q4q_real, MV);

    for (int p = 0; p < MINT; p++){
        // Transfer alpha[p] and alphaprime[p] from CPU to GPU
        cudaMemcpy(d_alphap, alpha[p], mem_size_real, cudaMemcpyHostToDevice);
        cudaMemcpy(d_alphaprimep, alphaprime[p], mem_size_real, cudaMemcpyHostToDevice);

        // Compute ghat_p(k) = alpha_p(k) * fhat(k)
		// alpha_p is a real number
        compute_product_alphareal_Gcomplex<<<nblocks,nthreads_per_block>>>(d_alphap, d_Ginput_spectral, d_gp_spectral, MV);

        // Do inverse FFT ghat_p -> g_p
        cufftExecZ2Z(plan, d_gp_spectral, d_gp_real, CUFFT_INVERSE);
        normalize<<<nblocks,nthreads_per_block>>>(d_gp_real, MV);

        // Reminder: g_p is real number
        kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_gp_real, MV);

        // Compute G_p(k) = g_p(k) * f(k)
		// g_p and f are real numbers
        kill_real_part<<<nblocks,nthreads_per_block>>>(d_bigGp_real, MV);
        kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_bigGp_real, MV);
        update_with_product_C_alphareal_betareal<<<nblocks,nthreads_per_block>>>(d_gp_real, d_Ginput_real, d_bigGp_real, 1., MV);

        // Do FFT G_p -> Ghat_p
        cufftExecZ2Z(plan, d_bigGp_real, d_bigGp_spectral, CUFFT_FORWARD);

        // Compute hprimehat_p(k) = alphaprime_p(k) * Ghat_p(k)
		// alphaprime_p is a real number
        compute_product_alphareal_Gcomplex<<<nblocks,nthreads_per_block>>>(d_alphaprimep, d_bigGp_spectral, d_hprimep_spectral, MV);

        // Do inverse FFTs hprimehat_p -> hprime_p
        cufftExecZ2Z(plan, d_hprimep_spectral, d_hprimep_real, CUFFT_INVERSE);
        normalize<<<nblocks,nthreads_per_block>>>(d_hprimep_real, MV);

        // Reminder: hprime_p is real number
        kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_hprimep_real, MV);

        // Update Q4q with C * hprime_p * f
        update_with_product_C_alphareal_betareal<<<nblocks,nthreads_per_block>>>(d_hprimep_real, d_Ginput_real, d_Q4q_real, C, MV);
    }

    // Destroy cuFFT plan since it is now useless
    cufftDestroy(plan);

    // Free arrays on GPU that are now useless
    cudaFree(d_alphap);
    cudaFree(d_alphaprimep);

    cudaFree(d_gp_spectral);
    cudaFree(d_gp_real);
    cudaFree(d_bigGp_spectral);
    cudaFree(d_bigGp_real);
    cudaFree(d_hprimep_spectral);
    cudaFree(d_hprimep_real);
}

void collision_Qtotal_cufft(double *Gloc, double *QGloc, double **alpha, double **alphaprime, double omega, PARAMETERS parameters){
    extern int nblocks;
    extern int nthreads_per_block;

    double pi = 4.*atan(1.);

    double elap_Q1c, elap_Q2c, elap_Q1q, elap_Q2q, elap_Q3q, elap_Q4q, tstart, tend;
    char hostname[MAXSTRINGLENGTH];
    char namefile[MAXSTRINGLENGTH];
    char suffix[MAXSTRINGLENGTH];
    FILE *f;

    cufftDoubleComplex *Gloc_cufft;
    cufftDoubleComplex *d_Ginput_real, *d_Ginput_spectral;
    cufftDoubleComplex *d_Qpart_real;
    cufftDoubleComplex *d_Qtotal_real;
    cufftDoubleComplex *QGloc_cufft;
    //cufftDoubleComplex *QGpart_cufft;

    int MVTAB[3];
    MVTAB[0] = parameters.nvx;
	MVTAB[1] = parameters.nvy;
    if (parameters.ndim == DIM3)
	    MVTAB[2] = parameters.nvz;
    else
        MVTAB[2] = 1;
	int MV = MVTAB[0]*MVTAB[1]*MVTAB[2];

    int MINT = parameters.mint;

    int ndim = parameters.ndim+1;

    int mem_size_complex = sizeof(cufftDoubleComplex) * MV;

    cufftHandle plan;
    if (ndim == 2)
        cufftPlan2d(&plan, MVTAB[0], MVTAB[1], CUFFT_Z2Z);
    else if (ndim == 3)
        cufftPlan3d(&plan, MVTAB[0], MVTAB[1], MVTAB[2], CUFFT_Z2Z);


    double myalpha, mu, myomega, phi;
	if (parameters.type_omega == OMEGA_IS_NOTUSED){
		mu = pow(parameters.L/pi, parameters.ndim+1);
		myomega = 1.;
	}
	else{
		mu = 1.;
		myomega = omega;
	}
	
	if (parameters.type_collision == QUANTUM_COLLISIONS){
		if (parameters.type_qparticles == BOSONS)
			myalpha = pow(parameters.hbar,parameters.ndim+1);
		else if (parameters.type_qparticles == FERMIONS)
			myalpha = -pow(parameters.hbar,parameters.ndim+1);
	}
	else if (parameters.type_collision == CLASSICAL_COLLISIONS)
		myalpha = 0.;
	phi = myalpha * mu * pow(pi*myomega/parameters.L, parameters.ndim+1);

    // Copy Gloc in a cufft array on CPU
    Gloc_cufft = (cufftDoubleComplex*)malloc(mem_size_complex);
    QGloc_cufft = (cufftDoubleComplex*)malloc(mem_size_complex);
    for (int i = 0; i < MV; i++){
        Gloc_cufft[i].x = Gloc[i];
        Gloc_cufft[i].y = 0.;
    }

    //QGpart_cufft = (cufftDoubleComplex*)malloc(mem_size_complex);

    // Transfer Gloc from CPU to GPU
    cudaMalloc((void**)&d_Ginput_real, mem_size_complex);
    cudaMemcpy(d_Ginput_real, Gloc_cufft, mem_size_complex, cudaMemcpyHostToDevice);

    // Perform FFT on Gloc to get Ghat on GPU
    cudaMalloc((void**)&d_Ginput_spectral, mem_size_complex);
    kill_real_part<<<nblocks,nthreads_per_block>>>(d_Ginput_spectral, MV);
    kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_Ginput_spectral, MV);
    cufftExecZ2Z(plan, d_Ginput_real, d_Ginput_spectral, CUFFT_FORWARD);

    // Prepare QGloc on GPU
    cudaMalloc((void**)&d_Qtotal_real, mem_size_complex);
    kill_real_part<<<nblocks,nthreads_per_block>>>(d_Qtotal_real, MV);
    kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_Qtotal_real, MV);

    cudaMalloc((void**)&d_Qpart_real, mem_size_complex);

    // Compute Q1c on GPU
    tstart = (double)clock();
    compute_Q1c_cufft(d_Ginput_spectral, d_Qpart_real, alpha, alphaprime, MINT, MVTAB, ndim);
    tend = (double)clock();
	elap_Q1c = (tend-tstart)/CLOCKS_PER_SEC;
    add_real_part<<<nblocks,nthreads_per_block>>>(d_Qpart_real, d_Qtotal_real, 1., MV);


    /*cudaMemcpy(QGpart_cufft, d_Qpart_real, mem_size_complex, cudaMemcpyDeviceToHost);
    f = fopen("Q1c_cuda.txt", "w");
    for (int i = 0; i < MV; i++)
        fprintf(f, "%5d, %.15e %.15e\n", i, QGpart_cufft[i].x, QGpart_cufft[i].y);
    fclose(f);*/

    // Compute Q2c on GPU
    tstart = (double)clock();
    compute_Q2c_cufft(d_Ginput_spectral, d_Ginput_real, d_Qpart_real, alpha, alphaprime, MINT, MVTAB, ndim);
    tend = (double)clock();
	elap_Q2c = (tend-tstart)/CLOCKS_PER_SEC;
    add_real_part<<<nblocks,nthreads_per_block>>>(d_Qpart_real, d_Qtotal_real, -1., MV);

    /*cudaMemcpy(QGpart_cufft, d_Qpart_real, mem_size_complex, cudaMemcpyDeviceToHost);
    f = fopen("Q2c_cuda.txt", "w");
    for (int i = 0; i < MV; i++)
        fprintf(f, "%5d, %.15e %.15e\n", i, QGpart_cufft[i].x, QGpart_cufft[i].y);
    fclose(f);*/

    if (parameters.type_collision == QUANTUM_COLLISIONS){
        // Compute Q1q on GPU
        tstart = (double)clock();
        compute_Q1q_cufft(d_Ginput_real, d_Qpart_real, alpha, alphaprime, MINT, MVTAB, ndim);
        tend = (double)clock();
	    elap_Q1q = (tend-tstart)/CLOCKS_PER_SEC;
        add_real_part<<<nblocks,nthreads_per_block>>>(d_Qpart_real, d_Qtotal_real, phi, MV);

        /*cudaMemcpy(QGpart_cufft, d_Qpart_real, mem_size_complex, cudaMemcpyDeviceToHost);
        f = fopen("Q1q_cuda.txt", "w");
        for (int i = 0; i < MV; i++)
            fprintf(f, "%5d, %.15e %.15e\n", i, QGpart_cufft[i].x, QGpart_cufft[i].y);
        fclose(f);*/

        // Compute Q2q on GPU
        tstart = (double)clock();
        compute_Q2q_cufft(d_Ginput_spectral, d_Ginput_real, d_Qpart_real, alpha, alphaprime, MINT, MVTAB, ndim);
        tend = (double)clock();
	    elap_Q2q = (tend-tstart)/CLOCKS_PER_SEC;
        add_real_part<<<nblocks,nthreads_per_block>>>(d_Qpart_real, d_Qtotal_real, phi, MV);

        /*cudaMemcpy(QGpart_cufft, d_Qpart_real, mem_size_complex, cudaMemcpyDeviceToHost);
        f = fopen("Q2q_cuda.txt", "w");
        for (int i = 0; i < MV; i++)
            fprintf(f, "%5d, %.15e %.15e\n", i, QGpart_cufft[i].x, QGpart_cufft[i].y);
        fclose(f);*/

        // Compute Q3q on GPU
        tstart = (double)clock();
        compute_Q3q_cufft(d_Ginput_spectral, d_Ginput_real, d_Qpart_real, alpha, alphaprime, MINT, MVTAB, ndim);
        tend = (double)clock();
        elap_Q3q = (tend-tstart)/CLOCKS_PER_SEC;
        add_real_part<<<nblocks,nthreads_per_block>>>(d_Qpart_real, d_Qtotal_real, -phi, MV);

        /*cudaMemcpy(QGpart_cufft, d_Qpart_real, mem_size_complex, cudaMemcpyDeviceToHost);
        f = fopen("Q3q_cuda.txt", "w");
        for (int i = 0; i < MV; i++)
            fprintf(f, "%5d, %.15e %.15e\n", i, QGpart_cufft[i].x, QGpart_cufft[i].y);
        fclose(f);*/

        // Compute Q4q on GPU
        tstart = (double)clock();
        compute_Q4q_cufft(d_Ginput_spectral, d_Ginput_real, d_Qpart_real, alpha, alphaprime, MINT, MVTAB, ndim);
        tend = (double)clock();
	    elap_Q4q = (tend-tstart)/CLOCKS_PER_SEC;
        add_real_part<<<nblocks,nthreads_per_block>>>(d_Qpart_real, d_Qtotal_real, -phi, MV);

        /*cudaMemcpy(QGpart_cufft, d_Qpart_real, mem_size_complex, cudaMemcpyDeviceToHost);
        f = fopen("Q4q_cuda.txt", "w");
        for (int i = 0; i < MV; i++)
            fprintf(f, "%5d, %.15e %.15e\n", i, QGpart_cufft[i].x, QGpart_cufft[i].y);
        fclose(f);*/
    }

    // Transfer QGloc from GPU to CPU
    cudaMemcpy(QGloc_cufft, d_Qtotal_real, mem_size_complex, cudaMemcpyDeviceToHost);
    for (int i = 0; i < MV; i++)
        QGloc[i] = QGloc_cufft[i].x;
    

    gethostname(hostname, MAXSTRINGLENGTH);
    strcpy(namefile, "");
    strcat(namefile, parameters.outputdir);
    strcat(namefile, "/");
    strcat(namefile, parameters.prefix);
    sprintf(suffix, "_timecost_collision_%d_blocks_%d_threads.txt", nblocks, nthreads_per_block);
    strcat(namefile, suffix);
    f = fopen(namefile, "w");
    fprintf(f, "# Kind of particles\n");
    if (parameters.type_collision == CLASSICAL_COLLISIONS)
        fprintf(f, "classical\n");
    else{
        if (parameters.type_qparticles == FERMIONS)
            fprintf(f, "fermions\n");
        else
            fprintf(f, "bosons\n");
    }
    fprintf(f, "# Number of grid points\n");
    fprintf(f, "%d %d", MVTAB[0], MVTAB[1]);
    if (parameters.ndim == DIM3)
        fprintf(f, " %d", MVTAB[2]);
    fprintf(f, "\n");

    fprintf(f, "# Number of blocks\n");
    fprintf(f, "%d\n", nblocks);

    fprintf(f, "# Number of threads per block\n");
    fprintf(f, "%d\n", nthreads_per_block);

    fprintf(f, "# Rank |    Time cost Q1c    |    Time cost Q2c    ");
    if (parameters.type_collision == QUANTUM_COLLISIONS)
        fprintf(f, "|    Time cost Q1q    |    Time cost Q2q    |    Time cost Q3q    |    Time cost Q4q    ");
    fprintf(f, "| Hostname\n");
    fprintf(f, "%7d|%.15e|%.15e", 0, elap_Q1c, elap_Q2c);
    if (parameters.type_collision == QUANTUM_COLLISIONS)
        fprintf(f, "|%.15e|%.15e|%.15e|%.15e", elap_Q1q, elap_Q2q, elap_Q3q, elap_Q4q);
    fprintf(f, "|%s", hostname);
    fprintf(f, "\n");
    fclose(f);
    
    cufftDestroy(plan);

    free(Gloc_cufft);
    cudaFree(d_Ginput_real);
    cudaFree(d_Ginput_spectral);
    cudaFree(d_Qpart_real);
    cudaFree(d_Qtotal_real);
    free(QGloc_cufft);
    //free(QGpart_cufft);
}





__global__ void compute_ABCGhat_spectral(cufftDoubleComplex *d_AGhat_spectral, cufftDoubleComplex *d_BGhat_spectral, cufftDoubleComplex *d_CGhat_spectral, cufftDoubleComplex *d_G_spectral, double *d_xitilde, double *d_sigma, double normxi, int *d_MVTAB, int ndim){
    int k = blockIdx.x * blockDim.x + threadIdx.x;

    int MV = d_MVTAB[0]*d_MVTAB[1];
    if (ndim == 3)
        MV *= d_MVTAB[2];

    if (k < MV){
        double pi = 4.*atan(1.);
        double mode;
        double A_re, A_im, B_re, B_im, C_re, C_im;
    
        int itemp;
        int modek[3];

        itemp = k/d_MVTAB[2];
        modek[2] = k - itemp*d_MVTAB[2];
        modek[0] = itemp/d_MVTAB[1];
        modek[1] = itemp-modek[0]*d_MVTAB[1];
        for (int d = 0; d < ndim; d++){
            if (modek[d] >= d_MVTAB[d]/2)
                modek[d] -= d_MVTAB[d];
        }

        mode = 0.;
        for (int d = 0; d <= ndim; d++)
            mode += (modek[d]*(0.5*d_xitilde[d] + 0.5*normxi*d_sigma[d]));
        A_re = cos(mode);
        A_im = sin(mode);

        mode = 0.;
        for (int d = 0; d <= ndim; d++)
            mode += (modek[d]*(0.5*d_xitilde[d] - 0.5*normxi*d_sigma[d]));
        B_re = cos(mode);
        B_im = sin(mode);

        mode = 0.;
        for (int d = 0; d <= ndim; d++)
            mode += (modek[d]*d_xitilde[d]);
        C_re = cos(mode);
        C_im = sin(mode);

        d_AGhat_spectral[k].x = A_re * d_G_spectral[k].x - A_im * d_G_spectral[k].y;
        d_AGhat_spectral[k].y = A_re * d_G_spectral[k].y + A_im * d_G_spectral[k].x;

        d_BGhat_spectral[k].x = B_re * d_G_spectral[k].x - B_im * d_G_spectral[k].y;
        d_BGhat_spectral[k].y = B_re * d_G_spectral[k].y + B_im * d_G_spectral[k].x;
        
        d_CGhat_spectral[k].x = C_re * d_G_spectral[k].x - C_im * d_G_spectral[k].y;
        d_CGhat_spectral[k].y = C_re * d_G_spectral[k].y + C_im * d_G_spectral[k].x;
    }
}



__global__ void update_Qtotal_quick_convolution(cufftDoubleComplex *d_AGhat_real, cufftDoubleComplex *d_BGhat_real, cufftDoubleComplex *d_CGhat_real, cufftDoubleComplex *d_G_real, double *d_QGloc_real, double W, double phi, int MV){
    int k = blockIdx.x * blockDim.x + threadIdx.x;

    if (k < MV){
        double contrib_Q1c, contrib_Q2c, contrib_Q1q, contrib_Q2q, contrib_Q3q, contrib_Q4q;
        // Add W[p] * IFFT(A[p]*Ghat) * IFFT(B[p]*Ghat) to Q1c
        contrib_Q1c = (d_AGhat_real[k].x*d_BGhat_real[k].x - d_AGhat_real[k].y*d_BGhat_real[k].y);

        // Add W[p] * IFFT(C[p]*Ghat) * G to Q2c
        contrib_Q2c = (d_CGhat_real[k].x*d_G_real[k].x);

        // Add W[p] * IFFT(A[p]*Ghat) * IFFT(B[p]*Ghat) * IFFT(C[p]*Ghat) to Q1q
        contrib_Q1q = ((d_AGhat_real[k].x*d_BGhat_real[k].x - d_AGhat_real[k].y*d_BGhat_real[k].y)*d_CGhat_real[k].x - (d_AGhat_real[k].x*d_BGhat_real[k].y + d_AGhat_real[k].y*d_BGhat_real[k].x)*d_CGhat_real[k].y);
        
        // Add W[p] * IFFT(A[p]*Ghat) * IFFT(B[p]*Ghat) * G to Q2q
        contrib_Q2q = ((d_AGhat_real[k].x*d_BGhat_real[k].x - d_AGhat_real[k].y*d_BGhat_real[k].y)*d_G_real[k].x);

        // Add W[p] * IFFT(A[p]*Ghat) * IFFT(C[p]*Ghat) * G to Q3q
        contrib_Q3q = ((d_AGhat_real[k].x*d_CGhat_real[k].x - d_AGhat_real[k].y*d_CGhat_real[k].y)*d_G_real[k].x);

        // Add W[p] * IFFT(B[p]*Ghat) * IFFT(C[p]*Ghat) * G to Q4q
        contrib_Q4q = ((d_BGhat_real[k].x*d_CGhat_real[k].x - d_BGhat_real[k].y*d_CGhat_real[k].y)*d_G_real[k].x);

        d_QGloc_real[k] += (W * (contrib_Q1c - contrib_Q2c + phi * (contrib_Q1q + contrib_Q2q - contrib_Q3q - contrib_Q4q)));
    }
}


__global__ void set_to_zero(double *d_tab, int MV){
    int k = blockIdx.x * blockDim.x + threadIdx.x;

    if (k < MV)
        d_tab[k] = 0.;
}



void collision_Qtotal_quick_convolution_cufft(double *Gloc, double *QGloc, double **xitilde, double **sigma, double *W, double omega, PARAMETERS parameters){
    double elap, tstart, tend;
    char hostname[MAXSTRINGLENGTH];
    char namefile[MAXSTRINGLENGTH];
    char suffix[MAXSTRINGLENGTH];
    FILE *f;

    extern int nblocks;
    extern int nthreads_per_block;

    double pi = 4.*atan(1.);
    int ndim = parameters.ndim+1;

    int MVTAB[3];
    MVTAB[0] = parameters.nvx;
	MVTAB[1] = parameters.nvy;
    MVTAB[2] = max(parameters.nvz,1);
	int MV = MVTAB[0]*MVTAB[1]*MVTAB[2];

    int *d_MVTAB;

    int mem_size_complex = sizeof(cufftDoubleComplex) * MV;
    int mem_size_double = sizeof(double) * MV;
    int mem_size_xitilde = sizeof(double)*ndim;

    cufftDoubleComplex *Gloc_cufft;
    cufftDoubleComplex *d_Ginput_real, *d_Ginput_spectral;
    cufftDoubleComplex *d_AGhat_spectral, *d_AGhat_real;
    cufftDoubleComplex *d_BGhat_spectral, *d_BGhat_real;
    cufftDoubleComplex *d_CGhat_spectral, *d_CGhat_real;
    double *d_Qtotal_real;

    double normxi;

    double *d_xitilde, *d_sigma;

    cufftHandle plan;
    if (ndim == 2)
        cufftPlan2d(&plan, MVTAB[0], MVTAB[1], CUFFT_Z2Z);
    else if (ndim == 3)
        cufftPlan3d(&plan, MVTAB[0], MVTAB[1], MVTAB[2], CUFFT_Z2Z);
    
    double myalpha, mu, myomega, phi;
	if (parameters.type_omega == OMEGA_IS_NOTUSED){
		mu = pow(parameters.L/pi, parameters.ndim+1);
		myomega = 1.;
	}
	else{
		mu = 1.;
		myomega = omega;
	}
	
	if (parameters.type_collision == QUANTUM_COLLISIONS){
		if (parameters.type_qparticles == BOSONS)
			myalpha = pow(parameters.hbar,parameters.ndim+1);
		else if (parameters.type_qparticles == FERMIONS)
			myalpha = -pow(parameters.hbar,parameters.ndim+1);
	}
	else if (parameters.type_collision == CLASSICAL_COLLISIONS)
		myalpha = 0.;
	phi = myalpha * mu * pow(pi*myomega/parameters.L, parameters.ndim+1);

    cudaMalloc((void**)&d_MVTAB, sizeof(int)*3);
    cudaMemcpy(d_MVTAB, MVTAB, sizeof(int)*3, cudaMemcpyHostToDevice);

    // Copy Gloc in a cufft array on CPU
    Gloc_cufft = (cufftDoubleComplex*)malloc(mem_size_complex);
    for (int i = 0; i < MV; i++){
        Gloc_cufft[i].x = Gloc[i];
        Gloc_cufft[i].y = 0.;
    }
    // Transfer Gloc from CPU to GPU
    cudaMalloc((void**)&d_Ginput_real, mem_size_complex);
    cudaMemcpy(d_Ginput_real, Gloc_cufft, mem_size_complex, cudaMemcpyHostToDevice);
    // Prepare QGloc on GPU
    cudaMalloc((void**)&d_Qtotal_real, mem_size_double);
    set_to_zero<<<nblocks,nthreads_per_block>>>(d_Qtotal_real, MV);
    cudaMalloc((void**)&d_AGhat_spectral, mem_size_complex);
    cudaMalloc((void**)&d_AGhat_real, mem_size_complex);
    cudaMalloc((void**)&d_BGhat_spectral, mem_size_complex);
    cudaMalloc((void**)&d_BGhat_real, mem_size_complex);
    cudaMalloc((void**)&d_CGhat_spectral, mem_size_complex);
    cudaMalloc((void**)&d_CGhat_real, mem_size_complex);
    cudaMalloc((void**)&d_xitilde, mem_size_xitilde);
    cudaMalloc((void**)&d_sigma, mem_size_xitilde);

    // Perform FFT on Gloc to get Ghat on GPU
    cudaMalloc((void**)&d_Ginput_spectral, mem_size_complex);
    kill_real_part<<<nblocks,nthreads_per_block>>>(d_Ginput_spectral, MV);
    kill_imaginary_part<<<nblocks,nthreads_per_block>>>(d_Ginput_spectral, MV);
    tstart = (double)clock();
    cufftExecZ2Z(plan, d_Ginput_real, d_Ginput_spectral, CUFFT_FORWARD);

    for (int p = 0; p < parameters.mint; p++){
		normxi = 0.;
		for (int d = 0; d <= parameters.ndim; d++)
			normxi += pow(xitilde[p][d], 2);
		normxi = sqrt(normxi);

        cudaMemcpy(d_xitilde, xitilde[p], mem_size_xitilde, cudaMemcpyHostToDevice);
        cudaMemcpy(d_sigma, sigma[p], mem_size_xitilde, cudaMemcpyHostToDevice);

        compute_ABCGhat_spectral<<<nblocks,nthreads_per_block>>>(d_AGhat_spectral, d_BGhat_spectral, d_CGhat_spectral, d_Ginput_spectral, d_xitilde, d_sigma, normxi, d_MVTAB, ndim);

        cufftExecZ2Z(plan, d_AGhat_spectral, d_AGhat_real, CUFFT_INVERSE);
        normalize<<<nblocks,nthreads_per_block>>>(d_AGhat_real, MV);

        cufftExecZ2Z(plan, d_BGhat_spectral, d_BGhat_real, CUFFT_INVERSE);
        normalize<<<nblocks,nthreads_per_block>>>(d_BGhat_real, MV);

        cufftExecZ2Z(plan, d_CGhat_spectral, d_CGhat_real, CUFFT_INVERSE);
        normalize<<<nblocks,nthreads_per_block>>>(d_CGhat_real, MV);

        update_Qtotal_quick_convolution<<<nblocks,nthreads_per_block>>>(d_AGhat_real, d_BGhat_real, d_CGhat_real, d_Ginput_real, d_Qtotal_real, W[p], phi, MV);
    }

    cudaMemcpy(QGloc, d_Qtotal_real, mem_size_double, cudaMemcpyDeviceToHost);
    tend = (double)clock();
    elap = tend-tstart;

    gethostname(hostname, MAXSTRINGLENGTH);
    strcpy(namefile, "");
    strcat(namefile, parameters.outputdir);
    strcat(namefile, "/");
    strcat(namefile, parameters.prefix);
    sprintf(suffix, "_timecost_collision_%d_blocks_%d_threads.txt", nblocks, nthreads_per_block);
    strcat(namefile, suffix);
    f = fopen(namefile, "w");
    fprintf(f, "# Kind of particles\n");
    if (parameters.type_collision == CLASSICAL_COLLISIONS)
        fprintf(f, "classical\n");
    else{
        if (parameters.type_qparticles == FERMIONS)
            fprintf(f, "fermions\n");
        else
            fprintf(f, "bosons\n");
    }
    fprintf(f, "# Number of grid points\n");
    fprintf(f, "%d %d", MVTAB[0], MVTAB[1]);
    if (parameters.ndim == DIM3)
        fprintf(f, " %d", MVTAB[2]);
    fprintf(f, "\n");

    fprintf(f, "# Number of blocks\n");
    fprintf(f, "%d\n", nblocks);

    fprintf(f, "# Number of threads per block\n");
    fprintf(f, "%d\n", nthreads_per_block);

    fprintf(f, "# Rank |    Time cost Qtotal   |     Hostname\n");
    fprintf(f, "%7d|%.15e|%s", 0, elap, hostname);
    fprintf(f, "\n");
    fclose(f);


    cufftDestroy(plan);

    cudaFree(d_MVTAB);
    cudaFree(d_Ginput_real);
    cudaFree(d_Ginput_spectral);

    cudaFree(d_AGhat_spectral);
    cudaFree(d_AGhat_real);
    cudaFree(d_BGhat_spectral);
    cudaFree(d_BGhat_real);
    cudaFree(d_CGhat_spectral);
    cudaFree(d_CGhat_real);

    cudaFree(d_xitilde);
    cudaFree(d_sigma);

    cudaFree(d_Qtotal_real);

    free(Gloc_cufft);
}




#endif