#include "./src/core/divergence.h"


#if defined(key_fftw_mpi) || defined(key_fftw) || defined(key_fftw_openmp)
void compute_deriv_vx_fftw(fftw_complex *G_real, fftw_complex *dvx_G_real, PARAMETERS parameters){

	int rank, commsize, nthreads;
#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &commsize);
	MPI_Status status;
#else
	rank = 0;
	commsize = 1;
#endif

#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif
#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif

	double pi = 4.*atan(1.);
	int NVXLOC = (int)parameters.nvxloc;
	int IVX0LOC = (int)parameters.ivx0loc;
	int MVX = parameters.nvx;
	int ivxabs;
	double mode;
	double *G_send, *G_recv, *dvxG_send, *dvxG_recv;

	fftw_complex *Gglobal_real, *dvx_Gglobal_real;
	fftw_complex *Gglobal_spectral, *dvx_Gglobal_spectral;
	fftw_plan p_G_forward, p_dvxG_backward;

	double C = (double)MVX / (2.*parameters.L);

	// WARNING : with FFTW-MPI, the data is distributed in vx-direction
	// --> We have to gather it on a single CPU before doing the computations
	// with FFTW-noMPI, the data is global in vx-direction
	if (rank == 0){
		Gglobal_real = fftw_malloc(sizeof(fftw_complex)*MVX);
		Gglobal_spectral = fftw_malloc(sizeof(fftw_complex)*MVX);
		dvx_Gglobal_spectral = fftw_malloc(sizeof(fftw_complex)*MVX);
		dvx_Gglobal_real = fftw_malloc(sizeof(fftw_complex)*MVX);

		// Receive all local arrays from others CPUs and store them in a global array
		if (parameters.nvxtab[0] > 0){
			for (int ivx = 0; ivx < parameters.nvxtab[0]; ivx++){
				Gglobal_real[parameters.ivx0tab[0]+ivx][0] = G_real[ivx][0];
				Gglobal_real[parameters.ivx0tab[0]+ivx][1] = G_real[ivx][1];
			}
		}
#ifdef key_mpi
		for (int r = 1; r < commsize; r++){
			if (parameters.nvxtab[r] > 0){
				G_recv = allocation_1d(parameters.nvxtab[r]);
				MPI_Recv(G_recv, parameters.nvxtab[r], MPI_DOUBLE, r, r, MPI_COMM_WORLD, &status);
				for (int ivx = 0; ivx < parameters.nvxtab[r]; ivx++)
					Gglobal_real[parameters.ivx0tab[r]+ivx][0] = G_recv[ivx];
				MPI_Recv(G_recv, parameters.nvxtab[r], MPI_DOUBLE, r, commsize+r, MPI_COMM_WORLD, &status);
				for (int ivx = 0; ivx < parameters.nvxtab[r]; ivx++)
					Gglobal_real[parameters.ivx0tab[r]+ivx][1] = G_recv[ivx];
				free_1d(G_recv);
			}
		}
#endif

		// Prepare FFTW plans
		for (int ivx = 0; ivx < MVX; ivx++){
			Gglobal_spectral[ivx][0] = 0.;
			Gglobal_spectral[ivx][1] = 0.;
			dvx_Gglobal_spectral[ivx][0] = 0.;
			dvx_Gglobal_spectral[ivx][1] = 0.;
			dvx_Gglobal_real[ivx][0] = 0.;
			dvx_Gglobal_real[ivx][1] = 0.;
		}
		p_G_forward = fftw_plan_dft_1d(MVX, Gglobal_real, Gglobal_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
		p_dvxG_backward = fftw_plan_dft_1d(MVX, dvx_Gglobal_spectral, dvx_Gglobal_real, FFTW_BACKWARD, FFTW_ESTIMATE);

		// Compute G_hat
		fftw_execute(p_G_forward);

		// Compute dvxG_hat
		for (int ivx = 0; ivx < MVX; ivx++){
			if (ivx < MVX/2)
				ivxabs = ivx;
			else if (ivx > MVX/2)
				ivxabs = ivx-MVX;
			else if (ivx == MVX)
				ivxabs = 0;

			mode = 2*pi*((double)ivxabs)/((double)MVX);

			dvx_Gglobal_spectral[ivx][0] = -mode * Gglobal_spectral[ivx][1];
			dvx_Gglobal_spectral[ivx][1] = mode * Gglobal_spectral[ivx][0];
		}

		// Compute dvxG
		fftw_execute(p_dvxG_backward);
		for (int ivx = 0; ivx < MVX; ivx++){
			dvx_Gglobal_real[ivx][0] /= (double)MVX;
			dvx_Gglobal_real[ivx][1] /= (double)MVX;
		}

		for (int ivx = 0; ivx < MVX; ivx++){
			dvx_Gglobal_real[ivx][0] *= C;
			dvx_Gglobal_real[ivx][1] *= C;
		}

		// Distribute the global array of dvxG to local arrays
		if (parameters.nvxtab[0] > 0){
			for (int ivx = 0; ivx < parameters.nvxtab[0]; ivx++){
				dvx_G_real[ivx][0] = dvx_Gglobal_real[parameters.ivx0tab[0]+ivx][0];
				dvx_G_real[ivx][1] = dvx_Gglobal_real[parameters.ivx0tab[0]+ivx][1];
			}
		}
#ifdef key_mpi
		for (int r = 1; r < commsize; r++){
			if (parameters.nvxtab[r] > 0){
				dvxG_send = allocation_1d(parameters.nvxtab[r]);
				for (int ivx = 0; ivx < parameters.nvxtab[r]; ivx++)
					dvxG_send[ivx] = dvx_Gglobal_real[parameters.ivx0tab[r]+ivx][0];
				MPI_Send(dvxG_send, parameters.nvxtab[r], MPI_DOUBLE, r, 2*commsize+r, MPI_COMM_WORLD);

				for (int ivx = 0; ivx < parameters.nvxtab[r]; ivx++)
					dvxG_send[ivx] = dvx_Gglobal_real[parameters.ivx0tab[r]+ivx][1];
				MPI_Send(dvxG_send, parameters.nvxtab[r], MPI_DOUBLE, r, 3*commsize+r, MPI_COMM_WORLD);
				free_1d(dvxG_send);
			}
		}
#endif

		fftw_destroy_plan(p_G_forward);
		fftw_destroy_plan(p_dvxG_backward);
		
		fftw_free(Gglobal_real);
		fftw_free(Gglobal_spectral);
		fftw_free(dvx_Gglobal_spectral);
		fftw_free(dvx_Gglobal_real);
	}
#ifdef key_mpi
	else{
		// Send local array to CPU 0
		if (NVXLOC > 0){
			G_send = allocation_1d(NVXLOC);
			for (int ivx = 0; ivx < NVXLOC; ivx++)
				G_send[ivx] = G_real[ivx][0];
			MPI_Send(G_send, NVXLOC, MPI_DOUBLE, 0, rank, MPI_COMM_WORLD);
			for (int ivx = 0; ivx < NVXLOC; ivx++)
				G_send[ivx] = G_real[ivx][1];
			MPI_Send(G_send, NVXLOC, MPI_DOUBLE, 0, commsize+rank, MPI_COMM_WORLD);
			free_1d(G_send);
		}

		// Receive local array from CPU 0
		if (NVXLOC > 0){
			dvxG_recv = allocation_1d(NVXLOC);
			MPI_Recv(dvxG_recv, NVXLOC, MPI_DOUBLE, 0, 2*commsize+rank, MPI_COMM_WORLD, &status);
			for (int ivx = 0; ivx < NVXLOC; ivx++)
				dvx_G_real[ivx][0] = dvxG_recv[ivx];
			MPI_Recv(dvxG_recv, NVXLOC, MPI_DOUBLE, 0, 3*commsize+rank, MPI_COMM_WORLD, &status);
			for (int ivx = 0; ivx < NVXLOC; ivx++)
				dvx_G_real[ivx][1] = dvxG_recv[ivx];
			free_1d(dvxG_recv);
		}
	}
#endif
}


void compute_deriv_vy_fftw(fftw_complex *G_real, fftw_complex *dvy_G_real, PARAMETERS parameters){

	int nthreads;
#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif
#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif

	double pi = 4.*atan(1.);
	int MVY = parameters.nvy;
	int ivyabs;
	double mode;
	fftw_complex *G_spectral, *dvy_G_spectral;
	fftw_plan p_G_forward, p_dvyG_backward;

	double C = (double)MVY / (2.*parameters.L);

	// WARNING : with FFTW-MPI, the data is distributed in vx-direction and is global in vy-direction

	// Prepare FFTW plans
	G_spectral = fftw_malloc(sizeof(fftw_complex)*MVY);
	dvy_G_spectral = fftw_malloc(sizeof(fftw_complex)*MVY);
	for (int ivy = 0; ivy < MVY; ivy++){
		G_spectral[ivy][0] = 0.;
		G_spectral[ivy][1] = 0.;
		dvy_G_spectral[ivy][0] = 0.;
		dvy_G_spectral[ivy][1] = 0.;
		dvy_G_real[ivy][0] = 0.;
		dvy_G_real[ivy][1] = 0.;
	}
	p_G_forward = fftw_plan_dft_1d(MVY, G_real, G_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	p_dvyG_backward = fftw_plan_dft_1d(MVY, dvy_G_spectral, dvy_G_real, FFTW_BACKWARD, FFTW_ESTIMATE);

	// Compute G_hat
	fftw_execute(p_G_forward);

	// Compute dvyG_hat
	for (int ivy = 0; ivy < MVY; ivy++){
		if (ivy < MVY/2)
			ivyabs = ivy;
		else if (ivy > MVY/2)
			ivyabs = ivy-MVY;
		else if (ivy == MVY)
			ivyabs = 0;

		mode = 2*pi*((double)ivyabs)/((double)MVY);

		dvy_G_spectral[ivy][0] = -mode * G_spectral[ivy][1];
		dvy_G_spectral[ivy][1] = mode * G_spectral[ivy][0];
	}

	// Compute dvyG
	fftw_execute(p_dvyG_backward);
	for (int ivy = 0; ivy < MVY; ivy++){
		dvy_G_real[ivy][0] /= (double)MVY;
		dvy_G_real[ivy][1] /= (double)MVY;
	}

	for (int ivy = 0; ivy < MVY; ivy++){
		dvy_G_real[ivy][0] *= C;
		dvy_G_real[ivy][1] *= C;
	}

	fftw_destroy_plan(p_G_forward);
	fftw_destroy_plan(p_dvyG_backward);
	fftw_free(G_spectral);
	fftw_free(dvy_G_spectral);
}


void compute_deriv_vz_fftw(fftw_complex *G_real, fftw_complex *dvz_G_real, PARAMETERS parameters){

	int nthreads;
#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif
#ifdef key_fftw_openmp
	fftw_plan_with_nthreads(nthreads);
#endif

	double pi = 4.*atan(1.);
	int MVZ = parameters.nvz;
	int ivzabs;
	double mode;
	fftw_complex *G_spectral, *dvz_G_spectral;
	fftw_plan p_G_forward, p_dvzG_backward;

	double C = (double)MVZ / (2.*parameters.L);

	// WARNING : with FFTW-MPI, the data is distributed in vx-direction and is global in vz-direction

	// Prepare FFTW plans
	G_spectral = fftw_malloc(sizeof(fftw_complex)*MVZ);
	dvz_G_spectral = fftw_malloc(sizeof(fftw_complex)*MVZ);
	for (int ivz = 0; ivz < MVZ; ivz++){
		G_spectral[ivz][0] = 0.;
		G_spectral[ivz][1] = 0.;
		dvz_G_spectral[ivz][0] = 0.;
		dvz_G_spectral[ivz][1] = 0.;
		dvz_G_real[ivz][0] = 0.;
		dvz_G_real[ivz][1] = 0.;
	}
	p_G_forward = fftw_plan_dft_1d(MVZ, G_real, G_spectral, FFTW_FORWARD, FFTW_ESTIMATE);
	p_dvzG_backward = fftw_plan_dft_1d(MVZ, dvz_G_spectral, dvz_G_real, FFTW_BACKWARD, FFTW_ESTIMATE);

	// Compute G_hat
	fftw_execute(p_G_forward);

	// Compute dvzG_hat
	for (int ivz = 0; ivz < MVZ; ivz++){
		if (ivz < MVZ/2)
			ivzabs = ivz;
		else if (ivz > MVZ/2)
			ivzabs = ivz-MVZ;
		else if (ivz == MVZ)
			ivzabs = 0;

		mode = 2*pi*((double)ivzabs)/((double)MVZ);

		dvz_G_spectral[ivz][0] = -mode * G_spectral[ivz][1];
		dvz_G_spectral[ivz][1] = mode * G_spectral[ivz][0];
	}

	// Compute dvzG
	fftw_execute(p_dvzG_backward);
	for (int ivz = 0; ivz < MVZ; ivz++){
		dvz_G_real[ivz][0] /= (double)MVZ;
		dvz_G_real[ivz][1] /= (double)MVZ;
	}

	for (int ivz = 0; ivz < MVZ; ivz++){
		dvz_G_real[ivz][0] *= C;
		dvz_G_real[ivz][1] *= C;
	}

	fftw_destroy_plan(p_G_forward);
	fftw_destroy_plan(p_dvzG_backward);
	fftw_free(G_spectral);
	fftw_free(dvz_G_spectral);
}


void compute_divergence_fftw(double **Gv_loc, double *div_G_loc, PARAMETERS parameters){
	double pi = 4.*atan(1.);
	fftw_complex *G_real, *dG_real;
	
	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(1,parameters.nvz);
	
	int kind;
	
	for (int k = 0; k < NVXLOC*NVYLOC*NVZLOC; k++)
		div_G_loc[k] = 0.;
	
	G_real = fftw_malloc(sizeof(fftw_complex)*NVXLOC);
	dG_real = fftw_malloc(sizeof(fftw_complex)*NVXLOC);
	for (int kvz = 0; kvz < NVZLOC; kvz++){
		for (int kvy = 0; kvy < NVYLOC; kvy++){
			// For each node in (vy,vz)-direction, we compute dvx_Gvx(.,vyj,vzk)
			for (int kvx = 0; kvx < NVXLOC; kvx++){
				kind = kvz+NVZLOC*(kvy+NVYLOC*kvx);
				G_real[kvx][0] = Gv_loc[0][kind];
				G_real[kvx][1] = 0.;
				dG_real[kvx][0] = 0.;
				dG_real[kvx][1] = 0.;
			}
			compute_deriv_vx_fftw(G_real, dG_real, parameters);
			for (int kvx = 0; kvx < NVXLOC; kvx++){
				kind = kvz+NVZLOC*(kvy+NVYLOC*kvx);
				div_G_loc[kind] += dG_real[kvx][0];
			}
		}
	}
	fftw_free(G_real);
	fftw_free(dG_real);
	
	G_real = fftw_malloc(sizeof(fftw_complex)*NVYLOC);
	dG_real = fftw_malloc(sizeof(fftw_complex)*NVYLOC);
	for (int kvz = 0; kvz < NVZLOC; kvz++){
		for (int kvx = 0; kvx < NVXLOC; kvx++){
			// For each node in (vx,vz)-direction, we compute dvy_Gvy(vxi,.,vzk)
			for (int kvy = 0; kvy < NVYLOC; kvy++){
				kind = kvz+NVZLOC*(kvy+NVYLOC*kvx);
				G_real[kvy][0] = Gv_loc[1][kind];
				G_real[kvy][1] = 0.;
				dG_real[kvy][0] = 0.;
				dG_real[kvy][1] = 0.;
			}
			compute_deriv_vy_fftw(G_real, dG_real, parameters);
			for (int kvy = 0; kvy < NVYLOC; kvy++){
				kind = kvz+NVZLOC*(kvy+NVYLOC*kvx);
				div_G_loc[kind] += dG_real[kvy][0];
			}
		}
	}
	fftw_free(G_real);
	fftw_free(dG_real);
	
	if (parameters.ndim == DIM3){
		G_real = fftw_malloc(sizeof(fftw_complex)*NVZLOC);
		dG_real = fftw_malloc(sizeof(fftw_complex)*NVZLOC);
		for (int kvx = 0; kvx < NVXLOC; kvx++){
			for (int kvy = 0; kvy < NVYLOC; kvy++){
				// For each node in (vx,vy)-direction, we compute dvzG(vxi,vyj,.)
				for (int kvz = 0; kvz < NVZLOC; kvz++){
					kind = kvz+NVZLOC*(kvy+NVYLOC*kvx);
					G_real[kvz][0] = Gv_loc[2][kind];
					G_real[kvz][1] = 0.;
					dG_real[kvz][0] = 0.;
					dG_real[kvz][1] = 0.;
				}
				compute_deriv_vz_fftw(G_real, dG_real, parameters);
				for (int kvz = 0; kvz < NVZLOC; kvz++){
					kind = kvz+NVZLOC*(kvy+NVYLOC*kvx);
					div_G_loc[kind] += dG_real[kvz][0];
				}
			}
		}
		fftw_free(G_real);
		fftw_free(dG_real);
	}
}


#endif
