#include "./src/core/fugacity.h"

void dgeev_(char *jobvl, char *jobvr, int *n, double *M, int *lda, double *wr, double *wi, double *vl, int *ldvl, double *vr, int *ldvr, double *work, int *lwork, int *info);



void find_legendre_roots(int n, double* roots, double* weights){
	int LDA, LDVL, LDVR, info, lwork, N, istatus;
	double *A, *b, *work, wkopt, *wr, *wi, *vr, *vl, *Pnm1, *Pn, *Pnp1, *temp;
	
	if (n < 1)
		error("Cannot run int_Legendre_1d for integers < 1.");

	N = n;
	LDA = N;
	LDVL = N;
	LDVR = N;
	A = allocation_1d(N*LDA);
	b = allocation_1d(N-1);
	for (int i = 1; i <= N-1; i++)
		b[i-1] = sqrt(i*i/(4.*i*i-1.));
	for (int i = 0; i < N*N; i++)
		A[i] = 0.;
	for (int i = 1; i <= N-1; i++)
		A[(i)+(i-1)*N] = b[i-1];
	for (int j = 1; j <= N-1; j++)
		A[(j-1)+(j)*N] = b[j-1];

	wr = allocation_1d(N); // right eigenvalues
	wi = allocation_1d(N); // left eigenvalues
	vr = allocation_1d(LDVR*N); // right eigenvectors - not computed
	vl = allocation_1d(LDVL*N); // left eigenvectors - not computed
	lwork = 4*N;
	work = allocation_1d(lwork);
	// Solve eigenproblem
	dgeev_("N", "N", &N, A, &LDA, wr, wi, vl, &LDVR, vr, &LDVR,	work, &lwork, &info);
	
	for (int i = 0; i < n; i++)
		roots[i] = wr[i];
	
	Pnm1 = allocation_1d(n);
	Pn = allocation_1d(n);
	Pnp1 = allocation_1d(n);
	temp = allocation_1d(n);
	for (int i = 0; i < n; i++){
		Pnm1[i] = 1.;
		Pn[i] = roots[i];
	}
	
	for (int k = 1; k <= n; k++){
		for (int i = 0; i < n; i++){
			Pnp1[i] = ( (2.*k+1.)*roots[i]*Pn[i]-k*Pnm1[i] )/(k+1.);
			temp[i] = Pnm1[i];
			Pnm1[i] = Pn[i];
			Pn[i] = Pnp1[i];
		}
	}
	for (int i = 0; i < n; i++)
		weights[i] = 2.*(1-roots[i]*roots[i])/(n*n*temp[i]*temp[i]);
	
	//for (int i = 0; i < n; i++)
		//printf("n = %d, i = %d, xi = %lf, wi = %lf\n", n, i, roots[i], weights[i]);
	
	free_1d(wr);
	free_1d(wi);
	free_1d(vr);
	free_1d(vl);
	free_1d(A);
	free_1d(b);
	free_1d(work);
	free_1d(Pnm1);
	free_1d(Pn);
	free_1d(Pnp1);
	free_1d(temp);
}



void read_legendre_roots(int n, double* roots, double* weights){
	char namefile[MAXSTRINGLENGTH];
	char buf[MAXSTRINGLENGTH];
	int istatus;
	char *sstatus;
	char *path, *file;
	FILE *f;
	int ndigits;

	int rank, commsize, init;
#ifdef key_mpi
	MPI_Initialized(&init);
	if (init){
		MPI_Comm_rank(MPI_COMM_WORLD, &rank);
		MPI_Comm_size(MPI_COMM_WORLD, &commsize);
	}
	else{
		rank = 0;
		commsize = 1;
	}
#else
	rank = 0;
	commsize = 1;
#endif

	strcpy(namefile, __FILE__);
	split_path_file(&path, &file, namefile);
	free_1d_char(file);
	
	ndigits = log10(ceil(n))+1;
	file = allocation_1d_char(8+1+ndigits+4+1);
	sprintf(file, "Legendre_%d.dat", n);
	sprintf(namefile, "%s/Legendre_roots/%s", path, file);
	
	if (rank == 0){
		if ((f = fopen(namefile, "r"))){
			// The file exists and in opened --> read it
			for (int i = 0; i < n; i++){
				sstatus = fgets(buf, sizeof(buf), f);
				istatus = sscanf(buf, "%lf %lf", &(roots[i]), &(weights[i]));
			}
			fclose(f);
		}
		else{
			// The file does not exist --> create it and fill it
			find_legendre_roots(n, roots, weights);
			f = fopen(namefile, "w");
			for (int i = 0; i < n; i++)
				fprintf(f, "%.15e %.15e\n", roots[i], weights[i]);
			fclose(f);
		}
	}
	
	free_1d_char(file);
	free_1d_char(path);
	
#ifdef key_mpi
	if (init){
		MPI_Bcast(roots, n, MPI_DOUBLE, 0, MPI_COMM_WORLD);
		MPI_Bcast(weights, n, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	}
#endif
}




double compute_Qnuz(double nu, double z, int type_qparticles){
	double *roots;
	double *weights;
	int n = 2;
	double pi = 4.*atan(1.);
	int foundnan;
	double valn, valnp1, fi, sign, gammanu;


	if (fabs(nu-0.5) < THRESHOLD)
		gammanu = sqrt(pi);
	else if (fabs(nu-1.5) < THRESHOLD)
		gammanu = 0.5*sqrt(pi);
	else if (fabs(nu-2.5) < THRESHOLD)
		gammanu = 0.75*sqrt(pi);
	else if (fabs(nu-1.) < THRESHOLD)
		gammanu = 1.;
	else if (fabs(nu-2.) < THRESHOLD)
		gammanu = 1.;
	else
		error("Qnuz function is only designed for nu = 0.5, 1, 1.5, 2, or 2.5");
	
	switch (type_qparticles){
		case BOSONS :
			sign = -1.;
			break;
		case FERMIONS :
			sign = 1.;
			break;
		default :
			error("Unknown kind of quantum particles.");
	}

	roots = allocation_1d(n);
	weights = allocation_1d(n);
	read_legendre_roots(n, roots, weights);
	valn = 0.;
	for (int i=0; i<n; i++){
		fi = 0.5*(roots[i]+1.);
		valn += 0.5*weights[i]*(pow(fi,nu-1.)/(exp(fi)+z*sign) + pow(fi,-1.-nu)/(exp(1./fi)+z*sign));
		//valn += weights[i] * pow(roots[i],nu-1.) / (1.+z*sign*exp(-roots[i]));
	}
	valn *= z;
	free_1d(roots);
	free_1d(weights);
	do{
		n++;
		roots = allocation_1d(n);
		weights = allocation_1d(n);
		read_legendre_roots(n, roots, weights);
		valnp1 = 0.;
		foundnan = (1 == 0);
		for (int i=0; i<n; i++){
			foundnan = (foundnan || (isnan(fabs(weights[i])) != 0));
			foundnan = (foundnan || (isnan(fabs(roots[i])) != 0));
			//printf("n = %d, i = %d, w[i] = %g, x[i] = %g | foundnan = %d\n", n, i, weights[i], roots[i], foundnan);
			//valnp1 += weights[i] * pow(roots[i],nu-1.) / (1.+z*sign*exp(-roots[i]));
			fi = 0.5*(roots[i]+1.);
			valnp1 += 0.5*weights[i]*(pow(fi,nu-1.)/(exp(fi)+z*sign) + pow(fi,-1.-nu)/(exp(1./fi)+z*sign));
		}
		if (foundnan)
			break;
		valnp1 *= z;
		//if (fabs(valn-valnp1)/fabs(valn) <= THRESHOLD)
		if (fabs(valn-valnp1)/fabs(valn) <= THRESHOLD)
			break;
		else
			valn = valnp1;
		//printf("n = %d, valn = %.15e\n", n, valn);
		free_1d(roots);
		free_1d(weights);
	} while (n <= 1000);

	// Clean Legendre roots and weights
	if (n <= 1000){
		free_1d(roots);
		free_1d(weights);
	}

	return valn/gammanu;
}


void find_fugacity_temperature_criticalmass(double rho, double *eint, double *T, double *z, double *m0, double hbar, int d, int type_qparticles){
	char buf[MAXSTRINGLENGTH];
	double alpha;
	double zeta_3o2;
	double zeta_5o2;
	double pi = 4.*atan(1.);
	int condensation;
	double threshold_condensation;
	
	double zmin, zmax, zmid, zshift;
	double Qdo2_zmin, Qdo2_zmax, Qdo2_zmid, Qdo2_zshift;
	double Qdo2p1_zmin, Qdo2p1_zmax, Qdo2p1_zmid, Qdo2p1_zshift;
	double test_zmin, test_zmax, test_zmid;

	double inpow;
	double r;
	int n;
	
	switch (type_qparticles){
		case FERMIONS:
			alpha = pow(hbar, d);
			break;
		case BOSONS:
			alpha = -pow(hbar, d);
			break;
		default:
			error("Unknown kind of quantum particles.");
	}
	
	if ((d == 3) && (type_qparticles == BOSONS)){
		// We verify if rho and eint (or T) can induce condensation
		zeta_3o2 = compute_Qnuz(1.5, 1., type_qparticles);
		zeta_5o2 = compute_Qnuz(2.5, 1., type_qparticles);
		if (*eint > 0.){
			// eint has been provided
			threshold_condensation = pow(zeta_3o2, 2.5) * pow(4.*pi/(3.*zeta_5o2), 1.5) / fabs(alpha);
			condensation = (rho > threshold_condensation * pow(*eint, 1.5));
		}
		else{
			// T has been provided instead of eint
			condensation = (rho - pow(2.*pi*(*T), 1.5)/fabs(alpha)*zeta_3o2 > 0.);
		}
	}
	else
		condensation = (1==0);
	
	
	if (condensation){
		if (*eint > 0)
			*T = (*eint)*(2./3.) * zeta_3o2/zeta_5o2;
		else
			*eint = 3.*(*T)/2. * zeta_5o2/zeta_3o2;
		*m0 = rho - pow(2.*pi*(*T), 1.5)*zeta_3o2/fabs(alpha);
		*z = 1.;
	}
	else{
		*m0 = -1.;
		
		if (*eint > 0){
			// First step : z is solution of
			// rho = f(z) = 1/|alpha| * (4*pi*eint / (d*Q_(d/2+1)(z)))^(d/2) * (Q_(d/2)(z))^(d/2+1)
			// or
			// (Q_(d/2)(z))^(1+2/d) * e - d/2 * (rho*|alpha|)^(d/2) * Q_(1+d/2)(z) = 0
	
			//
			// Second step : T is deduced with 
			// T = (2*eint*Q_(d/2)(z))/(d*Q_(d/2+1)(z))

			// First step
			switch (type_qparticles){
				case BOSONS :
					n = 1;
					do{
						zmin = 1./pow(2.,n);
						n++;
						Qdo2_zmin = compute_Qnuz(0.5*d, zmin, type_qparticles);
						Qdo2p1_zmin = compute_Qnuz(0.5*d+1., zmin, type_qparticles);
						inpow = (4*pi*(*eint)/d) * (Qdo2_zmin/Qdo2p1_zmin);
						test_zmin = (1./fabs(alpha)) * pow(inpow, 0.5*d) * Qdo2_zmin - rho;
						//printf("zmin = %e\n", zmin);
					} while ((test_zmin > 0.) && (zmin > (THRESHOLD)));
			
					n = 1;
					do{
						zmax = 1.-1./pow(2.,n);
						n++;
						Qdo2_zmax = compute_Qnuz(0.5*d, zmax, type_qparticles);
						Qdo2p1_zmax = compute_Qnuz(0.5*d+1., zmax, type_qparticles);
						inpow = (4*pi*(*eint)/d) * (Qdo2_zmax/Qdo2p1_zmax);
						test_zmax = (1./fabs(alpha)) * pow(inpow, 0.5*d) * Qdo2_zmax - rho;
						//printf("zmax = %e\n", zmax);
					} while ((test_zmax < 0.) && (zmax < 1.-(THRESHOLD)));
					break;
				case FERMIONS :
					zmin = THRESHOLD;
					zmax = 1.-THRESHOLD;
					break;
				default :
					error("Unknown kind of quantum particles.");
			}
	
			Qdo2_zmin = compute_Qnuz(0.5*d, zmin, type_qparticles);
			Qdo2p1_zmin = compute_Qnuz(0.5*d+1., zmin, type_qparticles);
	
			inpow = (4*pi*(*eint)/d) * (Qdo2_zmin/Qdo2p1_zmin);
			test_zmin = (1./fabs(alpha)) * pow(inpow, 0.5*d) * Qdo2_zmin - rho;
	
			Qdo2_zmax = compute_Qnuz(0.5*d, zmax, type_qparticles);
			Qdo2p1_zmax = compute_Qnuz(0.5*d+1., zmax, type_qparticles);
	
			inpow = (4*pi*(*eint)/d) * (Qdo2_zmax/Qdo2p1_zmax);
			test_zmax = (1./fabs(alpha)) * pow(inpow, 0.5*d) * Qdo2_zmax - rho;

			//printf("zmin = %e, zmax = %e, test_zmin = %e, test_zmax = %e\n", zmin, zmax, test_zmin, test_zmax);
	
			switch (type_qparticles){
				case BOSONS :
					if (test_zmin*test_zmax > 0.){
						// Problem : the root of z -> f(z)-rho is expected to be in ]0,1[ !
						error("Something is wrong with the computation of fugacity for bosons.");
					}
					break;
				case FERMIONS :
					while (test_zmin*test_zmax > 0.){
						// the sign of z -> f(z)-rho does not changes in ]0,1[
						// --> replace ]0,1[ by ]0,n[ and try again
						zmax++;
						Qdo2_zmax = compute_Qnuz(0.5*d, zmax, type_qparticles);
						Qdo2p1_zmax = compute_Qnuz(0.5*d+1., zmax, type_qparticles);
						inpow = (4*pi*(*eint)/d) * (Qdo2_zmax/Qdo2p1_zmax);
						test_zmax = (1./fabs(alpha)) * pow(inpow, 0.5*d) * Qdo2_zmax - rho;
						//printf("zmin = %e, zmax = %e, test_zmin = %e, test_zmax = %e\n", zmin, zmax, test_zmin, test_zmax);
					}
					break;
				default :
					error("Unknown kind of quantum particles.");
			}

			// the sign of z -> f(z)-rho changes in ]0,n[, with f(0+)-rho < 0 and f(n-)-rho > 0
			// Identify the root of this function in ]0,n[ by dichotomy
			while (zmax-zmin > THRESHOLD){
				zmid = 0.5*(zmin+zmax);
				Qdo2_zmid = compute_Qnuz(0.5*d, zmid, type_qparticles);
				Qdo2p1_zmid = compute_Qnuz(0.5*d+1., zmid, type_qparticles);
				test_zmid = (1./fabs(alpha)) * pow((4.*pi*(*eint))/(d*Qdo2p1_zmid), 0.5*d) * pow(Qdo2_zmid, 0.5*d+1.) - rho;
				//printf("zmin = %e, zmax = %e, test_zmid = %e\n", zmin, zmax, test_zmid);
				if (test_zmid*test_zmax > 0.){
					// The sign of f(z)-rho did not change on [zmid,zmax[
					// The root is located in ]zmin,zmid[
					zmax = zmid;
					test_zmax = test_zmid;
				}
				else{
					// The sign of f(z)-rho changed on [zmid,zmax[
					// The root is located in ]zmid,zmax[
					zmin = zmid;
					test_zmin = test_zmid;
				}
			}

			*z = zmid;

			// Second step
			*T = (2.*(*eint)*Qdo2_zmid)/(d*Qdo2p1_zmid);
		}
		else{
			// The goal is to find z such that Q_(d/2)(z) = r with 
			// - Q_(d/2) continuous on [0,+infty[ and strictly increasing on ]0,+infty[
			// - r = mass * hbar^d / (2*pi*T)^(d/2)
			// -----> A unique solution is expected
			
			r = rho*fabs(alpha)/pow(2.*pi*(*T), 0.5*d);
	
			// Find zmax such that Q_(d/2)(zshift) > r and Q_(d/2)(zshift-1) <= r
			switch (type_qparticles){
				case FERMIONS:
					// z is expected to lie in ]0,+infty[, however we should have found zmax such that z in [zmin,zmax]
					zshift = THRESHOLD;
					Qdo2_zshift = compute_Qnuz(0.5*d, zshift, type_qparticles);
					while ((Qdo2_zshift <= r) && (zmax <= 1e8)){
						zshift += 1.;
						Qdo2_zshift = compute_Qnuz(0.5*d, zshift, type_qparticles);
					}
					break;
				case BOSONS:
					zshift = 1.;
					// Note that if r > Q_(d/2)(1) for the 3D case, the computation of fugacity cannot work because
					// we have reached the condensation threshold --> Another model is required
					Qdo2_zshift = compute_Qnuz(0.5*d, zshift, type_qparticles);
					if ((d == 3) && (r > Qdo2_zshift)){
						message("The critical Bose-Einstein condensation temperature is reached and the current computations cannot provide a fugacity. We need\n");
						sprintf(buf, "    rho*hbar^3 / (2*pi*T)^(3/2)    in    ]0,zeta(3/2)]     with zeta(3/2) = %e\n", Qdo2_zshift);
						message(buf);
						sprintf(buf, "I have rho*hbar^3 / (2*pi*T)^(3/2) = %e here.\nSo you should consider a lower mass or a larger temperature.", r);
						error(buf);
					}
					break;
				default:
					error("Unknown kind of quantum particles.");
			}
			//printf("zshift = %e\n", zshift);
	
			// Find zmin = zshift-1+2^(-n) and zmax = zshift-2^(-n) such that Q(zmin) < r and Q(zmax) > r
			n = 1;
			zmin = zshift-1. + 1./pow(2.,n);
			Qdo2_zmin = compute_Qnuz(0.5*d, zmin, type_qparticles);
			while ((Qdo2_zmin > r) && (n < 1e8)){
				n++;
				zmin = zshift-1. + 1./pow(2.,n);
				Qdo2_zmin = compute_Qnuz(0.5*d, zmin, type_qparticles);
				//printf("zmin = %e, Q(zmin) = %e, r = %e, n = %d\n", zmin, Qdo2_zmin, r, n);
			}
			//printf("zmin = %e\n", zmin);
			n = 1;
			zmax = zshift - 1./pow(2.,n);
			Qdo2_zmax = compute_Qnuz(0.5*d, zmax, type_qparticles);
			while ((Qdo2_zmax < r) && (n < 1e8)){
				n++;
				zmax = zshift - 1./pow(2.,n);
				Qdo2_zmax = compute_Qnuz(0.5*d, zmax, type_qparticles);
				//printf("zmax = %e, Q(zmax) = %e, r = %e, n = %d\n", zmax, Qdo2_zmax, r, n);
			}
			//printf("[zmin,zmax] = [%e,%e]\n", zmin, zmax);

			// Use dichotomy to find z such that Q_(d/2)(z) = r
			Qdo2_zmax = compute_Qnuz(0.5*d, zmax, type_qparticles);
			Qdo2_zmin = compute_Qnuz(0.5*d, zmin, type_qparticles);
			//printf("Q(zmin) = %e, Q(zmax) = %e | r = %e\n", Qdo2_zmin, Qdo2_zmax, r);
			n = 0;
			while ((zmax-zmin > THRESHOLD) && (n < 1e8)){
				zmid = 0.5*(zmin+zmax);
				Qdo2_zmid = compute_Qnuz(0.5*d, zmid, type_qparticles);
				//printf("Q(zmin) = %e, Q(zmid) = %e, Q(zmax) = %e | r = %e\n", Qdo2zmin, Qdo2zmid, Qdo2zmax, r);
				if (Qdo2_zmid < r){
					// The solution of Q_(d/2)(z) = r lies in [zmid,zmax]
					zmin = zmid;
					Qdo2_zmin = Qdo2_zmid;
				}
				else{
					// The solution of Q_(d/2)(z) = r lies in [zmin,zmid]
					zmax = zmid;
					Qdo2_zmax = Qdo2_zmid;
				}
				n++;
				//printf("zmid = %e\n", zmid);
			}
	
			// Compute the internal energy: e = d/2 * T * Q_(d/2+1)(z) / Q_(d/2)(z)
			Qdo2p1_zmax = compute_Qnuz(0.5*d+1., zmax, type_qparticles);
			*eint = 0.5*d*(*T) * Qdo2p1_zmax / Qdo2_zmax;
			
			*z = zmax;
		
		}
	}
}



// void compute_F_zero(double *F, double *X, double rho, double alpha, double eint, int d, int type_qparticles){
// 	double pi = 4.*atan(1.);
// 	double Qdo2 = compute_Qnuz(0.5*d, X[0], type_qparticles);
// 	double Qdo2p1 = compute_Qnuz(0.5*d+1., X[0], type_qparticles);

// 	double r = fabs(alpha) * rho * pow(d/(4.*pi*eint), 0.5*d);
// }


double compute_A_zero(double mu, double nu, int type_qparticles){
	double z;
	double Qnu, Qnup1;

	if (type_qparticles == BOSONS)
		z = mu;
	else if (type_qparticles == FERMIONS)
		z = exp(mu);
	Qnu = compute_Qnuz(nu, z, type_qparticles);
	Qnup1 = compute_Qnuz(nu+1., z, type_qparticles);

	return pow(Qnu, nu+1.)/pow(Qnup1, nu);
}


int test_condensation(double rho, double eint, double T, double alpha, int d, int type_qparticles){
	double threshold_condensation, testing_condensation;
	double zeta_3o2, zeta_5o2;
	double pi = 4.*atan(1.);

	int condensation;

	switch (type_qparticles){
		case FERMIONS:
			if (eint > 0.){
				// The internal energy has been provided: finding a temperature is not insured
				if (d == 2)
					threshold_condensation = 2.;
				else if (d == 3)
					threshold_condensation = 5./3. * sqrt(10./pi);
				else
					error("Unknown dimensionality.");
				testing_condensation = fabs(alpha) * rho * pow(d/(4.*pi*eint), 0.5*d);
				condensation = (testing_condensation+THRESHOLD >= threshold_condensation);
			}
			else if (T > 0.){
				// The temperature is already provided. Condensation cannot occur.
				//printf("\nNo condensation test\n");
				condensation = (1==0);
			}
			else
				error("Input internal energy and temperature are both negative.");

			break;
		case BOSONS:
			if (d == 2){
				// Condensation can occur here
				condensation = (1==0);
			}
			if (d == 3){
				zeta_3o2 = compute_Qnuz(1.5, 1., type_qparticles);
				zeta_5o2 = compute_Qnuz(2.5, 1., type_qparticles);
				if (eint > 0.){
					// The internal energy has been provided. Condensation can occur.
					threshold_condensation = pow(zeta_3o2, 2.5) / pow(zeta_5o2, 1.5);
					testing_condensation = fabs(alpha) * rho * pow(d/(4.*pi*eint), 0.5*d);
					//printf("\n");
					//printf("|alpha| * rho * (d/(4*pi*e))^(d/2) = %e | threshold = %e\n", testing_condensation, threshold_condensation);
					condensation = (testing_condensation+THRESHOLD >= threshold_condensation);
				}
				else if (T > 0.){
					// The temperature has been provided. Condensation cannot occur.
					threshold_condensation = zeta_3o2;
					testing_condensation = rho * fabs(alpha) / pow(2.*pi*T, 1.5);
					//printf("\n");
					//printf("|alpha| * rho / (2*pi*T)^(3/2) = %e | threshold = %e\n", testing_condensation, threshold_condensation);
			 		condensation = (testing_condensation+THRESHOLD >= threshold_condensation);
				}
				else
					error("Input internal energy and temperature are both negative.");
			}
			break;
		default:
			error("Unknown type of particles.");
	}
	return condensation;
}


void find_fugacity_temperature_criticalmass_fixedpoint(double rho, double *eint, double *T, double *z, double *m0, double hbar, int d, int type_qparticles){
	double alpha;
	double zeta_3o2, zeta_5o2;
	double testing_condensation, threshold_condensation;
	double ak, akp1, bkm1, bk, bkp1;
	double Kak, Kakp1, Kbkm1, Kbk, Kbkp1;
	double m, s, r;
	double nu;
	int k;
	int condensation;
	char buf[MAXSTRINGLENGTH];

	double pi = 4.*atan(1.);
	
	switch (type_qparticles){
		case FERMIONS:
			alpha = pow(hbar, d);
			break;
		case BOSONS:
			alpha = -pow(hbar, d);
			break;
		default:
			error("Unknown kind of quantum particles.");
	}
	nu = 0.5*d;

	condensation = test_condensation(rho, *eint, *T, alpha, d, type_qparticles);

	if (!condensation){
		//printf("No condensation\n");
		if (*eint > 0.){
			// The internal energy is given. We solve the following problem
			//     0 = K(z) = A(z) - r
			// with r = |alpha| * rho * (nu/(2*pi*eint))^nu
			// with A(z) = Q_nu(z)^(nu+1) / Q_(nu+1)(z)^nu being strictly increasing
			//
			// For bosons, z is in ]0,1[
			// For fermions, z = e^mu with mu in ]-inf,+inf[

			//printf("Use solver with e > 0\n");
			r = fabs(alpha) * rho * pow(nu/(2*pi*(*eint)), nu);

			// We use Brent's method
			// Find a_0, b_0 such that K(a_0) * K(b_0) < 0
			if (type_qparticles == FERMIONS){
				// We work with z = e^mu
				//ak = 1e-8;
				//bk = 1.-1e-8;
				ak = -20.;
				bk = -19.;
				Kak = compute_A_zero(ak, nu, type_qparticles) - r;
				Kbk = compute_A_zero(bk, nu, type_qparticles) - r;
				k = 0;
				while ((k < 1e6) && (Kak*Kbk > 0.)){
					ak += 1.;
					bk += 1.;
					Kak = compute_A_zero(ak, nu, type_qparticles) - r;
					Kbk = compute_A_zero(bk, nu, type_qparticles) - r;
					k++;
				}
			}
			else{
				ak = 1e-8;
				bk = 1.-1e-8;
				Kak = compute_A_zero(ak, nu, type_qparticles) - r;
				Kbk = compute_A_zero(bk, nu, type_qparticles) - r;
				if (Kak*Kbk > 0.){
					sprintf(buf, "rho = %e\neint = %e\nhbar = %e\nr = %e\n", rho, *eint, hbar, r);
					message(buf);
					sprintf(buf, "Something is wrong with the computation of the fugacity for bosons. Either the input parameters are wrong or the discretization did not succeed in catching the correct profile.");
					error(buf);
				}
			}
			if (fabs(Kbk) < fabs(Kak)){
				// We do not need to swap a_0 and b_0
				// We set b_(-1) as a_0
				bkm1 = ak;
				Kbkm1 = Kak;
			}
			else{
				// We must swap a_0 and b_0, then set b_(-1) to a_0
				// Equivalently, we set b_(-1) to old b_0, then new b_0 to old a_0, and finally new a_0 to old b_0 (i.e. b_(-1))
				bkm1 = bk;
				Kbkm1 = Kbk;
				bk = ak;
				Kbk = Kak;
				ak = bkm1;
				Kak = Kbkm1;
			}
			
			k = 0;
			while ((k < 1e6) && (fabs(ak-bk) > THRESHOLD)){
				// Compute middle point of a_k and b_k
				m = 0.5*(ak+bk);

				// Compute the secant of z -> K(z) between b_(k-1) and b_k
				if (fabs(Kbk-Kbkm1) > THRESHOLD)
					s = bk - (bk-bkm1)/(Kbk-Kbkm1) * Kbk;
				else
					s = m;
				
				// Compute b_(k+1)
				if ((bk <= s) && (s <= m))
					bkp1 = s;
				else if ((m <= s) && (s <= bk))
					bkp1 = s;
				else
					bkp1 = m;
				
				// Compute a_(k+1)
				Kbkp1 = compute_A_zero(bkp1, nu, type_qparticles) - r;
				if (Kak*Kbkp1 < 0.){
					akp1 = ak;
					Kakp1 = Kak;
				}
				else{
					akp1 = bk;
					Kakp1 = Kbk;
				}

				if (fabs(Kakp1) < fabs(Kbkp1)){
					bkm1 = bk;
					Kbkm1 = Kbk;
					// Prepare the following iteration with a swap of a_(k+1) and b_(k+1)
					ak = bkp1;
					Kak = Kbkp1;
					bk = akp1;
					Kbk = Kakp1;
				}
				else{
					bkm1 = bk;
					Kbkm1 = Kbk;
					// Prepare the following iteration
					ak = akp1;
					Kak = Kakp1;
					bk = bkp1;
					Kbk = Kbkp1;
				}
				k++;
			}

			// We found the fugacity
			if (type_qparticles == BOSONS)
				*z = m;
			else
				*z = exp(m);
			// We deduce the temperature
			*T = (*eint)*2.*compute_Qnuz(nu, *z, type_qparticles)/(d*compute_Qnuz(nu+1., *z, type_qparticles));
		}
		else if (*T > 0.){
			// The temperature is provided. We solve the following problem
			//     0 = K(z) = Q_nu(z) - r
			// with r = rho * |alpha| / (2*pi*T)^nu
			// and Q_nu being strictly increasing

			//printf("Use solver with T > 0\n");

			r = fabs(alpha) * rho / pow(2.*pi*(*T), nu);
			//printf("r = %e\n", r);

			// We use Brent's method
			// Find a_0, b_0 such that K(a_0) * K(b_0) < 0
			
			if (type_qparticles == FERMIONS){
				ak = -20.;
				bk = -19.;
				Kak = compute_Qnuz(nu, exp(ak), type_qparticles) - r;
				Kbk = compute_Qnuz(nu, exp(bk), type_qparticles) - r;
				k = 0;
				while ((k < 1e6) && (Kak*Kbk > 0.)){
					ak += 1.;
					bk += 1.;
					Kak = compute_Qnuz(nu, exp(ak), type_qparticles) - r;
					Kbk = compute_Qnuz(nu, exp(bk), type_qparticles) - r;
					k++;
					//printf("ak = %e, bk = %e | K(exp(ak)) = %e, K(exp(bk)) = %e\n", ak, bk, Kak, Kbk);
				}
			}
			else{
				ak = 1e-8;
				bk = 1.-1e-8;
				Kak = compute_Qnuz(nu, ak, type_qparticles) - r;
				Kbk = compute_Qnuz(nu, bk, type_qparticles) - r;
				if (Kak*Kbk > 0.){
					sprintf(buf, "rho = %e\nT = %e\nhbar = %e\nr = %e\n", rho, *T, hbar, r);
					message(buf);
					sprintf(buf, "Something is wrong with the computation of the fugacity for bosons. Either the input parameters are wrong or the discretization did not succeed in catching the correct profile.");
					error(buf);
				}
			}
			if (fabs(Kbk) < fabs(Kak)){
				// We do not need to swap a_0 and b_0
				// We set b_(-1) as a_0
				bkm1 = ak;
				Kbkm1 = Kak;
			}
			else{
				// We must swap a_0 and b_0, then set b_(-1) to a_0
				// Equivalently, we set b_(-1) to old b_0, then new b_0 to old a_0, and finally new a_0 to old b_0 (i.e. b_(-1))
				bkm1 = bk;
				Kbkm1 = Kbk;
				bk = ak;
				Kbk = Kak;
				ak = bkm1;
				Kak = Kbkm1;
			}
			
			k = 0;
			while ((k < 1e6) && (fabs(ak-bk) > THRESHOLD)){
				// Compute middle point of a_k and b_k
				m = 0.5*(ak+bk);

				// Compute the secant of z -> K(z) between b_(k-1) and b_k
				if (fabs(Kbk-Kbkm1) > THRESHOLD)
					s = bk - (bk-bkm1)/(Kbk-Kbkm1) * Kbk;
				else
					s = m;
				
				// Compute b_(k+1)
				if ((bk <= s) && (s <= m))
					bkp1 = s;
				else if ((m <= s) && (s <= bk))
					bkp1 = s;
				else
					bkp1 = m;
				
				// Compute a_(k+1)
				if (type_qparticles == FERMIONS)
					Kbkp1 = compute_Qnuz(nu, exp(bkp1), type_qparticles) - r;
				else
					Kbkp1 = compute_Qnuz(nu, bkp1, type_qparticles) - r;
				if (Kak*Kbkp1 < 0.){
					akp1 = ak;
					Kakp1 = Kak;
				}
				else{
					akp1 = bk;
					Kakp1 = Kbk;
				}

				if (fabs(Kakp1) < fabs(Kbkp1)){
					bkm1 = bk;
					Kbkm1 = Kbk;
					// Prepare the following iteration with a swap of a_(k+1) and b_(k+1)
					ak = bkp1;
					Kak = Kbkp1;
					bk = akp1;
					Kbk = Kakp1;
				}
				else{
					bkm1 = bk;
					Kbkm1 = Kbk;
					// Prepare the following iteration
					ak = akp1;
					Kak = Kakp1;
					bk = bkp1;
					Kbk = Kbkp1;
				}
				k++;
			}

			// We found the fugacity
			if (type_qparticles == FERMIONS)
				*z = exp(m);
			else
				*z = m;
			// We deduce the internal energy
			*eint = 0.5*d*(*T) * compute_Qnuz(nu+1., *z, type_qparticles) / compute_Qnuz(nu, *z, type_qparticles);
		}
		else{
			error("Fugacity solver is not well prepared.");
		}

		// No critical mass
		*m0 = -1.;
	}
	else{
		//printf("Condensation\n");
		// Condensation or quantum degeneracy has been detected.
		// --> The quantum maxwellian limit model is no more valid
		switch (type_qparticles){
			case BOSONS:
				// The limit model is the quantum Dirac-Maxwell model
				zeta_3o2 = compute_Qnuz(1.5, 1., type_qparticles);
				zeta_5o2 = compute_Qnuz(2.5, 1., type_qparticles);
				if (*eint > 0.){
					// The internal energy is provided.
					// Compute the corresponding critical temperature
					*T = 2.*(*eint) * zeta_3o2 / (3.*zeta_5o2);
					// Compute the corresponding critical mass
					*m0 = rho - pow(2.*pi*(*T), 1.5)*zeta_3o2/fabs(alpha);
				}
				else if (*T > 0.){
					// The temperature is provided.
					// Compute the corresponding critical internal energy
					*eint = 3.*(*T) * zeta_5o2 / (2.*zeta_3o2);
					// Compute the corresponding critical mass
					*m0 = rho - pow(2.*pi*(*T), 1.5)*zeta_3o2/fabs(alpha);
				}
				else{
					error("Fugacity solver is not well prepared.");
				}
				*z = 1.;
				break;
			case FERMIONS:
				*m0 = -1;
				*z = -1.;
				break;
			default:
				error("Unknown type of quantum particles.");
		}
	}


	// if ((d == 3) && (type_qparticles == BOSONS)){
	// 	// We verify if rho and eint (or T) can induce condensation
	// 	zeta_3o2 = compute_Qnuz(1.5, 1., type_qparticles);
	// 	zeta_5o2 = compute_Qnuz(2.5, 1., type_qparticles);
	// 	if (*eint > 0.){
	// 		// The internal energy has been provided
	// 		threshold_condensation = pow(zeta_3o2, 2.5) * pow(4.*pi/(3.*zeta_5o2), 1.5) / fabs(alpha);
	// 		condensation = (rho > threshold_condensation * pow(*eint, 1.5));
	// 	}
	// 	else{
	// 		// The temperature has been provided
	// 		condensation = (rho - pow(2.*pi*(*T), 1.5)/fabs(alpha)*zeta_3o2 > 0.);
	// 	}
	// }
	// else
	// 	condensation = (1==0);
	
	// if (condensation){
	// 	if (*eint > 0){
	// 		// The internal energy has been provided --> compute the corresponding temperature
	// 		*T = (*eint)*(2./3.) * zeta_3o2/zeta_5o2;
	// 	}
	// 	else{
	// 		// The temperature has been provided --> compute the corresponding internal energy
	// 		*eint = 3.*(*T)/2. * zeta_5o2/zeta_3o2;
	// 	}
	// 	// Compute the critical mass
	// 	*m0 = rho - pow(2.*pi*(*T), 1.5)*zeta_3o2/fabs(alpha);
	// 	// Set the fugacity to 1
	// 	*z = 1.;
	// 	printf("condensation !!\n");
	// }
	// else{
	// 	// There is no condensation
	// 	*m0 = -1.;
		
	// 	if (*eint > 0){
	// 		// The internal energy has been provided --> compute the corresponding temperature and fugacity
	// 		// To do this, we solve the following problem :
	// 		//   0 = K(z) = r - A(z)
	// 		// with r = rho / ( (4*pi*eint/d)^(d/2) * (1/|alpha|^d) )
	// 		// with A(z) = Q_(d/2)(z)^(d/2+1) / Q_(d/2+1)(z)^(d/2)
	// 		// with z in ]0,+inf[ for fermions and z in ]0,1[ for bosons
			
	// 		/*if (*z < 0.){
	// 			// The fugacity has never been computed before --> Provide an initial value
	// 			znm1 = 1.;
	// 			zn = 5.;
	// 		}

	// 		r = rho / (pow(4.*pi*(*eint)/d, 0.5*d) / fabs(alpha));
	// 		rstar = 1.;

	// 		printf("r = %e\n", r);

	// 		Qdo2z = compute_Qnuz(0.5*d, znm1, type_qparticles);
	// 		Qdo2p1z = compute_Qnuz(0.5*d+1., znm1, type_qparticles);
	// 		Anm1 = pow(Qdo2z, 0.5*d+1.)/pow(Qdo2p1z, 0.5*d);

	// 		Qdo2z = compute_Qnuz(0.5*d, zn, type_qparticles);
	// 		Qdo2p1z = compute_Qnuz(0.5*d+1., zn, type_qparticles);
	// 		An = pow(Qdo2z, 0.5*d+1.)/pow(Qdo2p1z, 0.5*d);

	// 		if (r > rstar)
	// 			Kznm1 = r/Anm1 - 1.;
	// 		else
	// 			Kznm1 = Anm1/r - 1.;

	// 		if (r > rstar)
	// 			Kzn = r/An - 1.;
	// 		else
	// 			Kzn = An/r - 1.;

	// 		znp1 = zn - (zn-znm1)/(Kzn-Kznm1) * Kzn;

	// 		Qdo2z = compute_Qnuz(0.5*d, znp1, type_qparticles);
	// 		Qdo2p1z = compute_Qnuz(0.5*d+1., znp1, type_qparticles);
	// 		Anp1 = pow(Qdo2z, 0.5*d+1.)/pow(Qdo2p1z, 0.5*d);
	// 		if (r > rstar)
	// 			Kznp1 = r/Anp1 - 1.;
	// 		else
	// 			Kznp1 = Anp1/r - 1.;

	// 		printf("Anm1 = %e, An = %e, Anp1 = %e\n", Anm1, An, Anp1);
			

	// 		n = 0;

	// 		while ((fabs(Kznp1) > THRESHOLD) && (n < 1e8)){
	// 			znm1 = zn;
	// 			Kznm1 = Kzn;
	// 			Anm1 = An;

	// 			zn = znp1;
	// 			Kzn = Kznp1;
	// 			An = Anp1;

	// 			znp1 = zn - (zn-znm1)/(Kzn-Kznm1) * Kzn;

	// 			Qdo2z = compute_Qnuz(0.5*d, znp1, type_qparticles);
	// 			Qdo2p1z = compute_Qnuz(0.5*d+1., znp1, type_qparticles);
	// 			Anp1 = pow(Qdo2z, 0.5*d+1.)/pow(Qdo2p1z, 0.5*d);
	// 			if (r > rstar)
	// 				Kznp1 = r/Anp1 - 1.;
	// 			else
	// 				Kznp1 = Anp1/r - 1.;
				
	// 			printf("Anm1 = %e, An = %e, Anp1 = %e\n", Anm1, An, Anp1);
	// 			n++;
	// 		}
	// 		*z = znp1;

	// 		*T = (2.*(*eint)*Qdo2z)/(d*Qdo2p1z);*/

	// 		double Xn[2];
	// 		double Xnp1[2];
	// 		double Fn[2];
	// 		double Fn_norm;
	// 		double Fnp1[2];
	// 		double deltan[2];
	// 		double deltan_norm_square;
	// 		double deltaFn[2];
	// 		double Bn[2][2];
	// 		double Bnp1[2][2];
	// 		double det_Bn;
	// 		double zinit, Tinit;

	// 		double p1 = pow(rho*fabs(alpha), 2./d) * d / (4.*pi*(*eint));
	// 		double p2 = pow(d/(4.*pi*(*eint)), d/(d+2.)) * pow(rho*fabs(alpha), 2./(d+2.));
	// 		double q1, q2;
			
	// 		if (*z > 0.)
	// 			zinit = *z;
	// 		else{
	// 			zinit = 1e-8;
	// 			q1 = pow(compute_Qnuz(0.5*d, zinit, type_qparticles), 2./d);
	// 			q2 = pow(compute_Qnuz(0.5*d+1., zinit, type_qparticles), 2./(2.+d));
	// 			printf("z = %.4e | Q_(d/2)(z)^(2/d) = %e, threshold = %e | Q_(d/2+1)(z)^(2/(d+2)) = %e, threshold = %e \n", zinit, q1, p1, q2, p2);
	// 			while ((q1 < p1) || (q2 < p2)){
	// 				zinit++;
	// 				q1 = pow(compute_Qnuz(0.5*d, zinit, type_qparticles), 2./d);
	// 				q2 = pow(compute_Qnuz(0.5*d+1., zinit, type_qparticles), 2./(2.+d));
	// 				printf("z = %.4e | Q_(d/2)(z)^(2/d) = %e, threshold = %e | Q_(d/2+1)(z)^(2/(d+2)) = %e, threshold = %e \n", zinit, q1, p1, q2, p2);
	// 			}
	// 		}
			
	// 		if (*T > 0.)
	// 			Tinit = *T;
	// 		else
	// 			Tinit = 0.5*(*eint)/d;

	// 		// Initialize Broyden method
	// 		Xn[0] = zinit;
	// 		Xn[1] = Tinit;
	// 		Bn[0][0] = 1.;
	// 		Bn[1][0] = 0.;
	// 		Bn[0][1] = 0.;
	// 		Bn[1][1] = 1.;

	// 		// Compute F(X_n)
	// 		compute_F_zero(Fn, Xn, rho, alpha, *eint, d, type_qparticles);
	// 		Fn_norm = sqrt(pow(Fn[0],2) + pow(Fn[1],2));

	// 		// Start loop
	// 		n = 0;
	// 		printf("n = %d | (z,T) = (%.6e,%.6e) | F(z,T) = (%.6e,%.6e)\n", n, Xn[0], Xn[1], Fn[0], Fn[1]);
	// 		while ((Fn_norm > THRESHOLD) && (n < 1e4)){
	// 			// Solve B_n * delta_n = -F(X_n)
	// 			det_Bn = Bn[0][0]*Bn[1][1] - Bn[1][0]*Bn[0][1];
	// 			deltan[0] = -(Bn[1][1]*Fn[0] - Bn[0][1]*Fn[1]) / det_Bn;
	// 			deltan[1] = -(-Bn[1][0]*Fn[0] + Bn[0][0]*Fn[1]) / det_Bn;
	// 			deltan_norm_square = pow(deltan[0],2) + pow(deltan[1],2);

	// 			// Define X_(n+1) = X_n + delta_n
	// 			Xnp1[0] = Xn[0] + deltan[0];
	// 			Xnp1[1] = Xn[1] + deltan[1];

	// 			// Compute F(X_(n+1))
	// 			compute_F_zero(Fnp1, Xnp1, rho, alpha, *eint, d, type_qparticles);
				
	// 			// Define deltaF_n = F(X_(n+1)) - F(X_n)
	// 			deltaFn[0] = Fnp1[0] - Fn[0];
	// 			deltaFn[1] = Fnp1[1] - Fn[1];

	// 			// Define B_(n+1) = B_n + (deltaF_n - B_n * delta_n) * (delta_n)^T / |delta_n|^2
	// 			//                = B_n + F(X_(n+1)) * (delta_n)^T / |delta_n|^2
	// 			for (int i = 0; i < 2; i++){
	// 				for (int j = 0; j < 2; j++)
	// 					Bnp1[i][j] = Bn[i][j] + Fnp1[i]*deltan[j]/deltan_norm_square;
	// 			}

	// 			// Update X_n, F(X_n), B_n for the following iteration
	// 			for (int i = 0; i < 2; i++){
	// 				Xn[i] = Xnp1[i];
	// 				Fn[i] = Fnp1[i];
	// 				for (int j = 0; j < 2; j++)
	// 					Bn[i][j] = Bnp1[i][j];
	// 			}

	// 			Fn_norm = sqrt(pow(Fn[0],2) + pow(Fn[1],2));
	// 			n++;
	// 			printf("n = %d | (z,T) = (%.6e,%.6e) | F(z,T) = (%.6e,%.6e)\n", n, Xn[0], Xn[1], Fn[0], Fn[1]);
	// 		}
			
	// 	}
	// 	else{
	// 		// The goal is to find z such that Q_(d/2)(z) = r with 
	// 		// - Q_(d/2) continuous on [0,+infty[ and strictly increasing on ]0,+infty[
	// 		// - r = mass * hbar^d / (2*pi*T)^(d/2)
	// 		// -----> A unique solution is expected

	// 		r = rho*fabs(alpha)/pow(2.*pi*(*T), 0.5*d);

	// 		if (*z < 0.){
	// 			// The fugacity has never been computed before --> Provide an initial value
	// 			znm1 = 1e-6;
	// 			zn = 1.-1e-6;
	// 		}
	// 		Qdo2z = compute_Qnuz(0.5*d, znm1, type_qparticles);
	// 		if (r > 1.)
	// 			Kznm1 = r/Qdo2z - 1.;
	// 		else
	// 			Kznm1 = Qdo2z/r - 1.;

	// 		Qdo2z = compute_Qnuz(0.5*d, zn, type_qparticles);
	// 		if (r > 1.)
	// 			Kzn = r/Qdo2z - 1.;
	// 		else
	// 			Kzn = Qdo2z/r - 1.;

	// 		znp1 = zn - (zn-znm1)/(Kzn-Kznm1) * Kzn;

	// 		Qdo2z = compute_Qnuz(0.5*d, znp1, type_qparticles);
	// 		if (r > 1.)
	// 			Kznp1 = r/Qdo2z - 1.;
	// 		else
	// 			Kznp1 = Qdo2z/r - 1.;

	// 		n = 0;
	// 		while ((fabs(Kznp1) > THRESHOLD) && (n < 1e8)){
	// 			znm1 = zn;
	// 			Kznm1 = Kzn;

	// 			zn = znp1;
	// 			Kzn = Kznp1;

	// 			znp1 = zn - (zn-znm1)/(Kzn-Kznm1) * Kzn;

	// 			Qdo2z = compute_Qnuz(0.5*d, znp1, type_qparticles);
	// 			if (r > 1.)
	// 				Kznp1 = r/Qdo2z - 1.;
	// 			else
	// 				Kznp1 = Qdo2z/r - 1.;
	// 			n++;
	// 		}
	// 		*z = znp1;

	// 		// Compute the internal energy: e = d/2 * T * Q_(d/2+1)(z) / Q_(d/2)(z)
	// 		Qdo2p1z = compute_Qnuz(0.5*d+1., *z, type_qparticles);
	// 		*eint = 0.5*d*(*T) * Qdo2p1z / Qdo2z;
	// 	}
	// }
}
