#include "./src/core/initialization.h"


void find_initial_step(PARAMETERS *parameters){
	int ndigits, exists;
	char *namefile_data;
	char prefixfile[MAXSTRINGLENGTH];
	int rank, commsize;
	struct stat mystat;

#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &commsize);
#else
	commsize = 1;
	rank = 0;
#endif

	ndigits = fmin((int)(log10(ceil(parameters->nt)))+1, INT_MAX);
	namefile_data = allocation_1d_char(MAXSTRINGLENGTH+5+ndigits);
	strcpy(prefixfile, "");
	strcat(prefixfile, parameters->outputdir);
	strcat(prefixfile, "/");
	strcat(prefixfile, parameters->prefix);
	
	parameters->n_init = -1;
	exists = (1 == 1);
	for (int n = 0; n <= parameters->nt; n++){
		switch(parameters->output_fmt){
			case FMT_BP:
#if defined(key_adios) || defined(key_adios_nompi)
				sprintf(namefile_data, "%s_%0*d.bp", prefixfile, ndigits, n);
				break;
#else
				error("Problem: KINEBEC has not been compiled with Adios dependency");
#endif
			case FMT_H5:
#if defined(key_phdf5) || defined(key_hdf5)
				sprintf(namefile_data, "%s_%0*d.h5", prefixfile, ndigits, n);
				break;
#else
				error("Problem: KINEBEC has not been compiled with Hdf5 dependency");
#endif
			default:
				error("Unknown output format (only .bp or .h5 are allowed)");
		}
		
		// We test if the file exists
		exists = (1==0);
		if (stat(namefile_data, &mystat) == 0){
			// The file exists, we must perform an integrity test on it
			exists = integrity_test(namefile_data, parameters->ndim, parameters->type_collision, parameters->type_qparticles, parameters->exact_solution);

			// If the integrity test succeeded, we increment n_init and save the file name
			if (exists){
				parameters->n_init = n;
				strcpy(parameters->filef0, namefile_data);
			}
			else
				break;
		}
		else
			break;
	}
	
	cut_logfiles(prefixfile, parameters->n_init+1);
	
	free_1d_char(namefile_data);
}



void initialization_from_backup(DIAGS *mydiags, PARAMETERS *parameters){
	PARAMETERS newparameters;
	
	double pi = 4.*atan(1.);
	
	double lambda, mu, C;
	double **Gshifted_2d, ***Gshifted_3d, *Gloc;
	int NVXLOC, NVYLOC, NVZLOC, NVLOC;
	int IVX0LOC, IVY0LOC, IVZ0LOC;
	int itemp, itemp2, istatus;

	int k;

	int discrete_relaxation;

	int commsize;
#ifdef key_mpi
	MPI_Comm_size(MPI_COMM_WORLD, &commsize);
#else
	commsize = 1;
#endif

	// Compute the total number of integration points in the case where we need to enforce discrete convolutions
	switch (parameters->ndim){
		case DIM2:
			if (parameters->force_convolution)
				parameters->mint = parameters->P1_int * parameters->P2_int * parameters->P3_int;
			break;
		case DIM3:
			if (parameters->force_convolution)
				parameters->mint = parameters->P1_int * order_table(parameters->P2_int) * order_table(parameters->P3_int);
			break;
		default:
			error("Unknown dimensionality");
	}

	parameters->T0 = -1.;
	parameters->m0 = -1.;
	parameters->zfug = -1.;
	if (parameters->type_collision == QUANTUM_COLLISIONS){
		find_fugacity_temperature_criticalmass_fixedpoint(parameters->rho0, &(parameters->e0), &(parameters->T0), &(parameters->zfug), &(parameters->m0), parameters->hbar, parameters->ndim+1, parameters->type_qparticles);
		parameters->condensation = test_condensation(parameters->rho0, parameters->e0, parameters->T0, pow(parameters->hbar, parameters->ndim+1), parameters->ndim+1, parameters->type_qparticles);
		discrete_relaxation = search_in_file(parameters->filef0, "distribution/f_limit");
	}
	else{
		parameters->T0 = 2./(parameters->ndim+1.) * parameters->e0;
		parameters->condensation = (1==0);
		discrete_relaxation = (1==1);
	}
	
	parameters->Tinit = ((double)(fmax(0,parameters->n_init))) * parameters->Tfinal/((double)parameters->nt);

	// Read the rescaling parameters
	if (parameters->type_omega == OMEGA_IS_NOTUSED){
		mydiags->omega = 1.;
		lambda = 0.;
		mu = pow(parameters->L/pi, parameters->ndim+1);
	}
	else{
		read_0d_double(parameters->filef0, "distribution", "omega", &(mydiags->omega));
		lambda = 1.;
		mu = 1.;
	}
	C = (1./mu) * pow(parameters->L/(pi*mydiags->omega), parameters->ndim+1);
	
	// Read the non-rescaled distribution
	switch (parameters->ndim){
		case DIM2:
			read_2d(parameters->filef0, "distribution", "f", &Gshifted_2d, &newparameters);
			if ((parameters->nvx != newparameters.nvx) || (parameters->nvy != newparameters.nvy))
				error("Problem while resuming simulation. It appears that you may have modified the grid point numbers. Please restore them or run the code without the --resume flag.");
			break;
		
		case DIM3:
			read_3d(parameters->filef0, "distribution", "f", &Gshifted_3d, &newparameters);
			if ((parameters->nvx != newparameters.nvx) || (parameters->nvy != newparameters.nvy) || (parameters->nvz != newparameters.nvz))
				error("Problem while resuming simulation. It appears that you may have modified the grid point numbers. Please restore them or run the code without the --resume flag.");
			
			break;
		
		default:
			error("Unknown dimensionality");
	}
	
	// Extract the partition table of the velocity grid
	parameters->nvx = newparameters.nvx;
	parameters->nvy = newparameters.nvy;
	if (parameters->ndim == DIM3)
		parameters->nvz = newparameters.nvz;
	else
		parameters->nvz = 0;
	parameters->nvxloc = newparameters.nvxloc;
	parameters->nvyloc = newparameters.nvyloc;
	if (parameters->ndim == DIM3)
		parameters->nvzloc = newparameters.nvzloc;
	else
		parameters->nvzloc = 1;
	parameters->ivx0loc = newparameters.ivx0loc;
	parameters->ivy0loc = newparameters.ivy0loc;
	if (parameters->ndim == DIM3)
		parameters->ivz0loc = newparameters.ivz0loc;
	else
		parameters->ivz0loc = 0;
	
	parameters->ivx0tab = allocation_1d_int(commsize);
	itemp = (int)parameters->ivx0loc;
#ifdef key_mpi
	istatus = MPI_Allgather(&itemp, 1, MPI_INT, parameters->ivx0tab, 1, MPI_INT, MPI_COMM_WORLD);
#else
	parameters->ivx0tab[0] = itemp;
#endif
	parameters->nvxtab = allocation_1d_int(commsize);
	itemp2 = (int)parameters->nvxloc;
#ifdef key_mpi
	istatus = MPI_Allgather(&itemp2, 1, MPI_INT, parameters->nvxtab, 1, MPI_INT, MPI_COMM_WORLD);
#else
	parameters->nvxtab[0] = itemp2;
#endif

	parameters->ivy0tab = allocation_1d_int(commsize);
	itemp = (int)parameters->ivy0loc;
#ifdef key_mpi
	istatus = MPI_Allgather(&itemp, 1, MPI_INT, parameters->ivy0tab, 1, MPI_INT, MPI_COMM_WORLD);
#else
	parameters->ivy0tab[0] = itemp;
#endif
	parameters->nvytab = allocation_1d_int(commsize);
	itemp2 = (int)parameters->nvyloc;
#ifdef key_mpi
	istatus = MPI_Allgather(&itemp2, 1, MPI_INT, parameters->nvytab, 1, MPI_INT, MPI_COMM_WORLD);
#else
	parameters->nvytab[0] = itemp2;
#endif
	
	parameters->ivz0tab = allocation_1d_int(commsize);
	itemp = (int)parameters->ivz0loc;
#ifdef key_mpi
	istatus = MPI_Allgather(&itemp, 1, MPI_INT, parameters->ivz0tab, 1, MPI_INT, MPI_COMM_WORLD);
#else
	parameters->ivz0tab[0] = itemp;
#endif
	parameters->nvztab = allocation_1d_int(commsize);
	itemp2 = (int)parameters->nvzloc;
#ifdef key_mpi
	istatus = MPI_Allgather(&itemp2, 1, MPI_INT, parameters->nvztab, 1, MPI_INT, MPI_COMM_WORLD);
#else
	parameters->nvztab[0] = itemp2;
#endif

#ifdef key_fftw_mpi
	parameters->alloc_local = newparameters.alloc_local;
#endif
	NVXLOC = parameters->nvxloc;
	NVYLOC = parameters->nvyloc;
	NVZLOC = parameters->nvzloc;
	NVLOC = NVXLOC*NVYLOC*NVZLOC;
	IVX0LOC = parameters->ivx0loc;
	IVY0LOC = parameters->ivy0loc;
	IVZ0LOC = parameters->ivz0loc;
	
#ifdef key_force_vectorize
	parameters->kvxabstab = allocation_1d_int(NVLOC);
	parameters->kvyabstab = allocation_1d_int(NVLOC);
	parameters->kvzabstab = allocation_1d_int(NVLOC);
	#pragma omp parallel for private(k, itemp) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, parameters, IVX0LOC, IVY0LOC, IVZ0LOC)
	for (k = 0; k < NVLOC; k++){
		itemp = k/NVZLOC;
		parameters->kvzabstab[k] = k - itemp*NVZLOC;
		parameters->kvxabstab[k] = itemp/NVYLOC;
		parameters->kvyabstab[k] = itemp-parameters->kvxabstab[k]*NVYLOC;

		parameters->kvxabstab[k] += IVX0LOC;
		parameters->kvyabstab[k] += IVY0LOC;
		parameters->kvzabstab[k] += IVZ0LOC;
	}
	// Note : for the 2D case, kvzabstab is allocated and filled with zeros since NVZLOC = 1
#endif

	// Shift the non-rescaled distribution
	switch (parameters->ndim){
		case DIM2:
			mydiags->Gloc_2d = allocation_2d(NVXLOC, NVYLOC);
			shift_2d(Gshifted_2d, mydiags->Gloc_2d, *parameters);
			free_2d(Gshifted_2d);
			Gloc = mydiags->Gloc_2d[0];
			break;
		case DIM3:
			mydiags->Gloc_3d = allocation_3d(NVXLOC, NVYLOC, NVZLOC);
			shift_3d(Gshifted_3d, mydiags->Gloc_3d, *parameters);
			free_3d(Gshifted_3d);
			Gloc = mydiags->Gloc_3d[0][0];
			break;
		default:
			error("Unknown dimensionality");
	}

	// Rescale the distribution
	#pragma omp parallel for private(k) shared(NVLOC, Gloc, C)
	for (k = 0; k < NVLOC; k++)
		Gloc[k] *= C;
	
	// Read the limit non-rescaled distribution and shift it
	//if (!(parameters->condensation)){
	if (discrete_relaxation){
		switch (parameters->ndim){
			case DIM2:
				read_2d(parameters->filef0, "distribution", "f_limit", &Gshifted_2d, &newparameters);
				mydiags->Gloclimit_2d = allocation_2d(parameters->nvxloc, parameters->nvyloc);
				shift_2d(Gshifted_2d, mydiags->Gloclimit_2d, *parameters);
				free_2d(Gshifted_2d);
				Gloc = mydiags->Gloclimit_2d[0];
				break;
			case DIM3:			
				read_3d(parameters->filef0, "distribution", "f_limit", &Gshifted_3d, &newparameters);
				mydiags->Gloclimit_3d = allocation_3d(parameters->nvxloc, parameters->nvyloc, parameters->nvzloc);
				shift_3d(Gshifted_3d, mydiags->Gloclimit_3d, *parameters);
				free_3d(Gshifted_3d);
				Gloc = mydiags->Gloclimit_3d[0][0];
				break;
			default:
				error("Unknown dimensionality");
		}
	
		// Rescale the limit distribution
		#pragma omp parallel for private(k) shared(NVLOC, Gloc, C)
		for (k = 0; k < NVLOC; k++)
			Gloc[k] *= C;
	}

	if (parameters->exact_solution){
		// Read the exact non-rescaled distribution and shift it
		switch (parameters->ndim){
			case DIM2:
				read_2d(parameters->filef0, "distribution", "fexact", &Gshifted_2d, &newparameters);
				mydiags->Glocexact_2d = allocation_2d(parameters->nvxloc, parameters->nvyloc);
				shift_2d(Gshifted_2d, mydiags->Glocexact_2d, *parameters);
				free_2d(Gshifted_2d);
				Gloc = mydiags->Glocexact_2d[0];
				break;
			case DIM3:
				read_3d(parameters->filef0, "distribution", "fexact", &Gshifted_3d, &newparameters);
				mydiags->Glocexact_3d = allocation_3d(parameters->nvxloc, parameters->nvyloc, parameters->nvzloc);
				shift_3d(Gshifted_3d, mydiags->Glocexact_3d, *parameters);
				free_3d(Gshifted_3d);
				Gloc = mydiags->Glocexact_3d[0][0];
				break;
			default:
				error("Unknown dimensionality");
		}
		// Rescale the exact distribution
		#pragma omp parallel for private(k) shared(NVLOC, Gloc, C)
		for (k = 0; k < NVLOC; k++)
			Gloc[k] *= C;
	}
	
	// Read the diagnostics
	read_0d_double(parameters->filef0, "distribution", "Mass", &(mydiags->mass));
	mydiags->u_av = allocation_1d(parameters->ndim+1);
	read_0d_double(parameters->filef0, "distribution", "Average_ux", &(mydiags->u_av[0]));
	read_0d_double(parameters->filef0, "distribution", "Average_uy", &(mydiags->u_av[1]));
	if (parameters->ndim == DIM3)
		read_0d_double(parameters->filef0, "distribution", "Average_uz", &(mydiags->u_av[2]));
	mydiags->stress_energy = allocation_2d(parameters->ndim+1, parameters->ndim+1);
	read_0d_double(parameters->filef0, "distribution", "Stress_energy_xx", &(mydiags->stress_energy[0][0]));
	read_0d_double(parameters->filef0, "distribution", "Stress_energy_xy", &(mydiags->stress_energy[0][1]));
	read_0d_double(parameters->filef0, "distribution", "Stress_energy_yx", &(mydiags->stress_energy[1][0]));
	read_0d_double(parameters->filef0, "distribution", "Stress_energy_yy", &(mydiags->stress_energy[1][1]));
	if (parameters->ndim == DIM3){
		read_0d_double(parameters->filef0, "distribution", "Stress_energy_xz", &(mydiags->stress_energy[0][2]));
		read_0d_double(parameters->filef0, "distribution", "Stress_energy_yz", &(mydiags->stress_energy[1][2]));
		read_0d_double(parameters->filef0, "distribution", "Stress_energy_zx", &(mydiags->stress_energy[2][0]));
		read_0d_double(parameters->filef0, "distribution", "Stress_energy_zy", &(mydiags->stress_energy[2][1]));
		read_0d_double(parameters->filef0, "distribution", "Stress_energy_zz", &(mydiags->stress_energy[2][2]));
	}
	read_0d_double(parameters->filef0, "distribution", "Temperature", &(mydiags->temperature));
	if (parameters->type_collision == QUANTUM_COLLISIONS){
		read_0d_double(parameters->filef0, "distribution", "Fugacity", &(mydiags->fugacity));
		if ((parameters->type_qparticles == BOSONS) && (parameters->ndim == DIM3))
			read_0d_double(parameters->filef0, "distribution", "Critical_mass", &(mydiags->critical_mass));
	}
	read_0d_double(parameters->filef0, "distribution", "Internal_energy", &(mydiags->eint));
	read_0d_double(parameters->filef0, "distribution", "Kinetic_energy", &(mydiags->ekin));
	
	read_0d_double(parameters->filef0, "distribution", "L1_norm_space", &(mydiags->L1));
	read_0d_double(parameters->filef0, "distribution", "L2_norm_space", &(mydiags->L2));
	read_0d_double(parameters->filef0, "distribution", "Linf_norm_space", &(mydiags->Linf));
	read_0d_double(parameters->filef0, "distribution", "L1-L1_norm_time-space", &(mydiags->L1L1));
	read_0d_double(parameters->filef0, "distribution", "L1-L2_norm_time-space", &(mydiags->L1L2));
	read_0d_double(parameters->filef0, "distribution", "L1-Linf_norm_time-space", &(mydiags->L1Linf));
	read_0d_double(parameters->filef0, "distribution", "L2-L1_norm_time-space", &(mydiags->L2L1));
	read_0d_double(parameters->filef0, "distribution", "L2-L2_norm_time-space", &(mydiags->L2L2));
	read_0d_double(parameters->filef0, "distribution", "L2-Linf_norm_time-space", &(mydiags->L2Linf));
	read_0d_double(parameters->filef0, "distribution", "Linf-L1_norm_time-space", &(mydiags->LinfL1));
	read_0d_double(parameters->filef0, "distribution", "Linf-L2_norm_time-space", &(mydiags->LinfL2));
	read_0d_double(parameters->filef0, "distribution", "Linf-Linf_norm_time-space", &(mydiags->LinfLinf));
	read_0d_double(parameters->filef0, "distribution", "Max_f", &(mydiags->maxf));
	read_0d_double(parameters->filef0, "distribution", "Min_f", &(mydiags->minf));
	read_0d_double(parameters->filef0, "distribution", "Entropy", &(mydiags->entropy));
	//if (!(parameters->condensation))
	if (discrete_relaxation)
		read_0d_double(parameters->filef0, "distribution", "Relative_entropy", &(mydiags->rel_entropy));
	read_0d_double(parameters->filef0, "distribution", "4th_order_momentum", &(mydiags->mom4));
	read_0d_double(parameters->filef0, "distribution", "6th_order_momentum", &(mydiags->mom6));
	
	if (parameters->exact_solution){
		read_0d_double(parameters->filef0, "distribution", "L1_norm_exact_space", &(mydiags->L1ex));
		read_0d_double(parameters->filef0, "distribution", "L2_norm_exact_space", &(mydiags->L2ex));
		read_0d_double(parameters->filef0, "distribution", "Linf_norm_exact_space", &(mydiags->Linfex));
		read_0d_double(parameters->filef0, "distribution", "L1-L1_norm_exact_time-space", &(mydiags->L1L1ex));
		read_0d_double(parameters->filef0, "distribution", "L1-L2_norm_exact_time-space", &(mydiags->L1L2ex));
		read_0d_double(parameters->filef0, "distribution", "L1-Linf_norm_exact_time-space", &(mydiags->L1Linfex));
		read_0d_double(parameters->filef0, "distribution", "L2-L1_norm_exact_time-space", &(mydiags->L2L1ex));
		read_0d_double(parameters->filef0, "distribution", "L2-L2_norm_exact_time-space", &(mydiags->L2L2ex));
		read_0d_double(parameters->filef0, "distribution", "L2-Linf_norm_exact_time-space", &(mydiags->L2Linfex));
		read_0d_double(parameters->filef0, "distribution", "Linf-L1_norm_exact_time-space", &(mydiags->LinfL1ex));
		read_0d_double(parameters->filef0, "distribution", "Linf-L2_norm_exact_time-space", &(mydiags->LinfL2ex));
		read_0d_double(parameters->filef0, "distribution", "Linf-Linf_norm_exact_time-space", &(mydiags->LinfLinfex));
		
		read_0d_double(parameters->filef0, "distribution", "L1_norm_error_space", &(mydiags->errL1));
		read_0d_double(parameters->filef0, "distribution", "L2_norm_error_space", &(mydiags->errL2));
		read_0d_double(parameters->filef0, "distribution", "Linf_norm_error_space", &(mydiags->errLinf));
		read_0d_double(parameters->filef0, "distribution", "L1-L1_norm_error_time-space", &(mydiags->errL1L1));
		read_0d_double(parameters->filef0, "distribution", "L1-L2_norm_error_time-space", &(mydiags->errL1L2));
		read_0d_double(parameters->filef0, "distribution", "L1-Linf_norm_error_time-space", &(mydiags->errL1Linf));
		read_0d_double(parameters->filef0, "distribution", "L2-L1_norm_error_time-space", &(mydiags->errL2L1));
		read_0d_double(parameters->filef0, "distribution", "L2-L2_norm_error_time-space", &(mydiags->errL2L2));
		read_0d_double(parameters->filef0, "distribution", "L2-Linf_norm_error_time-space", &(mydiags->errL2Linf));
		read_0d_double(parameters->filef0, "distribution", "Linf-L1_norm_error_time-space", &(mydiags->errLinfL1));
		read_0d_double(parameters->filef0, "distribution", "Linf-L2_norm_error_time-space", &(mydiags->errLinfL2));
		read_0d_double(parameters->filef0, "distribution", "Linf-Linf_norm_error_time-space", &(mydiags->errLinfLinf));
	}
	
	//if (!(parameters->condensation)){
	if (discrete_relaxation){
		read_0d_double(parameters->filef0, "distribution", "Mass_limit", &(mydiags->mass_limit));
		mydiags->u_av_limit = allocation_1d(parameters->ndim+1);
		read_0d_double(parameters->filef0, "distribution", "Average_ux_limit", &(mydiags->u_av_limit[0]));
		read_0d_double(parameters->filef0, "distribution", "Average_uy_limit", &(mydiags->u_av_limit[1]));
		if (parameters->ndim == DIM3)
			read_0d_double(parameters->filef0, "distribution", "Average_uz_limit", &(mydiags->u_av_limit[2]));
		read_0d_double(parameters->filef0, "distribution", "Temperature_limit", &(mydiags->temperature_limit));
		if (parameters->type_collision == QUANTUM_COLLISIONS){
			read_0d_double(parameters->filef0, "distribution", "Fugacity_limit", &(mydiags->fugacity_limit));
			if ((parameters->type_qparticles == BOSONS) && (parameters->ndim == DIM3))
				read_0d_double(parameters->filef0, "distribution", "Critical_mass_limit", &(mydiags->critical_mass_limit));
		}

		read_0d_double(parameters->filef0, "distribution", "L1_norm_space_flimit", &(mydiags->L1limit));
		read_0d_double(parameters->filef0, "distribution", "L2_norm_space_flimit", &(mydiags->L2limit));
		read_0d_double(parameters->filef0, "distribution", "Linf_norm_space_flimit", &(mydiags->Linflimit));
		read_0d_double(parameters->filef0, "distribution", "L1_norm_space_relaxation", &(mydiags->relaxerrL1));
		read_0d_double(parameters->filef0, "distribution", "L2_norm_space_relaxation", &(mydiags->relaxerrL2));
		read_0d_double(parameters->filef0, "distribution", "Linf_norm_space_relaxation", &(mydiags->relaxerrLinf));
		
		read_0d_double(parameters->filef0, "distribution", "L1-L1_norm_time-space_relaxation", &(mydiags->relaxerrL1L1));
		read_0d_double(parameters->filef0, "distribution", "L1-L2_norm_time-space_relaxation", &(mydiags->relaxerrL1L2));
		read_0d_double(parameters->filef0, "distribution", "L1-Linf_norm_time-space_relaxation", &(mydiags->relaxerrL1Linf));
		read_0d_double(parameters->filef0, "distribution", "L2-L1_norm_time-space_relaxation", &(mydiags->relaxerrL2L1));
		read_0d_double(parameters->filef0, "distribution", "L2-L2_norm_time-space_relaxation", &(mydiags->relaxerrL2L2));
		read_0d_double(parameters->filef0, "distribution", "L2-Linf_norm_time-space_relaxation", &(mydiags->relaxerrL2Linf));
		read_0d_double(parameters->filef0, "distribution", "Linf-L1_norm_time-space_relaxation", &(mydiags->relaxerrLinfL1));
		read_0d_double(parameters->filef0, "distribution", "Linf-L2_norm_time-space_relaxation", &(mydiags->relaxerrLinfL2));
		read_0d_double(parameters->filef0, "distribution", "Linf-Linf_norm_time-space_relaxation", &(mydiags->relaxerrLinfLinf));
	}
}


void initialization_from_analytic(DIAGS *mydiags, PARAMETERS *parameters){
	PARAMETERS newparameters;
	
	int NVXLOC, NVYLOC, NVZLOC, NVLOC;
	int IVX0LOC, IVY0LOC, IVZ0LOC;
	int MVX, MVY, MVZ, MV;
	
	int kvxabs, kvyabs, kvzabs;
	double dxix, dxiy, dxiz;
	double dvx, dvy, dvz;
	double xi[3];
	double v[3];
	double vsquare, xisquare;
	double *Gloc, *Gloclimit, *Glocex, f;
	double ***Dirac;
	
	int itemp, insupport, istatus;
	
	double lambda, mu, C;
	double temperature_init, zfug, m0;
	double dt;
	double maxcut_loc, maxcut;
	
	double pi = 4.*atan(1.);

	int k, d;
	
	int commsize, rank;
#ifdef key_mpi
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &commsize);
#else
	rank = 0;
	commsize = 1;
#endif

	// Compute the total number of integration points in the case where we need to enforce discrete convolutions
	switch (parameters->ndim){
		case DIM2:
			if (parameters->force_convolution)
				parameters->mint = parameters->P1_int * parameters->P2_int * parameters->P3_int;
			break;
		case DIM3:
			if (parameters->force_convolution)
				parameters->mint = parameters->P1_int * order_table(parameters->P2_int) * order_table(parameters->P3_int);
			break;
		default:
			error("Unknown dimensionality");
	}
	
	MVX = parameters->nvx;
	MVY = parameters->nvy;
	MVZ = fmax(parameters->nvz,1);
	MV = MVX*MVY*MVZ;
	
	dxix = 2.*pi/MVX;
	dxiy = 2.*pi/MVY;
	if (parameters->ndim == DIM3)
		dxiz = 2.*pi/MVZ;
	else
		dxiz = 1.;
	
	// Build the grid partition table
#ifdef key_fftw_mpi
	switch (parameters->ndim){
		case DIM2:
			parameters->alloc_local = fftw_mpi_local_size_2d(MVX, MVY, MPI_COMM_WORLD, &parameters->nvxloc, &parameters->ivx0loc);
			break;
		case DIM3:
			parameters->alloc_local = fftw_mpi_local_size_3d(MVX, MVY, MVZ, MPI_COMM_WORLD, &parameters->nvxloc, &parameters->ivx0loc);
			break;
		default:
			error("Unknown dimensionality");	
	}
#else
	parameters->nvxloc = MVX;
	parameters->ivx0loc = 0;
#endif
	parameters->nvyloc = MVY;
	parameters->ivy0loc = 0;
	parameters->nvzloc = MVZ;
	parameters->ivz0loc = 0;
	
	NVXLOC = parameters->nvxloc;
	NVYLOC = parameters->nvyloc;
	NVZLOC = parameters->nvzloc;
	NVLOC = NVXLOC*NVYLOC*NVZLOC;
	IVX0LOC = parameters->ivx0loc;
	IVY0LOC = parameters->ivy0loc;
	IVZ0LOC = parameters->ivz0loc;

	parameters->ivx0tab = allocation_1d_int(commsize);
	parameters->ivy0tab = allocation_1d_int(commsize);
	parameters->ivz0tab = allocation_1d_int(commsize);
	parameters->nvxtab = allocation_1d_int(commsize);
	parameters->nvytab = allocation_1d_int(commsize);
	parameters->nvztab = allocation_1d_int(commsize);
	
#ifdef key_mpi
	istatus = MPI_Allgather(&IVX0LOC, 1, MPI_INT, parameters->ivx0tab, 1, MPI_INT, MPI_COMM_WORLD);
	istatus = MPI_Allgather(&IVY0LOC, 1, MPI_INT, parameters->ivy0tab, 1, MPI_INT, MPI_COMM_WORLD);
	istatus = MPI_Allgather(&IVZ0LOC, 1, MPI_INT, parameters->ivz0tab, 1, MPI_INT, MPI_COMM_WORLD);
	istatus = MPI_Allgather(&NVXLOC, 1, MPI_INT, parameters->nvxtab, 1, MPI_INT, MPI_COMM_WORLD);
	istatus = MPI_Allgather(&NVYLOC, 1, MPI_INT, parameters->nvytab, 1, MPI_INT, MPI_COMM_WORLD);
	istatus = MPI_Allgather(&NVZLOC, 1, MPI_INT, parameters->nvztab, 1, MPI_INT, MPI_COMM_WORLD);
#else
	parameters->ivx0tab[0] = IVX0LOC;
	parameters->ivy0tab[0] = IVY0LOC;
	parameters->ivz0tab[0] = IVZ0LOC;
	parameters->nvxtab[0] = NVXLOC;
	parameters->nvytab[0] = NVYLOC;
	parameters->nvztab[0] = NVZLOC;
#endif

#ifdef key_force_vectorize	
	parameters->kvxabstab = allocation_1d_int(NVLOC);
	parameters->kvyabstab = allocation_1d_int(NVLOC);
	parameters->kvzabstab = allocation_1d_int(NVLOC);
	#pragma omp parallel for private(k, itemp) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, parameters)
	for (k = 0; k < NVLOC; k++){
		itemp = k/NVZLOC;
		parameters->kvzabstab[k] = k - itemp*NVZLOC;
		parameters->kvxabstab[k] = itemp/NVYLOC;
		parameters->kvyabstab[k] = itemp-parameters->kvxabstab[k]*NVYLOC;

		parameters->kvxabstab[k] += IVX0LOC;
		parameters->kvyabstab[k] += IVY0LOC;
		parameters->kvzabstab[k] += IVZ0LOC;
	}
	// Note : for the 2D case, kvzabstab is allocated and filled with zeros since NVZLOC = 1
#endif

	maxcut_loc = 0.;
	// We compute the initial temperature T0 according to e0 and rho0
	switch (parameters->type_collision){
		case CLASSICAL_COLLISIONS:
			parameters->T0 = 2./(parameters->ndim+1.) * parameters->e0;
			parameters->zfug = -1.;
			parameters->m0 = -1.;
			break;

		case QUANTUM_COLLISIONS:
			parameters->T0 = -1.;
			parameters->zfug = -1.;
			parameters->m0 = -1.;
			find_fugacity_temperature_criticalmass_fixedpoint(parameters->rho0, &(parameters->e0), &(parameters->T0), &(parameters->zfug), &(parameters->m0), parameters->hbar, parameters->ndim+1, parameters->type_qparticles);
			/*printf("Predicted mass = %e\n", parameters->rho0);
			printf("Predicted u_av = (%e,%e)\n", parameters->u0[0], parameters->u0[1]);
			printf("Predicted internal energy = %e\n", parameters->e0);
			printf("Predicted temperature = %e\n", parameters->T0);
			printf("Predicted fugacity = %e\n", parameters->zfug);*/
			break;

		default:
			error("Unknown type of collisions.");
	}

	// Compute the rescaling parameters according to the definition of the test case
	if (parameters->type_omega != OMEGA_IS_NOTUSED){
		// If Fermi-Dirac quantum degeneracy occurs, we could not be able to compute the temperature at present time
		if (parameters->T0 < 0.)
			error("Fermi-Dirac quantum degeneracy has been detected. At present time, we cannot compute the temperature associated to the initial data. Please deactivate the velocity rescaling method to run the code.");

		// Compute omega from analytic momenta
		if (parameters->type_omega == OMEGA_IS_SQRTEKINORHO){
			vsquare = 0.;
			for (int d = 0; d <= parameters->ndim; d++)
				vsquare += pow(parameters->u0[d],2);
			mydiags->omega = sqrt(parameters->T0+vsquare/(parameters->ndim+1.));
		}
		else if (parameters->type_omega == OMEGA_IS_SQRTTEMPERATURE)
			mydiags->omega = sqrt(parameters->T0);
	}
	else
		mydiags->omega = 1.;

	if (parameters->type_omega == OMEGA_IS_NOTUSED){
		lambda = 0.;
		mu = pow(parameters->L/pi, parameters->ndim+1);
	}
	else{
		lambda = 1.;
		mu = 1.;
	}
	C = (1./mu) * pow(parameters->L/(pi*mydiags->omega), parameters->ndim+1);

	if (parameters->type_collision == QUANTUM_COLLISIONS)
		parameters->condensation = test_condensation(parameters->rho0, parameters->e0, parameters->T0, pow(parameters->hbar, parameters->ndim+1), parameters->ndim+1, parameters->type_qparticles);
	else
		parameters->condensation = (1==0);

	// Compute the initial distribution
	switch (parameters->ndim){
		case DIM2:
			mydiags->Gloc_2d = allocation_2d(NVXLOC, NVYLOC);
			Gloc = mydiags->Gloc_2d[0];
			mydiags->Gloclimit_2d = allocation_2d(NVXLOC, NVYLOC);
			Gloclimit = mydiags->Gloclimit_2d[0];
			if (parameters->exact_solution){
				mydiags->Glocexact_2d = allocation_2d(NVXLOC, NVYLOC);
				Glocex = mydiags->Glocexact_2d[0];
			}
			break;
		case DIM3:
			mydiags->Gloc_3d = allocation_3d(NVXLOC, NVYLOC, NVZLOC);
			Gloc = mydiags->Gloc_3d[0][0];
			if (!(parameters->condensation)){
				mydiags->Gloclimit_3d = allocation_3d(NVXLOC, NVYLOC, NVZLOC);
				Gloclimit = mydiags->Gloclimit_3d[0][0];
			}
			if (parameters->exact_solution){
				mydiags->Glocexact_3d = allocation_3d(NVXLOC, NVYLOC, NVZLOC);
				Glocex = mydiags->Glocexact_3d[0][0];
			}
			break;
		default:
			error("Unknown dimensionality");
	}

	// Project analytic initial distribution on the grid
	switch (parameters->type_f0){
		case CLASSICAL_MAXWELLIAN :
			#pragma omp parallel for private(k, d, kvxabs, kvyabs, kvzabs, itemp, xi, v, xisquare, insupport, f) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, MVX, MVY, MVZ, IVX0LOC, IVY0LOC, IVZ0LOC, parameters, dxix, dxiy, dxiz, pi, mydiags, lambda, C, Gloc, maxcut_loc)
			for (k = 0; k < NVLOC; k++){
#ifdef key_force_vectorize
				kvxabs = parameters->kvxabstab[k] - MVX*(parameters->kvxabstab[k] >= MVX/2);
				kvyabs = parameters->kvyabstab[k] - MVY*(parameters->kvyabstab[k] >= MVY/2);
				kvzabs = parameters->kvzabstab[k] - MVZ*(parameters->kvzabstab[k] >= MVZ/2);
#else
				itemp = k/NVZLOC;
				// itemp now contains kvyloc+NVYLOC*kvxloc;
				kvzabs = k - itemp*NVZLOC;
				kvxabs = itemp/NVYLOC;
				kvyabs = itemp-kvxabs*NVYLOC;
				kvxabs += IVX0LOC;
				kvyabs += IVY0LOC;
				kvzabs += IVZ0LOC;
				if (kvxabs >= MVX/2)
					kvxabs -= MVX;
				if (kvyabs >= MVY/2)
					kvyabs -= MVY;
				if (kvzabs >= MVZ/2)
					kvzabs -= MVZ;
#endif
				xi[0] = kvxabs * dxix;
				xi[1] = kvyabs * dxiy;
				if (parameters->ndim == DIM3)
					xi[2] = kvzabs * dxiz;
				v[0] = parameters->L*xi[0]/(pi*mydiags->omega) + lambda*parameters->u0[0];
				v[1] = parameters->L*xi[1]/(pi*mydiags->omega) + lambda*parameters->u0[1];
				if (parameters->ndim == DIM3)
					v[2] = parameters->L*xi[2]/(pi*mydiags->omega) + lambda*parameters->u0[2];
				
				xisquare = 0.;
				for (d = 0; d <= parameters->ndim; d++)
					xisquare += pow(xi[d],2);
				insupport = (xisquare <= pow(parameters->supportradius,2));
				f = classical_Maxwellian(v, parameters->rho0, parameters->u0, parameters->sigma0, parameters->ndim+1);
				Gloc[k] = C * insupport * f;
				#pragma omp critical
				maxcut_loc = fmax((1-insupport)*fabs(f), maxcut_loc);
			}
			break;
			
		case CLASSICAL_BIMAXWELLIAN :
			#pragma omp parallel for private(k, d, kvxabs, kvyabs, kvzabs, itemp, xi, v, xisquare, insupport, f) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, MVX, MVY, MVZ, IVX0LOC, IVY0LOC, IVZ0LOC, parameters, dxix, dxiy, dxiz, pi, mydiags, lambda, C, Gloc, maxcut_loc)
			for (k = 0; k < NVLOC; k++){
#ifdef key_force_vectorize
				kvxabs = parameters->kvxabstab[k] - MVX*(parameters->kvxabstab[k] >= MVX/2);
				kvyabs = parameters->kvyabstab[k] - MVY*(parameters->kvyabstab[k] >= MVY/2);
				kvzabs = parameters->kvzabstab[k] - MVZ*(parameters->kvzabstab[k] >= MVZ/2);
#else
				itemp = k/NVZLOC;
				// itemp now contains kvyloc+NVYLOC*kvxloc;
				kvzabs = k - itemp*NVZLOC;
				kvxabs = itemp/NVYLOC;
				kvyabs = itemp-kvxabs*NVYLOC;
				kvxabs += IVX0LOC;
				kvyabs += IVY0LOC;
				kvzabs += IVZ0LOC;
				if (kvxabs >= MVX/2)
					kvxabs -= MVX;
				if (kvyabs >= MVY/2)
					kvyabs -= MVY;
				if (kvzabs >= MVZ/2)
					kvzabs -= MVZ;
#endif
				xi[0] = kvxabs * dxix;
				xi[1] = kvyabs * dxiy;
				if (parameters->ndim == DIM3)
					xi[2] = kvzabs * dxiz;
				v[0] = parameters->L*xi[0]/(pi*mydiags->omega) + lambda*parameters->u0[0];
				v[1] = parameters->L*xi[1]/(pi*mydiags->omega) + lambda*parameters->u0[1];
				if (parameters->ndim == DIM3)
					v[2] = parameters->L*xi[2]/(pi*mydiags->omega) + lambda*parameters->u0[2];
					
				xisquare = 0.;
				for (d = 0; d <= parameters->ndim; d++)
					xisquare += pow(xi[d],2);
				insupport = (xisquare <= pow(parameters->supportradius,2));
				f = classical_biMaxwellian(v, parameters->rho0, parameters->u1, parameters->u2, parameters->sigma0, parameters->ndim+1);
				Gloc[k] = C * insupport * f;
				#pragma omp critical
				maxcut_loc = fmax((1-insupport)*fabs(f), maxcut_loc);
			}
			break;
		
		case CLASSICAL_MULTIMAXWELLIAN :
			#pragma omp parallel for private(k, d, kvxabs, kvyabs, kvzabs, itemp, xi, v, xisquare, insupport, f) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, MVX, MVY, MVZ, IVX0LOC, IVY0LOC, IVZ0LOC, parameters, dxix, dxiy, dxiz, pi, mydiags, lambda, C, Gloc, maxcut_loc)
			for (k = 0; k < NVLOC; k++){
#ifdef key_force_vectorize
				kvxabs = parameters->kvxabstab[k] - MVX*(parameters->kvxabstab[k] >= MVX/2);
				kvyabs = parameters->kvyabstab[k] - MVY*(parameters->kvyabstab[k] >= MVY/2);
				kvzabs = parameters->kvzabstab[k] - MVZ*(parameters->kvzabstab[k] >= MVZ/2);
#else
				itemp = k/NVZLOC;
				// itemp now contains kvyloc+NVYLOC*kvxloc;
				kvzabs = k - itemp*NVZLOC;
				kvxabs = itemp/NVYLOC;
				kvyabs = itemp-kvxabs*NVYLOC;
				kvxabs += IVX0LOC;
				kvyabs += IVY0LOC;
				kvzabs += IVZ0LOC;
				if (kvxabs >= MVX/2)
					kvxabs -= MVX;
				if (kvyabs >= MVY/2)
					kvyabs -= MVY;
				if (kvzabs >= MVZ/2)
					kvzabs -= MVZ;
#endif
				xi[0] = kvxabs * dxix;
				xi[1] = kvyabs * dxiy;
				if (parameters->ndim == DIM3)
					xi[2] = kvzabs * dxiz;
				v[0] = parameters->L*xi[0]/(pi*mydiags->omega) + lambda*parameters->u0[0];
				v[1] = parameters->L*xi[1]/(pi*mydiags->omega) + lambda*parameters->u0[1];
				if (parameters->ndim == DIM3)
					v[2] = parameters->L*xi[2]/(pi*mydiags->omega) + lambda*parameters->u0[2];
					
				xisquare = 0.;
				for (d = 0; d <= parameters->ndim; d++)
					xisquare += pow(xi[d],2);
				insupport = (xisquare <= pow(parameters->supportradius,2));
				f = classical_multiMaxwellian(v, parameters->rho0, parameters->u, parameters->sigma0, parameters->ndim+1, parameters->nmaxwellians);
				Gloc[k] = C * insupport * f;
				#pragma omp critical
				maxcut_loc = fmax((1-insupport)*fabs(f), maxcut_loc);
			}
			break;
			
		case QUANTUM_MAXWELLIAN :
			Dirac = allocation_3d(NVXLOC, NVYLOC, NVZLOC);
			if (parameters->condensation){
				if (parameters->type_qparticles == BOSONS){
					// A 3D Bose-Einstein condensate is considered. We have to include a Dirac distribution
					Dirac_3d(Dirac, parameters->m0, parameters->u0, mydiags->omega, *parameters);
				}
				else{
					// A 2D or 3D Fermi-Dirac quantum degeneracy is considered. We cannot handle it at present time
					error("Fermi-Dirac degeneracy has been detected. We cannot build at present time a quantum maxwellian corresponding to the input parameters.");
				}
			}
			else{
				#pragma omp parallel for private(k) shared(NVLOC, Dirac)
				for (k = 0; k < NVLOC; k++)
					Dirac[0][0][k] = 0.;
			}

			#pragma omp parallel for private(k, d, kvxabs, kvyabs, kvzabs, itemp, xi, v, xisquare, insupport, f) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, MVX, MVY, MVZ, parameters, IVX0LOC, IVY0LOC, IVZ0LOC, dxix, dxiy, dxiz, pi, mydiags, lambda, C, Dirac, Gloc, maxcut_loc)
			for (k = 0; k < NVLOC; k++){
#ifdef key_force_vectorize
				kvxabs = parameters->kvxabstab[k] - MVX*(parameters->kvxabstab[k] >= MVX/2);
				kvyabs = parameters->kvyabstab[k] - MVY*(parameters->kvyabstab[k] >= MVY/2);
				kvzabs = parameters->kvzabstab[k] - MVZ*(parameters->kvzabstab[k] >= MVZ/2);
#else
				itemp = k/NVZLOC;
				// itemp now contains kvyloc+NVYLOC*kvxloc;
				kvzabs = k - itemp*NVZLOC;
				kvxabs = itemp/NVYLOC;
				kvyabs = itemp-kvxabs*NVYLOC;
				kvxabs += IVX0LOC;
				kvyabs += IVY0LOC;
				kvzabs += IVZ0LOC;
				if (kvxabs >= MVX/2)
					kvxabs -= MVX;
				if (kvyabs >= MVY/2)
					kvyabs -= MVY;
				if (kvzabs >= MVZ/2)
					kvzabs -= MVZ;
#endif
				xi[0] = kvxabs * dxix;
				xi[1] = kvyabs * dxiy;
				if (parameters->ndim == DIM3)
					xi[2] = kvzabs * dxiz;
				v[0] = parameters->L*xi[0]/(pi*mydiags->omega) + lambda*parameters->u0[0];
				v[1] = parameters->L*xi[1]/(pi*mydiags->omega) + lambda*parameters->u0[1];
				if (parameters->ndim == DIM3)
					v[2] = parameters->L*xi[2]/(pi*mydiags->omega) + lambda*parameters->u0[2];
					
				xisquare = 0.;
				for (d = 0; d <= parameters->ndim; d++)
					xisquare += pow(xi[d],2);
				insupport = (xisquare <= pow(parameters->supportradius,2));
				f = quantum_Maxwellian(v, parameters->u0, parameters->T0, pow(parameters->hbar,parameters->ndim+1), parameters->zfug, parameters->type_qparticles, parameters->ndim+1);
				Gloc[k] = insupport * (C * f + Dirac[0][0][k]);
				#pragma omp critical
				maxcut_loc = fmax((1-insupport)*fabs(f), maxcut_loc);
			}
			free_3d(Dirac);
			
			break;
		
		case CLASSICAL_BKW :
			#pragma omp parallel for private(k, d, kvxabs, kvyabs, kvzabs, itemp, xi, v, xisquare, insupport, f) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, MVX, MVY, MVZ, parameters, IVX0LOC, IVY0LOC, IVZ0LOC, dxix, dxiy, dxiz, pi, mydiags, lambda, C, Gloc, Glocex, maxcut_loc)
			for (k = 0; k < NVLOC; k++){
#ifdef key_force_vectorize
				kvxabs = parameters->kvxabstab[k] - MVX*(parameters->kvxabstab[k] >= MVX/2);
				kvyabs = parameters->kvyabstab[k] - MVY*(parameters->kvyabstab[k] >= MVY/2);
				kvzabs = parameters->kvzabstab[k] - MVZ*(parameters->kvzabstab[k] >= MVZ/2);
#else
				itemp = k/NVZLOC;
				// itemp now contains kvyloc+NVYLOC*kvxloc;
				kvzabs = k - itemp*NVZLOC;
				kvxabs = itemp/NVYLOC;
				kvyabs = itemp-kvxabs*NVYLOC;
				kvxabs += IVX0LOC;
				kvyabs += IVY0LOC;
				kvzabs += IVZ0LOC;
				if (kvxabs >= MVX/2)
					kvxabs -= MVX;
				if (kvyabs >= MVY/2)
					kvyabs -= MVY;
				if (kvzabs >= MVZ/2)
					kvzabs -= MVZ;
#endif
				xi[0] = kvxabs * dxix;
				xi[1] = kvyabs * dxiy;
				if (parameters->ndim == DIM3)
					xi[2] = kvzabs * dxiz;
				v[0] = parameters->L*xi[0]/(pi*mydiags->omega) + lambda*parameters->u0[0];
				v[1] = parameters->L*xi[1]/(pi*mydiags->omega) + lambda*parameters->u0[1];
				if (parameters->ndim == DIM3)
					v[2] = parameters->L*xi[2]/(pi*mydiags->omega) + lambda*parameters->u0[2];
					
				xisquare = 0.;
				for (d = 0; d <= parameters->ndim; d++)
					xisquare += pow(xi[d],2);
				insupport = (xisquare <= pow(parameters->supportradius,2));
				
				f = classical_BKW_2d(v[0], v[1], parameters->sigma0, 0.);
				Gloc[k] = C * insupport * f;
				
				Glocex[k] = Gloc[k];
				#pragma omp critical
				maxcut_loc = fmax((1-insupport)*fabs(f), maxcut_loc);
			}
			break;
		
		case BALL_INDICATOR :
			#pragma omp parallel for private(k, d, kvxabs, kvyabs, kvzabs, itemp, xi, v, xisquare, insupport, f) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, MVX, MVY, MVZ, parameters, IVX0LOC, IVY0LOC, IVZ0LOC, dxix, dxiy, dxiz, pi, mydiags, lambda, C, Gloc, Glocex, maxcut_loc)
			for (k = 0; k < NVLOC; k++){
#ifdef key_force_vectorize
				kvxabs = parameters->kvxabstab[k] - MVX*(parameters->kvxabstab[k] >= MVX/2);
				kvyabs = parameters->kvyabstab[k] - MVY*(parameters->kvyabstab[k] >= MVY/2);
				kvzabs = parameters->kvzabstab[k] - MVZ*(parameters->kvzabstab[k] >= MVZ/2);
#else
				itemp = k/NVZLOC;
				// itemp now contains kvyloc+NVYLOC*kvxloc;
				kvzabs = k - itemp*NVZLOC;
				kvxabs = itemp/NVYLOC;
				kvyabs = itemp-kvxabs*NVYLOC;
				kvxabs += IVX0LOC;
				kvyabs += IVY0LOC;
				kvzabs += IVZ0LOC;
				if (kvxabs >= MVX/2)
					kvxabs -= MVX;
				if (kvyabs >= MVY/2)
					kvyabs -= MVY;
				if (kvzabs >= MVZ/2)
					kvzabs -= MVZ;
#endif
				xi[0] = kvxabs * dxix;
				xi[1] = kvyabs * dxiy;
				if (parameters->ndim == DIM3)
					xi[2] = kvzabs * dxiz;
				v[0] = parameters->L*xi[0]/(pi*mydiags->omega) + lambda*parameters->u0[0];
				v[1] = parameters->L*xi[1]/(pi*mydiags->omega) + lambda*parameters->u0[1];
				if (parameters->ndim == DIM3)
					v[2] = parameters->L*xi[2]/(pi*mydiags->omega) + lambda*parameters->u0[2];
					
				xisquare = 0.;
				for (d = 0; d <= parameters->ndim; d++)
					xisquare += pow(xi[d],2);
				insupport = (xisquare <= pow(parameters->supportradius,2));
				
				f = ball_indicator(v, parameters->rho0, parameters->u0, parameters->e0, parameters->ndim+1);
				Gloc[k] = C * insupport * f;
				#pragma omp critical
				maxcut_loc = fmax((1-insupport)*fabs(f), maxcut_loc);
			}

			break;
		default:
			error("Problem with the use of function initialization_from_analytic");
	}

#ifdef key_mpi
	parameters->maxcut = 0.;
	MPI_Allreduce(&maxcut_loc, &(parameters->maxcut), 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
#else
	parameters->maxcut = maxcut_loc;
#endif

/*
#ifdef key_openmp
	int nthreads = omp_get_max_threads();
	char namefile[MAXSTRINGLENGTH];
	FILE *f;
	sprintf(namefile, "./Ginit_%d.txt", nthreads);
	f = fopen(namefile, "w");
	for (int kvx = 0; kvx < NVXLOC; kvx++){
		for (int kvy = 0; kvy < NVYLOC; kvy++){
			for (int kvz = 0; kvz < NVZLOC; kvz++)
				fprintf(f, "(%d,%d,%d) | %e\n", kvx, kvy, kvz, Gloc[kvz+NVZLOC*(kvy+NVYLOC*kvx)]);
		}
	}
	fclose(f);
#endif
*/
	// Compute the discrete momenta associated to the discretized initial distribution

	// Set averaged velocity to u^0 (necessary for running function compute_diagnostic)
	mydiags->u_av = allocation_1d(parameters->ndim+1);
	for (int d = 0; d <= parameters->ndim; d++)
		mydiags->u_av[d] = parameters->u0[d];

	// Compute the numerical momenta of the initial distribution
	// 1- mass, averaged velocity, internal energy
	// 2- temperature
	// 3- omega (if necessary)
	// 4- the other ones
	compute_diagnostic(mydiags, MASS, *parameters);
	
	if (parameters->type_omega == OMEGA_IS_NOTUSED)
		compute_diagnostic(mydiags, AVERAGE_U, *parameters);
	compute_diagnostic(mydiags, INTERNAL_ENERGY, *parameters);
	compute_diagnostic(mydiags, TEMPERATURE, *parameters);
	/*printf("Discrete mass = %e\n", mydiags->mass);
	printf("Discrete u_av = (%e,%e)\n", mydiags->u_av[0], mydiags->u_av[1]);
	printf("Discrete internal energy = %e\n", mydiags->eint);
	printf("Discrete temperature = %e\n", mydiags->temperature);
	printf("Discrete fugacity = %e\n", mydiags->fugacity);*/
	//error("stop now");
	switch (parameters->type_omega){
		case OMEGA_IS_SQRTEKINORHO:
			vsquare = 0.;
			for (int d = 0; d <= parameters->ndim; d++)
				vsquare += pow(mydiags->u_av[d],2);
			mydiags->omega = sqrt(mydiags->temperature+vsquare/(parameters->ndim+1.));
			break;
		case OMEGA_IS_SQRTTEMPERATURE:
			mydiags->omega = sqrt(mydiags->temperature);
			break;
		case OMEGA_IS_NOTUSED:
			mydiags->omega = 1.;
			break;
		default:
			error("Unknown definition of velocity rescaling");
	}

	mydiags->stress_energy = allocation_2d(parameters->ndim+1, parameters->ndim+1);
	compute_diagnostic(mydiags, STRESS_ENERGY, *parameters);
	compute_diagnostic(mydiags, KINETIC_ENERGY, *parameters);
	
	if (parameters->type_collision == CLASSICAL_COLLISIONS){
		mydiags->fugacity = -1.;
		mydiags->critical_mass = -1.;
	}

	compute_diagnostic(mydiags, MAX_VALUE, *parameters);
	compute_diagnostic(mydiags, MIN_VALUE, *parameters);
	compute_diagnostic(mydiags, ENTROPY, *parameters);
	compute_diagnostic(mydiags, MOMENTUM4, *parameters);
	compute_diagnostic(mydiags, MOMENTUM6, *parameters);
	
	compute_diagnostic(mydiags, L1_NORM, *parameters);
	compute_diagnostic(mydiags, L2_NORM, *parameters);
	compute_diagnostic(mydiags, LINF_NORM, *parameters);
	mydiags->L1L1 = 0.;
	mydiags->L1L2 = 0.;
	mydiags->L1Linf = 0.;
	mydiags->L2L1 = 0.;
	mydiags->L2L2 = 0.;
	mydiags->L2Linf = 0.;
	mydiags->LinfL1 = 0.;
	mydiags->LinfL2 = 0.;
	mydiags->LinfLinf = 0.;

	// If there is no condensation, use the numerical momenta to define the rescaled limit state
	if (!(parameters->condensation)){
		mydiags->mass_limit = mydiags->mass;
		mydiags->u_av_limit = allocation_1d(parameters->ndim+1);
		for (int d = 0; d <= parameters->ndim; d++)
			mydiags->u_av_limit[d] = mydiags->u_av[d];
		mydiags->eint_limit = mydiags->eint;
		mydiags->temperature_limit = mydiags->temperature;
		if (parameters->type_collision == QUANTUM_COLLISIONS){
			mydiags->fugacity_limit = mydiags->fugacity;
			if ((parameters->type_qparticles == BOSONS) && (parameters->ndim == DIM3))
				mydiags->critical_mass_limit = mydiags->critical_mass;
		}
		/*printf("mass limit = %e\n", mydiags->mass_limit);
		printf("u_av limit = (%e,%e)\n", mydiags->u_av_limit[0], mydiags->u_av_limit[1]);
		printf("eint limit = %e\n", mydiags->eint_limit);
		printf("temperature limit = %e\n", mydiags->temperature_limit);
		printf("z limit = %e\n", mydiags->fugacity_limit);*/
		
		switch (parameters->type_collision){
			case CLASSICAL_COLLISIONS:
				classical_Maxwellian_stationary(Gloclimit, mydiags->mass_limit, mydiags->u_av_limit, mydiags->temperature_limit, mydiags->omega, mydiags->u_av, *parameters);
				break;
			
			case QUANTUM_COLLISIONS:
				quantum_Maxwellian_stationary(Gloclimit, mydiags->fugacity_limit, mydiags->u_av_limit, mydiags->temperature_limit, mydiags->critical_mass_limit, mydiags->omega, mydiags->u_av, *parameters);
				break;
			
			default:
				error("Unknown type of collisions");
		}
		
		compute_diagnostic(mydiags, L1_NORM_LIMIT, *parameters);
		compute_diagnostic(mydiags, L2_NORM_LIMIT, *parameters);
		compute_diagnostic(mydiags, LINF_NORM_LIMIT, *parameters);
		compute_diagnostic(mydiags, L1_ERROR_RELAXATION, *parameters);
		compute_diagnostic(mydiags, L2_ERROR_RELAXATION, *parameters);
		compute_diagnostic(mydiags, LINF_ERROR_RELAXATION, *parameters);
		mydiags->relaxerrL1L1 = 0.;
		mydiags->relaxerrL1L2 = 0.;
		mydiags->relaxerrL1Linf = 0.;
		mydiags->relaxerrL2L1 = 0.;
		mydiags->relaxerrL2L2 = 0.;
		mydiags->relaxerrL2Linf = 0.;
		mydiags->relaxerrLinfL1 = 0.;
		mydiags->relaxerrLinfL2 = 0.;
		mydiags->relaxerrLinfLinf = 0.;
		compute_diagnostic(mydiags, RELATIVE_ENTROPY, *parameters);
	}
	
	// If an exact solution exists, define it
	if (parameters->exact_solution){
		exact_solution(Glocex, 0., mydiags->omega, mydiags->u_av, *parameters);

		compute_diagnostic(mydiags, L1_NORM_EX, *parameters);
		compute_diagnostic(mydiags, L2_NORM_EX, *parameters);
		compute_diagnostic(mydiags, LINF_NORM_EX, *parameters);
		compute_diagnostic(mydiags, L1_ERROR, *parameters);
		compute_diagnostic(mydiags, L2_ERROR, *parameters);
		compute_diagnostic(mydiags, LINF_ERROR, *parameters);
		mydiags->errL1L1 = 0.;
		mydiags->errL1L2 = 0.;
		mydiags->errL1Linf = 0.;
		mydiags->errL2L1 = 0.;
		mydiags->errL2L2 = 0.;
		mydiags->errL2Linf = 0.;
		mydiags->errLinfL1 = 0.;
		mydiags->errLinfL2 = 0.;
		mydiags->errLinfLinf = 0.;
		mydiags->L1L1ex = 0.;
		mydiags->L1L2ex = 0.;
		mydiags->L1Linfex = 0.;
		mydiags->L2L1ex = 0.;
		mydiags->L2L2ex = 0.;
		mydiags->L2Linfex = 0.;
		mydiags->LinfL1ex = 0.;
		mydiags->LinfL2ex = 0.;
		mydiags->LinfLinfex = 0.;
	}

	parameters->Tinit = 0.;
	parameters->n_init = 0;
	dt = parameters->Tfinal/(double)(parameters->nt);
	compute_time_diagnostics(mydiags, dt, *parameters);
	write_enriched_parameters(*parameters);
}



void initialization_from_file(DIAGS *mydiags, PARAMETERS *parameters){
	PARAMETERS newparameters;
	double pi = 4.*atan(1.);
	
	int NVXLOC, NVYLOC, NVZLOC, NVLOC;
	int IVX0LOC, IVY0LOC, IVZ0LOC;
	int MVX, MVY, MVZ, MV;
	int itemp, itemp2, istatus;
	
	double **Gshifted_2d, ***Gshifted_3d, *Gloc;
	
	double lambda, mu, C;

	int k;
	
	double dt;

	int commsize;
#ifdef key_mpi
	MPI_Comm_size(MPI_COMM_WORLD, &commsize);
#else
	commsize = 1;
#endif

	// Compute the total number of integration points in the case where we need to enforce discrete convolutions
	switch (parameters->ndim){
		case DIM2:
			if (parameters->force_convolution)
				parameters->mint = parameters->P1_int * parameters->P2_int * parameters->P3_int;
			break;
		case DIM3:
			if (parameters->force_convolution)
				parameters->mint = parameters->P1_int * order_table(parameters->P2_int) * order_table(parameters->P3_int);
			break;
		default:
			error("Unknown dimensionality");
	}

	// Read the distribution
	switch (parameters->ndim){
		case DIM2:
			read_2d(parameters->filef0, "distribution", "f", &Gshifted_2d, &newparameters);
			break;
		
		case DIM3:
			read_3d(parameters->filef0, "distribution", "f", &Gshifted_3d, &newparameters);
			break;
		
		default:
			error("Unknown dimensionality");
	}

	// Extract the partition table of the velocity grid
	parameters->nvx = newparameters.nvx;
	parameters->nvy = newparameters.nvy;
	if (parameters->ndim == DIM3)
		parameters->nvz = newparameters.nvz;
	else
		parameters->nvz = 0;
	parameters->nvxloc = newparameters.nvxloc;
	parameters->nvyloc = newparameters.nvyloc;
	if (parameters->ndim == DIM3)
		parameters->nvzloc = newparameters.nvzloc;
	else
		parameters->nvzloc = 1;
	parameters->ivx0loc = newparameters.ivx0loc;
	parameters->ivy0loc = newparameters.ivy0loc;
	if (parameters->ndim == DIM3)
		parameters->ivz0loc = newparameters.ivz0loc;
	else
		parameters->ivz0loc = 0;

	parameters->ivx0tab = allocation_1d_int(commsize);
	itemp = (int)parameters->ivx0loc;
#ifdef key_mpi
	istatus = MPI_Allgather(&itemp, 1, MPI_INT, parameters->ivx0tab, 1, MPI_INT, MPI_COMM_WORLD);
#else
	parameters->ivx0tab[0] = itemp;
#endif
	parameters->nvxtab = allocation_1d_int(commsize);
	itemp2 = (int)parameters->nvxloc;
#ifdef key_mpi
	istatus = MPI_Allgather(&itemp2, 1, MPI_INT, parameters->nvxtab, 1, MPI_INT, MPI_COMM_WORLD);
#else
	parameters->nvxtab[0] = itemp2;
#endif

	parameters->ivy0tab = allocation_1d_int(commsize);
	itemp = (int)parameters->ivy0loc;
#ifdef key_mpi
	istatus = MPI_Allgather(&itemp, 1, MPI_INT, parameters->ivy0tab, 1, MPI_INT, MPI_COMM_WORLD);
#else
	parameters->ivy0tab[0] = itemp;
#endif
	parameters->nvytab = allocation_1d_int(commsize);
	itemp2 = (int)parameters->nvyloc;
#ifdef key_mpi
	istatus = MPI_Allgather(&itemp2, 1, MPI_INT, parameters->nvytab, 1, MPI_INT, MPI_COMM_WORLD);
#else
	parameters->nvytab[0] = itemp2;
#endif

	parameters->ivz0tab = allocation_1d_int(commsize);
	itemp = (int)parameters->ivz0loc;
#ifdef key_mpi
	istatus = MPI_Allgather(&itemp, 1, MPI_INT, parameters->ivz0tab, 1, MPI_INT, MPI_COMM_WORLD);
#else
	parameters->ivz0tab[0] = itemp;
#endif
	parameters->nvztab = allocation_1d_int(commsize);
	itemp2 = (int)parameters->nvzloc;
#ifdef key_mpi
	istatus = MPI_Allgather(&itemp2, 1, MPI_INT, parameters->nvztab, 1, MPI_INT, MPI_COMM_WORLD);
#else
	parameters->nvztab[0] = itemp2;
#endif

#ifdef key_fftw_mpi
	parameters->alloc_local = newparameters.alloc_local;
#endif
	NVXLOC = parameters->nvxloc;
	NVYLOC = parameters->nvyloc;
	NVZLOC = parameters->nvzloc;
	NVLOC = NVXLOC*NVYLOC*NVZLOC;
	IVX0LOC = parameters->ivx0loc;
	IVY0LOC = parameters->ivy0loc;
	IVZ0LOC = parameters->ivz0loc;
	MVX = parameters->nvx;
	MVY = parameters->nvy;
	MVZ = fmax(parameters->nvz,1);
	MV = MVX*MVY*MVZ;
	
#ifdef key_force_vectorize	
	parameters->kvxabstab = allocation_1d_int(NVLOC);
	parameters->kvyabstab = allocation_1d_int(NVLOC);
	parameters->kvzabstab = allocation_1d_int(NVLOC);
	#pragma omp parallel for private(k, itemp) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, IVX0LOC, IVY0LOC, IVZ0LOC, parameters) 
	for (k = 0; k < NVLOC; k++){
		itemp = k/NVZLOC;
		parameters->kvzabstab[k] = k - itemp*NVZLOC;
		parameters->kvxabstab[k] = itemp/NVYLOC;
		parameters->kvyabstab[k] = itemp-parameters->kvxabstab[k]*NVYLOC;

		parameters->kvxabstab[k] += IVX0LOC;
		parameters->kvyabstab[k] += IVY0LOC;
		parameters->kvzabstab[k] += IVZ0LOC;
	}
	// Note : for the 2D case, kvzabstab is allocated and filled with zeros since NVZLOC = 1
#endif

	// Read the value of L
	read_0d_double(parameters->filef0, "distribution", "L", &(parameters->L));
		
	// Read the configuration of velocity rescaling method
	read_0d_int(parameters->filef0, "distribution", "type_omega", &(parameters->type_omega));
	if (parameters->type_omega != OMEGA_IS_NOTUSED)
		read_0d_double(parameters->filef0, "distribution", "omega", &(mydiags->omega));
	else
		mydiags->omega = 1.;
	
	if (parameters->type_omega == OMEGA_IS_NOTUSED){
		lambda = 0.;
		mu = pow(parameters->L/pi, parameters->ndim+1);
	}
	else{
		lambda = 1.;
		mu = 1.;
	}
	C = (1./mu) * pow(parameters->L/(pi*mydiags->omega), parameters->ndim+1);

	// Read the distribution and shift it
	switch (parameters->ndim){
		case DIM2:
			mydiags->Gloc_2d = allocation_2d(parameters->nvxloc, parameters->nvyloc);
			shift_2d(Gshifted_2d, mydiags->Gloc_2d, *parameters);
			free_2d(Gshifted_2d);
			Gloc = mydiags->Gloc_2d[0];
			break;
		case DIM3:
			mydiags->Gloc_3d = allocation_3d(parameters->nvxloc, parameters->nvyloc, parameters->nvzloc);
			shift_3d(Gshifted_3d, mydiags->Gloc_3d, *parameters);
			free_3d(Gshifted_3d);
			Gloc = mydiags->Gloc_3d[0][0];
			break;
		default:
			error("Unknown dimensionality");
	}
	// Rescale the distribution
	#pragma omp parallel for private(k) shared(NVXLOC, NVYLOC, NVZLOC, Gloc, C)
	for (int k = 0; k < NVXLOC*NVYLOC*NVZLOC; k++){
		#pragma omp atomic
		Gloc[k] *= C;
	}

	
	// Compute the numerical momenta of the initial distribution
	mydiags->u_av = allocation_1d(parameters->ndim+1);
	mydiags->stress_energy = allocation_2d(parameters->ndim+1, parameters->ndim+1);
	
	compute_diagnostic(mydiags, MASS, *parameters);
	compute_diagnostic(mydiags, AVERAGE_U, *parameters);
	compute_diagnostic(mydiags, INTERNAL_ENERGY, *parameters);
	compute_diagnostic(mydiags, STRESS_ENERGY, *parameters);
	compute_diagnostic(mydiags, KINETIC_ENERGY, *parameters);
	compute_diagnostic(mydiags, INTERNAL_ENERGY, *parameters);
	compute_diagnostic(mydiags, TEMPERATURE, *parameters);
	
	compute_diagnostic(mydiags, MAX_VALUE, *parameters);
	compute_diagnostic(mydiags, MIN_VALUE, *parameters);
	compute_diagnostic(mydiags, ENTROPY, *parameters);
	compute_diagnostic(mydiags, MOMENTUM4, *parameters);
	compute_diagnostic(mydiags, MOMENTUM6, *parameters);
	
	compute_diagnostic(mydiags, L1_NORM, *parameters);
	compute_diagnostic(mydiags, L2_NORM, *parameters);
	compute_diagnostic(mydiags, LINF_NORM, *parameters);
	mydiags->L1L1 = 0.;
	mydiags->L1L2 = 0.;
	mydiags->L1Linf = 0.;
	mydiags->L2L1 = 0.;
	mydiags->L2L2 = 0.;
	mydiags->L2Linf = 0.;
	mydiags->LinfL1 = 0.;
	mydiags->LinfL2 = 0.;
	mydiags->LinfLinf = 0.;
	
	// Identify if a condensate is expected
	parameters->condensation = test_condensation(mydiags->mass, mydiags->eint, mydiags->temperature, pow(parameters->hbar, parameters->ndim+1), parameters->ndim+1, parameters->type_qparticles);

	// If there is no condensation, use the numerical momenta to define the limit state
	if (!(parameters->condensation)){
		mydiags->mass_limit = mydiags->mass;
		mydiags->u_av_limit = allocation_1d(parameters->ndim+1);
		for (int d = 0; d <= parameters->ndim; d++)
			mydiags->u_av_limit[d] = mydiags->u_av[d];
		mydiags->temperature_limit = mydiags->temperature;
		if (parameters->type_collision == QUANTUM_COLLISIONS){
			mydiags->fugacity_limit = mydiags->fugacity;
			if ((parameters->type_qparticles == BOSONS) && (parameters->ndim == DIM3))
				mydiags->critical_mass_limit = mydiags->critical_mass;
		}

		switch (parameters->ndim){
			case DIM2:
				mydiags->Gloclimit_2d = allocation_2d(NVXLOC, NVYLOC);
				Gloc = mydiags->Gloclimit_2d[0];
				break;
			case DIM3:
				mydiags->Gloclimit_3d = allocation_3d(NVXLOC, NVYLOC, NVZLOC);
				Gloc = mydiags->Gloclimit_3d[0][0];
				break;
			default:
				error("Unknown dimensionality");
		}
		
		switch (parameters->type_collision){
			case CLASSICAL_COLLISIONS:
				classical_Maxwellian_stationary(Gloc, mydiags->mass_limit, mydiags->u_av_limit, mydiags->temperature_limit, mydiags->omega, mydiags->u_av, *parameters);
				break;
			
			case QUANTUM_COLLISIONS:
				quantum_Maxwellian_stationary(Gloc, mydiags->fugacity_limit, mydiags->u_av_limit, mydiags->temperature_limit, mydiags->critical_mass_limit, mydiags->omega, mydiags->u_av, *parameters);
				break;
			
			default:
				error("Unknown type of collisions");
		}
		
		compute_diagnostic(mydiags, L1_NORM_LIMIT, *parameters);
		compute_diagnostic(mydiags, L2_NORM_LIMIT, *parameters);
		compute_diagnostic(mydiags, LINF_NORM_LIMIT, *parameters);
		compute_diagnostic(mydiags, L1_ERROR_RELAXATION, *parameters);
		compute_diagnostic(mydiags, L2_ERROR_RELAXATION, *parameters);
		compute_diagnostic(mydiags, LINF_ERROR_RELAXATION, *parameters);
		mydiags->relaxerrL1L1 = 0.;
		mydiags->relaxerrL1L2 = 0.;
		mydiags->relaxerrL1Linf = 0.;
		mydiags->relaxerrL2L1 = 0.;
		mydiags->relaxerrL2L2 = 0.;
		mydiags->relaxerrL2Linf = 0.;
		mydiags->relaxerrLinfL1 = 0.;
		mydiags->relaxerrLinfL2 = 0.;
		mydiags->relaxerrLinfLinf = 0.;
		compute_diagnostic(mydiags, RELATIVE_ENTROPY, *parameters);
	}
	
	// Read the initial time value
	read_0d_double(parameters->filef0, "distribution", "Time", &(parameters->Tinit));
		
	// If it is asked in the setup file, set the initial time to 0
	if (parameters->type_Tinit == SETTOZERO)
		parameters->Tinit = 0.;

	// Read the configuration of the cut-off
	if (parameters->cutoff <= 0.)
		read_0d_double(parameters->filef0, "distribution", "Cut-off", &(parameters->cutoff));
	
	parameters->n_init = 0;
	dt = (parameters->Tfinal-parameters->Tinit)/(double)(parameters->nt);
	compute_time_diagnostics(mydiags, dt, *parameters);
}


void initialization(DIAGS *mydiags, PARAMETERS *parameters){
	// If it is asked to resume a run, we find the last file to load
	if (parameters->resume)
		find_initial_step(parameters);
	else
		parameters->n_init = 0;

	// Confirm that there is a run that has been aborted
	parameters->resume = ((parameters->resume) && (parameters->n_init >= 0));
	
	if (parameters->resume){
		// Initialize the run with a full reading of a data file
		initialization_from_backup(mydiags, parameters);
	}
	else{
		if (parameters->type_f0 == FROM_FILE){
			// Initialize the run by reading the distribution in a data file
			initialization_from_file(mydiags, parameters);
		}
		else{
			// Initialize the run with an analytically defined distribution
			initialization_from_analytic(mydiags, parameters);
		}
	}
}