#include "./src/core/stationary_functions.h"


void Dirac_3d(double ***delta, double m0, double *u, double omega, PARAMETERS parameters){
	double pi = 4.*atan(1.);

	int NVXLOC = (int)parameters.nvxloc;
	int IVX0LOC = (int)parameters.ivx0loc;
	int NVYLOC = (int)parameters.nvyloc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int NVZLOC = (int)parameters.nvzloc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = parameters.nvz;
	int kvxabs, kvyabs, kvzabs;
	int itemp;

	int k;
	
	double mu, myomega, lambda;
	int nonzeromode = (1==1);
	double inexp_vx, inexp_vy, inexp_vz;
	
	if (parameters.type_omega == OMEGA_IS_NOTUSED){
		mu = pow(parameters.L/pi, parameters.ndim+1);
		myomega = 1.;
		lambda = 0.;
	}
	else{
		mu = 1.;
		myomega = omega;
		lambda = 1.;
	}

	for (int d = 0; d <= parameters.ndim; d++)
		nonzeromode = (nonzeromode && ((1.-lambda)*u[d] >= -parameters.L/myomega) && ((1.-lambda)*u[d] <= parameters.L/myomega));

#if defined(key_fftw) || defined(key_fftw_mpi) || defined(key_fftw_openmp)
	fftw_complex *delta_spectral, *delta_real;
	fftw_plan plan_delta_backward;

#ifdef key_fftw_mpi
	delta_spectral = fftw_alloc_complex(parameters.alloc_local);
	delta_real = fftw_alloc_complex(parameters.alloc_local);
	plan_delta_backward = fftw_mpi_plan_dft_3d(MVX, MVY, MVZ, delta_spectral, delta_real, MPI_COMM_WORLD, FFTW_BACKWARD, FFTW_ESTIMATE);
#else
	delta_spectral = fftw_malloc(sizeof(fftw_complex)*NVXLOC*NVYLOC*NVZLOC);
	delta_real = fftw_malloc(sizeof(fftw_complex)*NVXLOC*NVYLOC*NVZLOC);
	plan_delta_backward = fftw_plan_dft_3d(MVX, MVY, MVZ, delta_spectral, delta_real, FFTW_BACKWARD, FFTW_ESTIMATE);
#endif
	
	if (nonzeromode){
		#pragma omp parallel for private(k, kvxabs, kvyabs, kvzabs, itemp, inexp_vx, inexp_vy, inexp_vz) shared(NVXLOC, NVYLOC, NVZLOC, MVX, MVY, MVZ, parameters, IVX0LOC, IVY0LOC, IVZ0LOC, pi, myomega, lambda, u, m0, mu, delta_spectral)
		for (k = 0; k < NVXLOC*NVYLOC*NVZLOC; k++){
#ifdef key_force_vectorize
			kvxabs = parameters.kvxabstab[k] - MVX*(parameters.kvxabstab[k] >= MVX/2);
			kvyabs = parameters.kvyabstab[k] - MVY*(parameters.kvyabstab[k] >= MVY/2);
			kvzabs = parameters.kvzabstab[k] - MVZ*(parameters.kvzabstab[k] >= MVZ/2);
#else
			// k = kvzloc+NVZLOC*(kvyloc+NVYLOC*kvxloc)
			itemp = k/NVZLOC;
			// itemp now contains kvyloc+NVYLOC*kvxloc;
			kvzabs = k - itemp*NVZLOC;
			kvxabs = itemp/NVYLOC;
			kvyabs = itemp-kvxabs*NVYLOC;
											
			kvxabs += IVX0LOC;
			kvyabs += IVY0LOC;
			kvzabs += IVZ0LOC;
											
			if (kvxabs >= MVX/2)
				kvxabs -= MVX;
			if (kvyabs >= MVY/2)
				kvyabs -= MVY;
			if (kvzabs >= MVZ/2)
				kvzabs -= MVZ;
#endif
			inexp_vx = -2.*pi*myomega/parameters.L * kvxabs*(1.-lambda)*u[0];
			inexp_vy = -2.*pi*myomega/parameters.L * kvyabs*(1.-lambda)*u[1];
			inexp_vz = -2.*pi*myomega/parameters.L * kvzabs*(1.-lambda)*u[2];
			
			delta_spectral[k][0] = fmax(m0,0.)/(mu*pow(2.*pi,3))*cos(inexp_vx+inexp_vy+inexp_vz);
			delta_spectral[k][1] = fmax(m0,0.)/(mu*pow(2.*pi,3))*sin(inexp_vx+inexp_vy+inexp_vz);
		}
		
		fftw_execute(plan_delta_backward);
		
		#pragma omp parallel for private(k) shared(NVXLOC, NVYLOC, NVZLOC, delta, delta_real)
		for (k = 0; k < NVXLOC*NVYLOC*NVZLOC; k++)
			delta[0][0][k] = delta_real[k][0];
	}
	else{
		#pragma omp parallel for private(k) shared(NVXLOC, NVYLOC, NVZLOC, delta)
		for (k = 0; k < NVXLOC*NVYLOC*NVZLOC; k++)
			delta[0][0][k] = 0.;
	}

	fftw_destroy_plan(plan_delta_backward);
	fftw_free(delta_spectral);
	fftw_free(delta_real);
#endif
}


double classical_Maxwellian(double *v, double rho, double *u_shift, double sigma, int ndim){
	double pi = 4.*atan(1.);
	double vsquare = 0.;
	for (int d = 0; d < ndim; d++)
		vsquare += pow(v[d]-u_shift[d],2);
	vsquare /= (2.*sigma);
	double theta = rho/pow(2.*pi*sigma, 0.5*ndim);
	return theta*exp(-vsquare);
}


double classical_biMaxwellian(double *v, double rho, double *u1_shift, double *u2_shift, double sigma, int ndim){
	double pi = 4.*atan(1.);
	double vsquare_1 = 0.;
	double vsquare_2 = 0.;
	for (int d = 0; d < ndim; d++){
		vsquare_1 += pow(v[d]-u1_shift[d],2);
		vsquare_2 += pow(v[d]-u2_shift[d],2);
	}
	vsquare_1 /= (2.*sigma);
	vsquare_2 /= (2.*sigma);
	double theta = 0.5 * rho/pow(2.*pi*sigma, 0.5*ndim);
	return theta*(exp(-vsquare_1)+exp(-vsquare_2));
}


double classical_multiMaxwellian(double *v, double rho, double **u_shift, double sigma, int ndim, int nmaxwellians){
	double pi = 4.*atan(1.);
	double vsquare_k;
	double res = 0.;

	for (int k = 0; k < nmaxwellians; k++){
		vsquare_k = 0.;
		for (int d = 0; d < ndim; d++)
			vsquare_k += pow(v[d]-u_shift[k][d], 2);
		vsquare_k /= (2.*sigma);
		res += exp(-vsquare_k);
	}
	res *= (rho/pow(2.*pi*sigma, 0.5*ndim)) / nmaxwellians;
	return res;
}


double quantum_Maxwellian(double *v, double *u_shift, double T, double theta0, double fugacity, int type_qparticles, int ndim){
	double vsquare = 0.;
	for (int d = 0; d < ndim; d++)
		vsquare += pow(v[d]-u_shift[d],2);
	vsquare /= (2.*T);
	switch (type_qparticles){
		case BOSONS :
			return 1./(theta0*(exp(vsquare)/fugacity-1.));
		case FERMIONS :
			return 1./(theta0*(exp(vsquare)/fugacity+1.));
		default :
			error("Unknown kind of quantum particles\n");
	}
}


double ball_indicator(double *v, double rho, double *u_shift, double eint, int ndim){
	double A = sqrt(2.*eint*(ndim+2.)/ndim);
	double pi = 4.*atan(1.);
	double ballvolume, gamma, vsquare;
	
	// Compute Gamma(d/2+1)
	if (ndim == 2)
		gamma = 1.;
	else if (ndim == 3)
		gamma = 0.75*sqrt(pi);
	
	// Compute the volume of the ball B(u0,A)
	ballvolume = pow(pi, 0.5*ndim) * pow(A, ndim) / gamma;

	vsquare = 0.;
	for (int d = 0; d < ndim; d++)
		vsquare += pow(v[d]-u_shift[d],2);
	
	if (vsquare <= A*A)
		return rho/ballvolume;
	else
		return 0.;
}
