#include "./src/core/stationary_states.h"


void classical_Maxwellian_stationary(double *Gstable, double rhostable, double *ushiftstable, double Tstable, double omega, double *u, PARAMETERS parameters){
	int kvxabs, kvyabs, kvzabs;
	int itemp;

	double pi = 4.*atan(1.);
	double lambda, myomega, mu, C;

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(1,parameters.nvz);
	
	double v[3];
	double xi[3];
	double dxix, dxiy, dxiz;
	double fstable;

	int k;
	
	dxix = 2.*pi/MVX;
	dxiy = 2.*pi/MVY;
	if (parameters.ndim == DIM3)
		dxiz = 2.*pi/MVZ;
	else
		dxiz = 1.;
	
	
	if (parameters.type_omega == OMEGA_IS_NOTUSED){
		lambda = 0.;
		myomega = 1.;
		mu = pow(parameters.L/pi, parameters.ndim+1);
	}
	else{
		lambda = 1.;
		myomega = omega;
		mu = 1.;
	}

	C = (1./mu) * pow(parameters.L/(pi*myomega), parameters.ndim+1);
	
	#pragma omp parallel for private(k, kvxabs, kvyabs, kvzabs, itemp, xi, v, fstable) shared(NVXLOC, NVYLOC, NVZLOC, MVX, MVY, MVZ, parameters, IVX0LOC, IVY0LOC, IVZ0LOC, dxix, dxiy, dxiz, pi, myomega, lambda, u, rhostable, ushiftstable, Tstable, C, Gstable)
	for (k = 0; k < NVXLOC*NVYLOC*NVZLOC; k++){
#ifdef key_force_vectorize
		kvxabs = parameters.kvxabstab[k] - MVX*(parameters.kvxabstab[k] >= MVX/2);
		kvyabs = parameters.kvyabstab[k] - MVY*(parameters.kvyabstab[k] >= MVY/2);
		kvzabs = parameters.kvzabstab[k] - MVZ*(parameters.kvzabstab[k] >= MVZ/2);
#else
		// k = kvzloc+NVZLOC*(kvyloc+NVYLOC*kvxloc)
		itemp = k/NVZLOC;
		// itemp now contains kvyloc+NVYLOC*kvxloc;
		kvzabs = k - itemp*NVZLOC;
		kvxabs = itemp/NVYLOC;
		kvyabs = itemp-kvxabs*NVYLOC;
											
		kvxabs += IVX0LOC;
		kvyabs += IVY0LOC;
		kvzabs += IVZ0LOC;
											
		if (kvxabs >= MVX/2)
			kvxabs -= MVX;
		if (kvyabs >= MVY/2)
			kvyabs -= MVY;
		if (kvzabs >= MVZ/2)
			kvzabs -= MVZ;
#endif
		xi[0] = kvxabs * dxix;
		xi[1] = kvyabs * dxiy;
		if (parameters.ndim == DIM3)
			xi[2] = kvzabs * dxiz;
		v[0] = parameters.L*xi[0]/(pi*myomega) + lambda*u[0];
		v[1] = parameters.L*xi[1]/(pi*myomega) + lambda*u[1];
		if (parameters.ndim == DIM3)
			v[2] = parameters.L*xi[2]/(pi*myomega) + lambda*u[2];
			
		// Compute the non-rescaled distribution f
		fstable = classical_Maxwellian(v, rhostable, ushiftstable, Tstable, parameters.ndim+1);
				
		// Compute the rescaled distribution G
		//Gstable[k] = fstable / (mu * pow(pi*myomega/parameters.L, parameters.ndim+1));
		Gstable[k] = fstable * C;
	}
}



void quantum_Maxwellian_stationary(double *Gstable, double zstable, double *ushiftstable, double Tstable, double m0stable, double omega, double *u, PARAMETERS parameters){
	int kvxabs, kvyabs, kvzabs;
	int itemp;

	double pi = 4.*atan(1.);
	double lambda, myomega, mu, C;

	int NVXLOC = (int)parameters.nvxloc;
	int NVYLOC = (int)parameters.nvyloc;
	int NVZLOC = (int)parameters.nvzloc;
	int IVX0LOC = (int)parameters.ivx0loc;
	int IVY0LOC = (int)parameters.ivy0loc;
	int IVZ0LOC = (int)parameters.ivz0loc;
	int MVX = parameters.nvx;
	int MVY = parameters.nvy;
	int MVZ = fmax(1,parameters.nvz);
	
	double ***Dirac;

	double v[3];
	double xi[3];	
	double dxix, dxiy, dxiz;
	double fstable;

	int k;
	
	dxix = 2.*pi/MVX;
	dxiy = 2.*pi/MVY;
	if (parameters.ndim == DIM3)
		dxiz = 2.*pi/MVZ;
	else
		dxiz = 1.;
	
	if (parameters.type_omega == OMEGA_IS_NOTUSED){
		lambda = 0.;
		myomega = 1.;
		mu = pow(parameters.L/pi, parameters.ndim+1);
	}
	else{
		lambda = 1.;
		myomega = omega;
		mu = 1.;
	}

	C = (1./mu) * pow(parameters.L/(pi*myomega), parameters.ndim+1);
	
	if (parameters.condensation){
		if ((parameters.type_qparticles == BOSONS) && (parameters.ndim == DIM3)){
			// A Bose-Einstein condensate is considered. We have to include a Dirac distribution
			Dirac = allocation_3d(NVXLOC, NVYLOC, NVZLOC);
			Dirac_3d(Dirac, m0stable, ushiftstable, myomega, parameters);
		}
		else{
			// A 2D or 3D Fermi-Dirac quantum degeneracy has been detected. We cannot handle it at present time
			error("Fermi-Dirac degeneracy has been detected. We cannot build at present time a quantum maxwellian corresponding to the input parameters.");
		}
	}
	
	#pragma omp parallel for private(k, kvxabs, kvyabs, kvzabs, itemp, xi, v, fstable) shared(NVXLOC, NVYLOC, NVZLOC, parameters, MVX, MVY, MVZ, IVX0LOC, IVY0LOC, IVZ0LOC, dxix, dxiy, dxiz, pi, myomega, lambda, u, ushiftstable, Tstable, zstable, Gstable, C, Dirac)
	for (k = 0; k < NVXLOC*NVYLOC*NVZLOC; k++){
#ifdef key_force_vectorize
		kvxabs = parameters.kvxabstab[k] - MVX*(parameters.kvxabstab[k] >= MVX/2);
		kvyabs = parameters.kvyabstab[k] - MVY*(parameters.kvyabstab[k] >= MVY/2);
		kvzabs = parameters.kvzabstab[k] - MVZ*(parameters.kvzabstab[k] >= MVZ/2);
#else
		// k = kvzloc+NVZLOC*(kvyloc+NVYLOC*kvxloc)
		itemp = k/NVZLOC;
		// itemp now contains kvyloc+NVYLOC*kvxloc;
		kvzabs = k - itemp*NVZLOC;
		kvxabs = itemp/NVYLOC;
		kvyabs = itemp-kvxabs*NVYLOC;
											
		kvxabs += IVX0LOC;
		kvyabs += IVY0LOC;
		kvzabs += IVZ0LOC;
											
		if (kvxabs >= MVX/2)
			kvxabs -= MVX;
		if (kvyabs >= MVY/2)
			kvyabs -= MVY;
		if (kvzabs >= MVZ/2)
			kvzabs -= MVZ;
#endif
		xi[0] = kvxabs * dxix;
		xi[1] = kvyabs * dxiy;
		if (parameters.ndim == DIM3)
			xi[2] = kvzabs * dxiz;
		v[0] = parameters.L*xi[0]/(pi*myomega) + lambda*u[0];
		v[1] = parameters.L*xi[1]/(pi*myomega) + lambda*u[1];
		if (parameters.ndim == DIM3)
			v[2] = parameters.L*xi[2]/(pi*myomega) + lambda*u[2];
		
		// Compute the non-rescaled distribution f
		fstable = quantum_Maxwellian(v, ushiftstable, Tstable, pow(parameters.hbar,parameters.ndim+1), zstable, parameters.type_qparticles, parameters.ndim+1);
		
		// Compute the rescaled distribution G
		//Gstable[k] = fstable / (mu * pow(pi*myomega/parameters.L, parameters.ndim+1));
		Gstable[k] = C*fstable;
		if (parameters.condensation){
			#pragma omp atomic
			Gstable[k] += Dirac[0][0][k];
		}
	}
	
	if (parameters.condensation){
		if ((parameters.type_qparticles == BOSONS) && (parameters.ndim == DIM3))
			free_3d(Dirac);
	}
}
