#include "./src/exec/kinebec.h"

int crossmode_done;
double ***crossmode_cos, ***crossmode_sin;

int nb_alloc_1d, nb_alloc_2d, nb_alloc_3d;
int nb_alloc_1d_char;
int nb_alloc_1d_int;

int nblocks, nthreads_per_block;


void print_help_kinebec(){
	message(BOLD RED"Error:"RESET" Wrong use of kinebec executable.\n");
	message("\n");
	
	message("The way to use this program is the following\n");
#ifdef key_cufft
	message("    kinebec <setup_file> -nb <number of CUDA blocks> -nt <number of threads per CUDA block> [options]\n");
#elif defined(key_mpi)
    message("    mpirun -np <nCPUs> kinebec <setup_file> [options]\n");
#else
	message("    kinebec <setup_file> [options]\n");
#endif
	message("\n");
	message("Options :\n");
	message(BOLD"    --resume"RESET"    Add this flag for resuming a previous run on the given setup file and that has been aborted.\n");
	message("                Warning: the run to be resumed must include result saving under HDF5/ADIOS format.\n");
#ifdef key_mpi
	// Close MPI environment
	MPI_Finalize();
#endif
	exit(0);
}



int main(int argc, char** argv){
	PARAMETERS parameters;
	char setupfile[MAXSTRINGLENGTH];
	int resume;
	int istatus;

	DIAGS mydiags;
	
	nb_alloc_1d = 0;
	nb_alloc_2d = 0;
	nb_alloc_3d = 0;
	nb_alloc_1d_char = 0;
	nb_alloc_1d_int = 0;
	
	int rank, commsize;
	int nthreads;

	int n;
	char buf[MAXSTRINGLENGTH];

	double tinit, tend, elap;
	
#ifdef key_mpi
	// Open MPI environment
	MPI_Init(&argc, &argv);

	// Find rank
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#else
	rank = 0;
	commsize = 1;
#endif

#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

	switch (argc){
		case 2:
#ifdef key_cufft
			print_help_kinebec();
#else
			strcpy(setupfile, argv[1]);
			resume = (1==0);
			nblocks = 1;
			nthreads_per_block = 1;
#endif
			break;
		case 3:
#ifdef key_cufft
			print_help_kinebec();
#else
			n = 1;
			resume = (1==0);
			while (n < argc){
				if (strcmp(argv[n], "--resume") == 0){
					resume = (1==1);
					n++;
				}
				else{
					strcpy(setupfile, argv[n]);
					n++;
				}
			}
			if (!resume)
				print_help_kinebec();
			
			nblocks = 1;
			nthreads_per_block = 1;
#endif
			break;
		case 6:
#ifdef key_cufft
            n = 1;
			nblocks = -1;
			nthreads_per_block = -1;
			resume = (1==0);
            while (n < argc){
                if (strcmp(argv[n], "-nb") == 0){
                    n++;
                    istatus = sscanf(argv[n], "%d", &nblocks);
                    n++;
                }
                else if (strcmp(argv[n], "-nt") == 0){
                    n++;
                    istatus = sscanf(argv[n], "%d", &nthreads_per_block);
                    n++;
                }
                else{
                    strcpy(setupfile, argv[n]);
                    n++;
                }
            }
			if ((nblocks < 0) || (nthreads_per_block < 0))
				print_help_kinebec();
#else
			print_help_kinebec();
#endif
			break;
		case 7:
#ifdef key_cufft
            n = 1;
			nblocks = -1;
			nthreads_per_block = -1;
			resume = (1==0);
            while (n < argc){
				if (strcmp(argv[n], "--resume") == 0){
					resume = (1==1);
					n++;
				}
                else if (strcmp(argv[n], "-nb") == 0){
                    n++;
                    istatus = sscanf(argv[n], "%d", &nblocks);
                    n++;
                }
                else if (strcmp(argv[n], "-nt") == 0){
                    n++;
                    istatus = sscanf(argv[n], "%d", &nthreads_per_block);
                    n++;
                }
                else{
                    strcpy(setupfile, argv[n]);
                    n++;
                }
            }
			if ((!resume) || (nblocks < 0) || (nthreads_per_block < 0))
				print_help_kinebec();
#else
			print_help_kinebec();
#endif
			break;
		default:
			print_help_kinebec();
	}


#ifdef key_fftw_mpi
	// Open FFTW-MPI environment
	fftw_mpi_init();
#endif

#ifdef key_fftw_openmp
	// Open FFTW-OpenMP environment
	istatus = fftw_init_threads();
	fftw_plan_with_nthreads(nthreads);
#endif

#ifdef key_mpi
	tinit = MPI_Wtime();
#elif defined(key_openmp)
	tinit = omp_get_wtime();
#else
	tinit = (double)clock();
#endif

	parameters = read_setup_file(setupfile);
	parameters.resume = resume;
	parameters.n_init = -1;

	crossmode_done = (1==0);

	// Build initial state
	message("Initialization... ");
	initialization(&mydiags, &parameters);
	message("done\n");

	// Run time loop
	time_loop(&mydiags, parameters);

	// Clean memory associated to final state
	free_diags(mydiags, parameters);

	// Clean memory associated to crossmodes
	free_crossmode(parameters);

	// Clean memory associated to parameters
	free_parameters(parameters);

#ifdef key_mpi
	tend = MPI_Wtime();
	elap = tend-tinit;
#elif defined(key_openmp)
	tend = omp_get_wtime();
	elap = tend-tinit;
#else
	tend = (double)clock();
	elap = (tend-tinit)/CLOCKS_PER_SEC;
#endif

	/*printf("nb_alloc_1d = %d\n", nb_alloc_1d);
	printf("nb_alloc_2d = %d\n", nb_alloc_2d);
	printf("nb_alloc_3d = %d\n", nb_alloc_3d);
	printf("nb_alloc_1d_char = %d\n", nb_alloc_1d_char);
	printf("nb_alloc_1d_int = %d\n", nb_alloc_1d_int);*/

	//printf("[%d] Elapsed time = %e\n", rank, elap);

#ifdef key_fftw_openmp
	// Clean FFTW-OpenMP environment
    fftw_cleanup_threads();
#endif

	exitsuccess("You can now visualize the results.");
}
