#include "./src/exec/spectral_accuracy.h"

#ifdef key_cufft
extern void collision_Qtotal_cufft(double *, double *, double **, double **, double, PARAMETERS);
extern void collision_Qtotal_quick_convolution_cufft(double *, double *, double **, double **, double *, double, PARAMETERS);
#endif

int crossmode_done;
double ***crossmode_cos, ***crossmode_sin;

int nb_alloc_1d, nb_alloc_2d, nb_alloc_3d;
int nb_alloc_1d_char;
int nb_alloc_1d_int;

int nblocks, nthreads_per_block;


void print_help_spectral_accuracy(){
	message(BOLD RED"Error:"RESET" Wrong use of spectral_accuracy executable.\n");
	message("\n");

	message("The way to use this program is the following\n");
#ifdef key_cufft
	message("    spectral_accuracy <setup_file> -nb <number of CUDA blocks> -nt <number of threads per CUDA block> \n");
#elif defined(key_mpi)
    message("    mpirun -np <nCPUs> spectral_accuracy <setup_file>\n");
#else
    message("    spectral_accuracy <setup_file>\n");
#endif

#ifdef key_mpi
	// Close MPI environment
	MPI_Finalize();
#endif
	exit(0);
}



void truncate_distribution(double *Gloc, PARAMETERS parameters){
    int NVXLOC, NVYLOC, NVZLOC, NVLOC;
    int IVX0LOC, IVY0LOC, IVZ0LOC;
    int MVX, MVY, MVZ, MV;

    double dxix, dxiy, dxiz;
    double pi = 4.*atan(1.);

    double xi[3];
    double xisquare;

    int kvxabs, kvyabs, kvzabs, itemp;

    int insupport;

    int k, d;

    NVXLOC = parameters.nvxloc;
    NVYLOC = parameters.nvyloc;
    NVZLOC = parameters.nvzloc;
    NVLOC = NVXLOC*NVYLOC*NVZLOC;
    IVX0LOC = parameters.ivx0loc;
    IVY0LOC = parameters.ivy0loc;
    IVZ0LOC = parameters.ivz0loc;
    MVX = parameters.nvx;
    MVY = parameters.nvy;
    MVZ = fmax(parameters.nvz, 1);
    MV = MVX*MVY*MVZ;

    dxix = 2.*pi/MVX;
    dxiy = 2.*pi/MVY;
    if (parameters.ndim == DIM3)
        dxiz = 2.*pi/MVZ;
    else
        dxiz = 1.;
    
    #pragma omp parallel private(k, d, kvxabs, kvyabs, kvzabs, itemp, xi, insupport) shared(NVLOC, NVXLOC, NVYLOC, NVZLOC, parameters, MVX, MVY, MVZ, IVX0LOC, IVY0LOC, IVZ0LOC, dxix, dxiy, dxiz, Gloc) 
    for (k = 0; k < NVLOC; k++){
#ifdef key_force_vectorize
		kvxabs = parameters.kvxabstab[k] - MVX*(parameters.kvxabstab[k] >= MVX/2);
		kvyabs = parameters.kvyabstab[k] - MVY*(parameters.kvyabstab[k] >= MVY/2);
		kvzabs = parameters.kvzabstab[k] - MVZ*(parameters.kvzabstab[k] >= MVZ/2);
#else
		itemp = k/NVZLOC;
		// itemp now contains kvyloc+NVYLOC*kvxloc;
		kvzabs = k - itemp*NVZLOC;
		kvxabs = itemp/NVYLOC;
		kvyabs = itemp-kvxabs*NVYLOC;
		kvxabs += IVX0LOC;
		kvyabs += IVY0LOC;
		kvzabs += IVZ0LOC;
		if (kvxabs >= MVX/2)
			kvxabs -= MVX;
		if (kvyabs >= MVY/2)
			kvyabs -= MVY;
		if (kvzabs >= MVZ/2)
			kvzabs -= MVZ;
#endif
		xi[0] = kvxabs * dxix;
		xi[1] = kvyabs * dxiy;
		if (parameters.ndim == DIM3)
			xi[2] = kvzabs * dxiz;
				
		xisquare = 0.;
		for (d = 0; d <= parameters.ndim; d++)
			xisquare += pow(xi[d],2);
		insupport = (xisquare <= pow(parameters.supportradius,2));

        #pragma omp atomic
        Gloc[k] *= insupport;
    }
}




int main(int argc, char** argv){
	PARAMETERS parameters;
	char setupfile[MAXSTRINGLENGTH];
    char resfile[MAXSTRINGLENGTH];
    char buf[MAXSTRINGLENGTH];
	int classical_accuracy, quantum_accuracy;

    int n;

	DIAGS mydiags;
    double *Gloc, *QGloc;
    double **alpha, **alphaprime;
    int NVXLOC, NVYLOC, NVZLOC, NVLOC;
    double Linfloc, Linf;
    double mu;
    double pi = 4.*atan(1.);
    //double ***A, ***B, ***C, *W;
    double **xitilde, **sigma, *W;
    int P;

    double contrib;

    FILE *f;
	
	nb_alloc_1d = 0;
	nb_alloc_2d = 0;
	nb_alloc_3d = 0;
	nb_alloc_1d_char = 0;
	nb_alloc_1d_int = 0;
	
	int rank, commsize;

    int k;

    int nthreads, istatus;
	
#ifdef key_mpi
	// Open MPI environment
	MPI_Init(&argc, &argv);

	// Find rank
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#else
	rank = 0;
	commsize = 1;
#endif

#ifdef key_openmp
	nthreads = omp_get_max_threads();
#else
	nthreads = 1;
#endif

	switch (argc){
		case 2:
#if defined(key_cufft)
            print_help_spectral_accuracy();
#else
            strcpy(setupfile, argv[1]);
            nblocks = 1;
            nthreads_per_block = 1;
#endif
			break;
        
        case 6:
#ifdef key_cufft
            n = 1;
            while (n < argc){
                if (strcmp(argv[n], "-nb") == 0){
                    n++;
                    istatus = sscanf(argv[n], "%d", &nblocks);
                    n++;
                }
                else if (strcmp(argv[n], "-nt") == 0){
                    n++;
                    istatus = sscanf(argv[n], "%d", &nthreads_per_block);
                    n++;
                }
                else{
                    strcpy(setupfile, argv[n]);
                    n++;
                }
            }
#else
            print_help_spectral_accuracy();
#endif
            break;
		default:
			print_help_spectral_accuracy();
	}


#ifdef key_fftw_mpi
	// Open FFTW-MPI environment
	fftw_mpi_init();
#endif

#ifdef key_fftw_openmp
	// Open FFTW-OpenMP environment
	istatus = fftw_init_threads();
	fftw_plan_with_nthreads(nthreads);
#endif

    // Read setup file
	parameters = read_setup_file(setupfile);
	parameters.resume = (1==0);
	parameters.n_init = -1;

	crossmode_done = (1==0);

    // Verify if the setup file can be used
    classical_accuracy = ((parameters.type_f0 == CLASSICAL_MAXWELLIAN) && (parameters.type_collision == CLASSICAL_COLLISIONS));
    quantum_accuracy = ((parameters.type_f0 == QUANTUM_MAXWELLIAN) && (parameters.type_collision == QUANTUM_COLLISIONS));

    if ((!classical_accuracy) && (!quantum_accuracy))
        error("Spectral accuracy can be run only for test cases that are initialized with their limit state.");

	// Build initial state
    message("Discretizing the distribution... ");
	initialization(&mydiags, &parameters);
    message("done\n");

    // Verify if there is no condensation
    if (parameters.condensation)
        error("Spectral accuracy cannot be run when quantum degeneracy occurs.");
    
    NVXLOC = parameters.nvxloc;
    NVYLOC = parameters.nvyloc;
    NVZLOC = parameters.nvzloc;
    NVLOC = NVXLOC*NVYLOC*NVZLOC;

    // Allocate Q(Glimit)
    QGloc = allocation_1d(NVLOC);

    // Compute cross-section coefficients
    if (parameters.force_convolution){
		/*A = allocation_3d(parameters.mint, NVLOC, 2);
		B = allocation_3d(parameters.mint, NVLOC, 2);
		C = allocation_3d(parameters.mint, NVLOC, 2);
		W = allocation_1d(parameters.mint);
		compute_ABCW_force_convolution(W, A, B, C, parameters);*/
      	compute_points_weights_force_convolution(parameters.cutoff, parameters.P1_int, parameters.P2_int, parameters.P3_int, parameters.ndim+1, &xitilde, &sigma, &W, &P);
        if (P != parameters.mint)
            error("Problem with the number of integration points.");
    }
    else{
        alpha = allocation_2d(parameters.mint, NVLOC);
	    alphaprime = allocation_2d(parameters.mint, NVLOC);
	    expanded_collision_modes(alpha, alphaprime, parameters);
    }
    
    switch (parameters.ndim){
        case DIM2:
            Gloc = mydiags.Gloclimit_2d[0];
            break;
        case DIM3:
            Gloc = mydiags.Gloclimit_3d[0][0];
            break;
        default:
            error("Unknown dimensionality");
    }
    message("Cross-section modes computed\n");
    //truncate_distribution(Gloc, parameters);

#ifdef key_cufft
    if (parameters.force_convolution)
        collision_Qtotal_quick_convolution_cufft(Gloc, QGloc, xitilde, sigma, W, mydiags.omega, parameters);
    else
    	collision_Qtotal_cufft(Gloc, QGloc, alpha, alphaprime, mydiags.omega, parameters);
    message("Collision term computed\n");
#elif defined(key_fftw) || defined(key_fftw_mpi) || defined(key_fftw_openmp)
    if (parameters.force_convolution){
        double t1, t2;
        /*t1 = (double)clock();
        collision_Qtotal_force_convolution_fftw(Gloc, QGloc, A, B, C, W, mydiags.omega, parameters);
        t2 = (double)clock();
        printf("elapsed time (force): %e\n", (t2-t1)/CLOCKS_PER_SEC);*/

        t1 = (double)clock();
        //collision_Qtotal_quick_convolution_fftw(Gloc, QGloc, A, B, C, W, mydiags.omega, parameters);
        collision_Qtotal_quick_convolution_fftw(Gloc, QGloc, xitilde, sigma, W, mydiags.omega, parameters);
        t2 = (double)clock();
        printf("elapsed time (quick): %e\n", (t2-t1)/CLOCKS_PER_SEC);
    }
    else
        collision_Qtotal_fftw(Gloc, QGloc, alpha, alphaprime, mydiags.omega, parameters);
    message("Collision term computed\n");
#endif
    if (parameters.type_omega == OMEGA_IS_NOTUSED)
        mu = pow(parameters.L/mydiags.omega, parameters.ndim+1);
    else
        mu = 1.;
    
    #pragma omp parallel for private(k, contrib) shared(NVLOC, QGloc, pi, mydiags, parameters, mu)
    for (k = 0; k < NVLOC; k++){
        contrib = pow(pi*mydiags.omega/parameters.L, parameters.ndim+1-parameters.gamma)*pow(mu,2);
        #pragma omp atomic
        QGloc[k] *= contrib;
    }

    Linfloc = 0.;
    #pragma omp parallel for private(k) shared(NVLOC, Linfloc, QGloc)
    for (int k = 0; k < NVLOC; k++){
        contrib = Linfloc*(fabs(QGloc[k]) <= Linfloc) + fabs(QGloc[k])*(fabs(QGloc[k]) > Linfloc) - Linfloc;
        #pragma omp atomic
        Linfloc += contrib;
    }
#ifdef key_mpi
    MPI_Allreduce(&Linfloc, &Linf, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
#else
    Linf = Linfloc;
#endif

    // Save data
    strcpy(resfile, "");
	strcat(resfile, parameters.outputdir);
	strcat(resfile, "/");
	strcat(resfile, parameters.prefix);
    strcat(resfile, ".h5");
    save_all_data(mydiags, 0., 1., resfile, parameters);

    // Save Linf-norm of the residual
    strcpy(resfile, "");
	strcat(resfile, parameters.outputdir);
	strcat(resfile, "/");
	strcat(resfile, parameters.prefix);
    strcat(resfile, "_LinfnormQ.dat");
    if (rank == 0){
        f = fopen(resfile, "w");
        fprintf(f, "%e %.15e", 0., Linf);
        fclose(f);
    }

    printf("residual = %e\n", Linf);

    // Clean memory of cross-section
    if (parameters.force_convolution){
        /*free_3d(A);
        free_3d(B);
        free_3d(C);
        free_1d(W);*/
        free_2d(xitilde);
        free_2d(sigma);
        free_1d(W);
    }
    else{
        free_2d(alpha);
        free_2d(alphaprime);
    }

    // Clean memory of Q(Glimit)
    free_1d(QGloc);

	// Clean memory associated to final state
	free_diags(mydiags, parameters);

	// Clean memory associated to crossmodes
	free_crossmode(parameters);

	// Clean memory associated to parameters
	free_parameters(parameters);

#ifdef key_fftw_openmp
	// Clean FFTW-OpenMP environment
    fftw_cleanup_threads();
#endif

	exitsuccess("You can now visualize the results.");
}